﻿module Game.CController;
import std.array, std.conv, std.json, std.stream, std.string, std.typecons;
import derelict.opengl3.gl3;
import Tractor.Controller, Tractor.Tractor;
import Game.CGameInfo, Game.CMenuController, Game.CInstructionsController, Game.IUIEvents;
import QBasic.CCP437Encoder, QBasic.IInputExternal, QBasic.Interpreter.CInterpreter, QBasic.Parser.CParser,
		QBasic.Program.CProgram;
import Render.CDisplay;
import Windows.CFileSystem;

enum EGameState { Menu, Game, Instructions }

class CController : IController, IUIEvents, IInterpreterLog, IInputExternal
{
	private ITractor mTractor;
	private CGameInfo mGameInfo;
	private EGameState mGameState;
	private CMenuController mMenu;
	private CInstructionsController mInstructions;
	private CDisplay mDisplay;
	private CInterpreter mInterpreter;
	private CFileSystem mFileSystem;
	private CProgram mProgram = null;
	private Nullable!(QBasic.EKey.EKey) mCurrentKey;
	private enum int eKeyQueueMaxSize = 6;
	private QBasic.EKey.EKey[] mKeyQueue;
	private enum int eKeyMinDelay = 80;
	private ITimer mKeyMinTimer;
	private int mKeyRepeatDelay = 90;
	private ITimer mKeyRepeatTimer;
	private bool mKeyKillSwitch = false;

	public ~this()
	{
		delete mMenu;
		delete mInstructions;
	}



	public CControllerOptions GetOptions()
	{
		return new CControllerOptions(960, 576, "Otheon's Adventure");
	}



	public void Initialise(ITractor aTractor)
	{
		mTractor = aTractor;
		LoadGameInfo();
		mGameState = EGameState.Menu;
		mMenu = new CMenuController(aTractor, this, mGameInfo);
		mMenu.Initialise();
		mInstructions = new CInstructionsController(aTractor, this, mGameInfo);
		mDisplay = new CDisplay(mTractor);
		mDisplay.Initialise();
		Stream QBStream = mTractor.Core.OpenDataFile("RPG.BAS");
		CCP437Encoder Encoder = new CCP437Encoder();
		CParser Parser = new CParser(Encoder, QBStream);
		Parser.Parse();
		QBStream.close();
		if (!Parser.Success())
			mTractor.Core.Log(format("Failed to load RPG.BAS. Error on line %d, char %d: %s",
					Parser.GetErrorLine() + 1, Parser.GetErrorPos(), to!string(Parser.GetError())));
		else
		{
			mProgram = Parser.GetProgram();
			mFileSystem = new CFileSystem(mTractor);
		}
		mMenu.Reset(false);
		mKeyQueue.reserve(eKeyQueueMaxSize);
		mKeyMinTimer = mTractor.Core.CreateTimer(&KeyMinTimer);
		mKeyRepeatTimer = mTractor.Core.CreateTimer(&KeyRepeatTimer);
	}



	public void OnFrame()
	{
		final switch (mGameState)
		{
			case EGameState.Menu:
				mMenu.Render();
				break;
			case EGameState.Game:
				mDisplay.Render();
				break;
			case EGameState.Instructions:
				mInstructions.Render();
				break;
		}
	}



	public void OnKeyDown(EKey aKey)
	{
		final switch (mGameState)
		{
			case EGameState.Menu:
				mMenu.OnKeyDown(aKey);
				break;
			case EGameState.Game:
				GameKeyPress(aKey);
				break;
			case EGameState.Instructions:
				mInstructions.OnKeyDown(aKey);
				break;
		}
	}



	private void GameKeyPress(EKey aKey)
	{
		if (aKey == EKey.Escape)
		{
			mGameState = EGameState.Menu;
			mMenu.Reset(true);
			mTractor.Core.Redraw();
			return;
		}
		if (!mInterpreter.IsWaitingForKey)
			return;
		auto QBKey = KeyToQBKey(aKey);
		if (QBKey.isNull || (!mCurrentKey.isNull && QBKey == mCurrentKey) || mKeyKillSwitch)
			return;
		mCurrentKey = QBKey;
		if (!mKeyMinTimer.Started)
		{
			mKeyMinTimer.Start(eKeyMinDelay);
			if (mKeyQueue.length == 0)
			{
				if (mKeyRepeatTimer.Started)
					mKeyRepeatTimer.Stop();
				mKeyRepeatTimer.Start(mKeyRepeatDelay);
			}
			mInterpreter.SetKey(QBKey);
			Step();
			if (!mInterpreter.IsWaitingForKey)
			{
				mKeyMinTimer.Stop();
				mKeyRepeatTimer.Stop();
			}
		}
		else
		{
			if (mKeyQueue.length < eKeyQueueMaxSize)
				mKeyQueue ~= QBKey;
		}
	}



	private void KeyMinTimer()
	{
		if (mKeyQueue.length != 0)
		{
			mInterpreter.SetKey(mKeyQueue[0]);
			mKeyQueue.replaceInPlace(0, 1, cast(QBasic.EKey.EKey[])[]);
			Step();
			if (!mInterpreter.IsWaitingForKey)
			{
				mKeyMinTimer.Stop();
				if (mKeyRepeatTimer.Started)
					mKeyRepeatTimer.Stop();
				return;
			}
			if (mKeyQueue.length == 0 && !mCurrentKey.isNull)
				mKeyRepeatTimer.Start(mKeyRepeatDelay);
		}
		if (mKeyQueue.length == 0)
			mKeyMinTimer.Stop();
	}



	private void KeyRepeatTimer()
	in
	{
		assert(!mCurrentKey.isNull);
	}
	body
	{
		assert(mInterpreter.IsWaitingForKey);
		mKeyMinTimer.Start(eKeyMinDelay);
		mInterpreter.SetKey(mCurrentKey);
		Step();
		if (!mInterpreter.IsWaitingForKey)
		{
			mKeyMinTimer.Stop();
			mKeyRepeatTimer.Stop();
		}
	}



	public void OnKeyUp(EKey aKey)
	{
		if (mKeyKillSwitch)
		{
			mKeyKillSwitch = false;
			return;
		}
		auto QBKey = KeyToQBKey(aKey);
		if (QBKey.isNull || mCurrentKey.isNull || QBKey != mCurrentKey)
			return;
		mCurrentKey = Nullable!(QBasic.EKey.EKey)();
		mKeyRepeatTimer.Stop();
	}



	private Nullable!(QBasic.EKey.EKey) KeyToQBKey(EKey aKey)
	{
		switch (aKey)
		{
			case EKey.Left:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Left);
			case EKey.Right:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Right);
			case EKey.Up:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Up);
			case EKey.Down:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Down);
			case EKey.Enter:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Enter);
			case EKey.Delete:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Delete);
			case EKey.Escape:
				return Nullable!(QBasic.EKey.EKey)(QBasic.EKey.EKey.Escape);
			default:
				return Nullable!(QBasic.EKey.EKey)();
		}
	}



	public void KeyRate(int aRate)
	{
		mKeyRepeatDelay = aRate;
		if (mKeyRepeatTimer.Started)
		{
			mKeyRepeatTimer.Stop();
			mKeyRepeatTimer.Start(mKeyRepeatDelay);
		}
	}



	public void KeyStop()
	{
		if (!mCurrentKey.isNull)
		{
			mCurrentKey = Nullable!(QBasic.EKey.EKey)();
			mKeyRepeatTimer.Stop();
			mKeyKillSwitch = true;
		}
		mKeyQueue.length = 0;
	}



	public void NewGame()
	{
		StartGame("-newgame");
	}



	public void LoadGame(int aSlot)
	{
		StartGame(to!string(aSlot));
	}



	public void Resume()
	{
		mGameState = EGameState.Game;
		mTractor.Core.Redraw();
	}



	public void Instructions()
	{
		mGameState = EGameState.Instructions;
		mTractor.Core.Redraw();
	}



	public void MainMenu()
	{
		mGameState = EGameState.Menu;
		mTractor.Core.Redraw();
	}



	private void StartGame(string aCommandLine)
	{
		mGameState = EGameState.Game;
		mDisplay.Reset();
		mInterpreter = new CInterpreter(mFileSystem, mDisplay, this, mProgram, this, aCommandLine);
		mDisplay.SetScreen(mInterpreter.ScreenData, mInterpreter.ScreenXSize, mInterpreter.ScreenYSize);
		Step();
	}



	private void Step()
	{
		mInterpreter.Step();
		mTractor.Core.Redraw();
		if (mInterpreter.Error)
			mTractor.Core.Log(format("Run-time error on line %d char %d: %s",
					mInterpreter.ErrorLine + 1, mInterpreter.ErrorPos, mInterpreter.ErrorMessage));
		else if (!mInterpreter.IsWaitingForKey)
			mTractor.Core.Quit();
	}



	public void TraceLog(string aLog)
	{
		mTractor.Core.Log(aLog);
	}



	private void LoadGameInfo()
	{
		auto GameInfoFile = mTractor.Core.OpenDataFile("GameInfo.json");
		JSONValue GameInfoJSON = parseJSON(GameInfoFile.toString());
		mTractor.Core.CloseFile(GameInfoFile);
		string Title = GameInfoJSON["Title"].str;
		string TitleFont = GameInfoJSON["TitleFont"].str;
		int TitleFontSize = cast(int)GameInfoJSON["TitleFontSize"].integer;
		string MenuFont = GameInfoJSON["MenuFont"].str;
		int MenuFontSize = cast(int)GameInfoJSON["MenuFontSize"].integer;
		mGameInfo = new CGameInfo(Title, TitleFont, TitleFontSize, MenuFont, MenuFontSize);
	}
}