﻿module Game.CMenuController;
import std.file, std.string;
import derelict.opengl3.gl3;
import Tractor.Tractor;
import Game.CGameInfo, Game.IUIEvents;
import Render.CScene;
import Text.CProcessor, Text.CText;

class CMenuController
{
	private ITractor mTractor;
	private IUIEvents mUIEvents;
	private CGameInfo mGameInfo;
	private int mMenuY;
	private ITexture mTitleImage;
	private ITexture mCursorImage;
	private IFont mFont;
	private CScene mScene;
	private uint mTitleSprite;
	private uint mCursorSprite;
	private int mNumStaticSprites;
	private int mMenu;
	private int mNumMenuItems;
	private bool mHasResume;
	private bool[4] mSlotAvailable;

	public this(ITractor aTractor, IUIEvents aUIEvents, CGameInfo aGameInfo)
	{
		mTractor = aTractor;
		mUIEvents = aUIEvents;
		mGameInfo = aGameInfo;
	}



	public ~this()
	{
		delete mScene;
	}



	public void Initialise()
	{
		mTitleImage = mTractor.Textures.LoadImage("Title.png");
		glBindTexture(GL_TEXTURE_2D, mTitleImage.ID);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		mCursorImage = mTractor.Textures.LoadImage("Menu cursor.png");
		glBindTexture(GL_TEXTURE_2D, mCursorImage.ID);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		assert(glGetError() == GL_NO_ERROR);
		mFont = mTractor.Fonts.LoadFont(mGameInfo.MenuFont);
		mScene = new CScene(mTractor, 20);
		mTitleSprite = mScene.CreateSprite();
		mScene.SetSprite(mTitleSprite, 0, 0, 960, 576, mTitleImage);
		mCursorSprite = mScene.CreateSprite();
		auto Processor = new CProcessor();
		auto Text = new CText(mTractor, Processor, mScene, 576, 480, 35, 920, 0.5f, 0, 0);
		Text.SetFont(mTractor.Fonts.LoadFont(mGameInfo.TitleFont), mGameInfo.TitleFontSize);
		Text.Write(mGameInfo.Title);
		Text.End();
		mNumStaticSprites = mScene.NumSprites;
	}



	public void Reset(bool aHasResume)
	{
		mHasResume = aHasResume;
		mMenu = 0;
		mMenuY = 0;
		UpdateMenuText();
		UpdateCursor();
	}



	private void UpdateMenuText()
	{
		mScene.Reset(mNumStaticSprites);
		auto Processor = new CProcessor();
		final switch (mMenu)
		{
			case 0:
				mNumMenuItems = mHasResume ? 5 : 4;
				break;
			case 1:
				mNumMenuItems = 5;
				break;
		}
		auto Text = new CText(mTractor, Processor, mScene, 576, 400, 544 - mNumMenuItems * 50,
				-1, 0, 23, 0);
		Text.SetFont(mFont, mGameInfo.MenuFontSize);
		final switch (mMenu)
		{
			case 0:
				if (mHasResume)
					Text.Write("Resume\n");
				Text.Write("New Game\n");
				Text.Write("Load Game\n");
				Text.Write("Instructions\n");
				Text.Write("Quit");
				break;
			case 1:
				Text.Write("Back\n");
				auto Files = mTractor.Core.GetLocalFiles();
				for (int i = 0; i < 4; i++)
				{
					string Filename = format("save%d.sav", i);
					DirEntry* File = Filename in Files;
					mSlotAvailable[i] = File !is null;
					if (File is null)
						Text.Write("-\n");
					else
						Text.Write(format("Slot %d - %s\n", i, mTractor.Core.FormatDateTime(File.timeLastModified)));
				}
				break;
		}
		Text.End();
	}



	private void UpdateCursor()
	{
		mScene.SetSprite(mCursorSprite, 375, 545 - (mNumMenuItems - mMenuY) * 50, 13, 17, mCursorImage);
	}



	public void Render()
	{
		mScene.Render();
	}



	public void OnKeyDown(EKey aKey)
	{
		switch (aKey)
		{
			case EKey.Up:
				if (mMenuY != 0)
				{
					mMenuY--;
					UpdateCursor();
					mTractor.Core.Redraw();
				}
				break;
			case EKey.Down:
				if (mMenuY != mNumMenuItems - 1)
				{
					mMenuY++;
					UpdateCursor();
					mTractor.Core.Redraw();
				}
				break;
			case EKey.Enter:
				MenuItem();
				break;
			default:
				break;
		}
	}



	private void MenuItem()
	{
		final switch (mMenu)
		{
			case 0:
				int Option = mHasResume ? mMenuY : mMenuY + 1;
				final switch (Option)
				{
					case 0:
						mUIEvents.Resume();
						break;
					case 1:
						mUIEvents.NewGame();
						break;
					case 2:
						ChangeMenu(1);
						break;
					case 3:
						mUIEvents.Instructions();
						break;
					case 4:
						mTractor.Core.Quit();
						break;
				}
				break;
			case 1:
				if (mMenuY == 0)
					ChangeMenu(0);
				else if (mSlotAvailable[mMenuY - 1])
					mUIEvents.LoadGame(mMenuY - 1);
				break;
		}
	}



	private void ChangeMenu(int aMenu)
	{
		mMenu = aMenu;
		mMenuY = 0;
		UpdateMenuText();
		UpdateCursor();
		mTractor.Core.Redraw();
	}
}