﻿module Render.CScene;
import derelict.opengl3.gl3;
import Tractor.Tractor;

class CScene
{
	private ITractor mTractor;
	private int mMaxSprites = 0;
	private int mNumSprites = 0;
	private GLuint mVertexArray;
	private GLuint mPositionsVABuffer;
	private GLuint mTextureCoordsVABuffer;
	private IShaderProgram mSpriteProgram;
	private float[] mPositionData;
	private float[] mTextureCoordData;
	private ITexture[] mSpriteTextures;

	public this(ITractor aTractor, int aMaxSprites)
	{
		mTractor = aTractor;
		mMaxSprites = aMaxSprites;
		mPositionData = new float[mMaxSprites * 8];
		mTextureCoordData = new float[mMaxSprites * 8];
		mSpriteTextures = new ITexture[mMaxSprites];
		glGenVertexArrays(1, &mVertexArray);
		glBindVertexArray(mVertexArray);
		glGenBuffers(1, &mPositionsVABuffer);
		glBindBuffer(GL_ARRAY_BUFFER, mPositionsVABuffer);
		glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, null);
		glEnableVertexAttribArray(0);
		glGenBuffers(1, &mTextureCoordsVABuffer);
		glBindBuffer(GL_ARRAY_BUFFER, mTextureCoordsVABuffer);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, null);
		glEnableVertexAttribArray(1);
		auto SpriteVertexShader = mTractor.Shaders.LoadCustomVertexShader("SpriteVert.glsl");
		auto SpriteFragmentShader = mTractor.Shaders.LoadCustomFragmentShader("SpriteFrag.glsl");
		mSpriteProgram = mTractor.Shaders.CreateCustomProgram(SpriteVertexShader, SpriteFragmentShader);
		mSpriteProgram.BindAttribute(0, "Position");
		mSpriteProgram.BindAttribute(1, "TextureCoordIn");
		mSpriteProgram.Link();
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}



	public ~this()
	{
		glDeleteVertexArrays(1, &mVertexArray);
		glDeleteBuffers(1, &mPositionsVABuffer);
		glDeleteBuffers(1, &mTextureCoordsVABuffer);
		glBindTexture(GL_TEXTURE_2D, 0);
	}



	public void Reset(int aKeepSprites)
	{
		mNumSprites = aKeepSprites;
	}



	public int NumSprites()
	{
		return mNumSprites;
	}



	public uint CreateSprite()
	in
	{
		assert(mNumSprites < mMaxSprites);
	}
	body
	{
		return mNumSprites++;
	}



	public void SetSprite(uint aID, int aX, int aY, int aXSize, int aYSize, ITexture aTexture)
	in
	{
		assert(aID < mNumSprites);
	}
	body
	{
		mPositionData[aID * 8] = aX;
		mPositionData[aID * 8 + 1] = aY;
		mPositionData[aID * 8 + 2] = aX + aXSize;
		mPositionData[aID * 8 + 3] = aY;
		mPositionData[aID * 8 + 4] = aX + aXSize;
		mPositionData[aID * 8 + 5] = aY + aYSize;
		mPositionData[aID * 8 + 6] = aX;
		mPositionData[aID * 8 + 7] = aY + aYSize;
		mTextureCoordData[aID * 8 + 0] = 0;
		mTextureCoordData[aID * 8 + 1] = 0;
		mTextureCoordData[aID * 8 + 2] = aTexture.TextureXSize;
		mTextureCoordData[aID * 8 + 3] = 0;
		mTextureCoordData[aID * 8 + 4] = aTexture.TextureXSize;
		mTextureCoordData[aID * 8 + 5] = aTexture.TextureYSize;
		mTextureCoordData[aID * 8 + 6] = 0;
		mTextureCoordData[aID * 8 + 7] = aTexture.TextureYSize;
		mSpriteTextures[aID] = aTexture;
	}



	public void MoveSprite(uint aID, int aDeltaX, int aDeltaY)
	{
		mPositionData[aID * 8] += aDeltaX;
		mPositionData[aID * 8 + 1] += aDeltaY;
		mPositionData[aID * 8 + 2] += aDeltaX;
		mPositionData[aID * 8 + 3] += aDeltaY;
		mPositionData[aID * 8 + 4] += aDeltaX;
		mPositionData[aID * 8 + 5] += aDeltaY;
		mPositionData[aID * 8 + 6] += aDeltaX;
		mPositionData[aID * 8 + 7] += aDeltaY;
	}



	public void Render()
	{
		glBindBuffer(GL_ARRAY_BUFFER, mPositionsVABuffer);
		glBufferData(GL_ARRAY_BUFFER, mPositionData.length * float.sizeof, mPositionData.ptr,
				GL_STATIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, mTextureCoordsVABuffer);
		glBufferData(GL_ARRAY_BUFFER, mTextureCoordData.length * float.sizeof,
				mTextureCoordData.ptr, GL_STATIC_DRAW);
		glBindVertexArray(mVertexArray);
		mTractor.Shaders.UseCustomProgram(mSpriteProgram);
		GLint SpriteTextureLocation = glGetUniformLocation(mSpriteProgram.ID, "SpriteTexture");
		glUniform1i(SpriteTextureLocation, 0);
		for (int i = 0; i < mNumSprites; i++)
		{
			glBindTexture(GL_TEXTURE_2D, mSpriteTextures[i].ID);
			glDrawArrays(GL_TRIANGLE_FAN, i * 4, 4);
		}
		assert(glGetError() == GL_NO_ERROR);
	}
}