﻿module Text.CProcessor;
import std.algorithm, std.utf;
import freetype.ft;
import Tractor.Tractor;
import Text.CTextFragment, Text.CWrite;

class CIntermediate
{
	public int Line;
	public FT_Pos X;
	public CWrite Write;
	public int TextStartIndex;
	public int TextEndIndex;

	public this(int aLine, FT_Pos aX, CWrite aWrite, int aTextStartIndex)
	in
	{
		assert(aTextStartIndex >= 0 && aTextStartIndex < aWrite.Text.length);
	}
	body
	{
		Line = aLine;
		X = aX;
		Write = aWrite;
		TextStartIndex = aTextStartIndex;
		TextEndIndex = aTextStartIndex;
	}
}



class CLineInfo
{
	public FT_F26Dot6 XSize;
	public FT_F26Dot6 AscenderSize;
	public FT_F26Dot6 DescenderSize;

	public this(FT_F26Dot6 aXSize, FT_F26Dot6 aAscenderSize, FT_F26Dot6 aDescenderSize)
	{
		XSize = aXSize;
		AscenderSize = aAscenderSize;
		DescenderSize = aDescenderSize;
	}
}



enum EElementProcessMode { None, Spaces, Word }

class CProcessor
{
	private bool mWrap;
	private FT_F26Dot6 mXSize;
	private FT_Pos mX;
	private int mCurrentCharacterIndex;
	private EElementProcessMode mElementProcessMode;
	private FT_F26Dot6 mSpacesXSize;
	private FT_F26Dot6 mWordXSize;
	private FT_F26Dot6 mWordMaxAscenderSize;
	private FT_F26Dot6 mWordMaxDescenderSize;
	private int mWordStartCharacterIndex;
	private bool mLineStarted;
	private FT_F26Dot6 mLineXSize;
	private FT_F26Dot6 mLineMaxAscenderSize;
	private FT_F26Dot6 mLineMaxDescenderSize;
	private CIntermediate[] mIntermediates;
	private string mCurrentIntermediateString;
	private int mCommittedIntermediateIndex;
	private int mCommittedCharacterIndex;
	private CLineInfo[] mLines;

	public CTextFragment[] Process(FT_Pos aOriginX, FT_Pos aOriginY, FT_F26Dot6 aXSize,
			float aXAlign, FT_F26Dot6 aLineSpacing, float aYAlign, CWrite[] aWrites)
	in
	{
		assert(aWrites.length != 0);
	}
	body
	{
		mXSize = aXSize;
		mWrap = mXSize != -1;
		mX = 0;
		mElementProcessMode = EElementProcessMode.None;
		mLineStarted = false;
		mCommittedIntermediateIndex = 0;
		mCommittedCharacterIndex = 0;
		foreach (int i, CWrite Write; aWrites)
		{
			if (!mLineStarted)
				StartLine();
			mCurrentCharacterIndex = 0;
			if (Write.IsImage)
			{
				mIntermediates ~= new CIntermediate(mLines.length, mX, Write, -1);
				ProcessImage(Write.Texture, Write.Font, Write.FontSize);
			}
			else
			{
				mIntermediates ~= new CIntermediate(mLines.length, mX, Write, 0);
				size_t Index = 0;
				while (Index < Write.Text.length)
				{
					auto LastIndex = Index;
					dchar Char = decode(Write.Text, Index);
					ProcessCharacter(Char, Write.Font, Write.FontSize);
					mCurrentCharacterIndex += Index - LastIndex;
				}
			}
		}
		if (mLineStarted)
		{
			if (mElementProcessMode != EElementProcessMode.None)
				CommitWord();
			EndLine();
		}
		CTextFragment[] Fragments;
		FT_F26Dot6 TotalYSize = mLines.map!(l => l.AscenderSize + l.DescenderSize)
				.reduce!((y0, y1) => y0 + y1) + (mLines.length - 1) * aLineSpacing;
		FT_Pos BaselineY;
		FT_Pos NextLineY = aOriginY + cast(FT_F26Dot6)(TotalYSize * aYAlign);
		int LastLineIndex = -1;
		foreach (int i, CIntermediate Intermediate; mIntermediates)
		{
			auto Write = Intermediate.Write;
			auto Line = mLines[Intermediate.Line];
			if (Intermediate.Line != LastLineIndex)
			{
				BaselineY = NextLineY - Line.AscenderSize;
				NextLineY -= Line.AscenderSize + Line.DescenderSize + aLineSpacing;
				LastLineIndex = Intermediate.Line;
			}
			FT_Pos X = aOriginX + Intermediate.X - cast(FT_F26Dot6)(Line.XSize * aXAlign);
			FT_F26Dot6 XSize;
			if (i + 1 == mIntermediates.length || mIntermediates[i + 1].Line != Intermediate.Line)
				XSize = Line.XSize - Intermediate.X;
			else
				XSize = mIntermediates[i + 1].X - Intermediate.X;
			if (Intermediate.Write.IsImage)
			{
				auto YSize = ToFTSize(Write.Texture.YSize);
				Fragments ~= new CTextFragment(Write.Colour, Write.Alpha, Write.Texture, X,
						BaselineY, BaselineY, XSize, YSize);
			}
			else
			{
				auto DescenderSize = Write.Font.DescenderSize(Write.FontSize);
				auto YSize = Write.Font.AscenderSize(Write.FontSize) - DescenderSize;
				auto Text = Intermediate.Write.Text[Intermediate.TextStartIndex ..
						Intermediate.TextEndIndex];
				Fragments ~= new CTextFragment(Write.Font, Write.FontSize, Write.Colour,
						Write.Alpha, Text, X, BaselineY + DescenderSize, BaselineY, XSize,
						YSize);
			}
		}
		return Fragments;
	}



	private void ProcessCharacter(dchar aChar, IFont aFont, FT_F26Dot6 aFontSize)
	{
		if (aChar == '\n')
		{
			if (mElementProcessMode != EElementProcessMode.None)
				CommitWord();
			// TODO: need to insert a blank line if the line hasn't been started
			EndLine();
			auto LastIntermediate = mIntermediates[$ - 1];
			if (mCurrentCharacterIndex + 1 != LastIntermediate.Write.Text.length)
			{
				StartLine();
				mIntermediates ~= new CIntermediate(mLines.length, mX, LastIntermediate.Write,
						mCurrentCharacterIndex + 1);
				LastIntermediate.TextEndIndex = mCurrentCharacterIndex;
			}
		}
		else
		{
			auto Pen = FT_Vector(0, 0);
			auto CharacterXSize = aFont.GetCharacterXSize(aChar, Pen, aFontSize);
			ProcessElement(aChar == ' ', CharacterXSize, aFont.AscenderSize(aFontSize),
					-aFont.DescenderSize(aFontSize));
		}
	}



	private FT_F26Dot6 ToFTSize(int aSize)
	{
		return aSize * 64;
	}



	private void ProcessImage(ITexture aTexture, IFont aFont, FT_F26Dot6 aFontSize)
	{
		ProcessElement(false, ToFTSize(aTexture.XSize), ToFTSize(aTexture.YSize),
				-aFont.DescenderSize(aFontSize));
	}



	private void ProcessElement(bool aIsSpace, FT_F26Dot6 aXSize, FT_F26Dot6 aAscenderSize,
			FT_F26Dot6 aDescenderSize)
	in
	{
		assert(!mWrap || aXSize < mXSize);   // TODO: support elements bigger than the wrap size. put them on a line by themselves
	}
	body
	{
		final switch (mElementProcessMode)
		{
			case EElementProcessMode.None:
				StartWord(aIsSpace);
				break;
			case EElementProcessMode.Spaces:
				if (!aIsSpace)
				{
					mElementProcessMode = EElementProcessMode.Word;
					mWordStartCharacterIndex = mCurrentCharacterIndex;
					if (mWrap && mX + aXSize > mXSize)
						RestartLine();
				}
				break;
			case EElementProcessMode.Word:
				if (aIsSpace)
				{
					CommitWord();
					StartWord(true);
				}
				break;
		}
		mX += aXSize;
		if (aIsSpace)
			mSpacesXSize += aXSize;
		else
			mWordXSize += aXSize;
		mWordMaxAscenderSize = max(mWordMaxAscenderSize, aAscenderSize);
		mWordMaxDescenderSize = max(mWordMaxDescenderSize, aDescenderSize);
	}



	private void StartWord(bool aIsSpace)
	in
	{
		assert(mElementProcessMode == EElementProcessMode.None);
	}
	body
	{
		mElementProcessMode = aIsSpace ? EElementProcessMode.Spaces : EElementProcessMode.Word;
		mSpacesXSize = 0;
		mWordXSize = 0;
		mWordMaxAscenderSize = 0;
		mWordMaxDescenderSize = 0;
	}



	private void CommitWord()
	in
	{
		assert(mElementProcessMode != EElementProcessMode.None);
	}
	body
	{
		mElementProcessMode = EElementProcessMode.None;
		mIntermediates[$ - 1].TextEndIndex = mCurrentCharacterIndex;
		mCommittedIntermediateIndex = mIntermediates.length - 1;
		mCommittedCharacterIndex = mCurrentCharacterIndex;
		mLineXSize = mX;
		mLineMaxAscenderSize = max(mLineMaxAscenderSize, mWordMaxAscenderSize);
		mLineMaxDescenderSize = max(mLineMaxDescenderSize, mWordMaxDescenderSize);
		mWordStartCharacterIndex = -1;
	}



	private void StartLine()
	in
	{
		assert(!mLineStarted);
	}
	body
	{
		mLineStarted = true;
		mX = 0;
		mLineXSize = 0;
		mLineMaxAscenderSize = 0;
		mLineMaxDescenderSize = 0;
	}



	private void RestartLine()
	in
	{
		assert(mLineStarted);
	}
	body
	{
		EndLine();
		mLineStarted = true;
		mLineXSize = 0;
		auto LastIntermediate = mIntermediates[$ - 1];
		mIntermediates ~= new CIntermediate(mLines.length, 0, LastIntermediate.Write,
				mWordStartCharacterIndex);
		if (mElementProcessMode == EElementProcessMode.Word)
		{
			mX = mWordXSize;
			mLineMaxAscenderSize = mWordMaxAscenderSize;
			mLineMaxDescenderSize = mWordMaxDescenderSize;
		}
		else
		{
			mX = 0;
			mLineMaxAscenderSize = 0;
			mLineMaxDescenderSize = 0;
		}
	}



	private void EndLine()
	in
	{
		assert(mLineStarted);
	}
	body
	{
		mLineStarted = false;
		mLines ~= new CLineInfo(mLineXSize, mLineMaxAscenderSize, mLineMaxDescenderSize);
	}



	// simple test with word-wrap disabled
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, -1, 0, 5, 0, Writes);
		assert(Fragments.length == 1);
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 30);
		assert(Fragments[0].YSize == 10);
	}



	// simple test with word-wrap enabled, but no breaks
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 35, 0, 5, 0, Writes);
		assert(Fragments.length == 1);
		assert(Fragments[0].Text == "foo");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 30);
		assert(Fragments[0].YSize == 10);
	}



	// test word-wrap with a break during the space
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo bar"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 35, 0, 5, 0, Writes);
		assert(Fragments.length == 2);
		assert(Fragments[0].Text == "foo");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 30);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "bar");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 30);
		assert(Fragments[1].YSize == 10);
	}



	// test word-wrap with a break during the word after the space
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo bar"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 45, 0, 5, 0, Writes);
		assert(Fragments.length == 2);
		assert(Fragments[0].Text == "foo");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 30);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "bar");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 30);
		assert(Fragments[1].YSize == 10);
	}



	/**
	Test for bug where previous spaces increase the skip length and also test that the first
	letter in a word contributes toward the line length when the break occurs after the
	first character in a word.
	*/
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "word wrap test"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 115, 0, 5, 0, Writes);
		assert(Fragments.length == 2);
		assert(Fragments[0].Text == "word wrap");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 90);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "test");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 40);
		assert(Fragments[1].YSize == 10);
	}



	// test two writes on separate lines
	unittest
	{
		CWriteInfo[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo\n"),
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "bar"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, -1, 0, 5, 0, Writes);
		assert(Fragments.length == 2);
		assert(Fragments[0].Text == "foo");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 30);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "bar");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 30);
		assert(Fragments[1].YSize == 10);
	}



	// test an image
	unittest
	{
		ITexture Texture = new CTestTexture();
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, Texture)
		];
		new CGetMetrics(Writes, -1, new CTestFontProvider(), LineXSizes, LineYSizes,
				FirstBaselineYPos, WrapBreaks);
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, -1, 0, 5, 0, Writes);
		assert(Fragments.length == 1);
		assert(Fragments[0].Texture == Texture);
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 64);
		assert(Fragments[0].YSize == 66);
	}



	// test a single word too long to fit on a line is broken up by characters
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "foo"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 25, 0, 5, 0, Writes);
		assert(Fragments.length == 2);
		assert(Fragments[0].Text == "fo");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 20);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "o");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 10);
		assert(Fragments[1].YSize == 10);
	}



	/**
	Test a single word too long to fit on a line (following another word) is broken up by
	characters.
	*/
	unittest
	{
		CWrite[] Writes = [
			new CWrite(new CTestFont(), 1, new CColour(0xffffff), 1, "a foo"),
		];
		CTextFragment[] Fragments = new CProcessor().Process(0, 0, 25, 0, 5, 0, Writes);
		assert(Fragments.length == 3);
		assert(Fragments[0].Text == "a");
		assert(Fragments[0].X == 0);
		assert(Fragments[0].Y == 0);
		assert(Fragments[0].XSize == 10);
		assert(Fragments[0].YSize == 10);
		assert(Fragments[1].Text == "fo");
		assert(Fragments[1].X == 0);
		assert(Fragments[1].Y == 15);
		assert(Fragments[1].XSize == 20);
		assert(Fragments[1].YSize == 10);
		assert(Fragments[2].Text == "o");
		assert(Fragments[2].X == 0);
		assert(Fragments[2].Y == 30);
		assert(Fragments[2].XSize == 10);
		assert(Fragments[2].YSize == 10);
	}
}



version(unittest)
{
	class CTestFont : IFont
	{
		public FT_Face Face()
		{
			throw new Exception();
		}
	}



	class CTestTexture : ITexture
	{
		public uint ID()
		{
			return 1;
		}



		public int XSize()
		{
			return 64;
		}



		public int YSize()
		{
			return 64;
		}



		public float TextureXSize()
		{
			return 64;
		}



		public float TextureYSize()
		{
			return 64;
		}
	}
}