﻿module QBasic.Interpreter.CEvaluator;
import std.algorithm, std.array, std.conv, std.math, std.random;
import QBasic.EKey, QBasic.Program.CType, QBasic.Program.CValue, QBasic.Program.EBuiltinFunction,
		QBasic.Program.EOperator, QBasic.System.IFileSystem;

interface IEvaluatorState
{
	void SetError(string aMessage);
	IFileSystem FileSystem();
	bool GetKey(out EKey aKey);
	immutable(ubyte[]) CommandLine();
}

class CEvaluator
{
	private IEvaluatorState mEvaluatorState;

	public this(IEvaluatorState aEvaluatorState)
	{
		mEvaluatorState = aEvaluatorState;
	}



	public CValue BinaryOperator(CValue aOperand0, CValue aOperand1, EOperator aOperator,
			EPrimitiveType aResultType)
	out(aResult)
	{
		assert(aResult is null || aResult.Type == aResultType);
	}
	body
	{
		final switch (aOperator)
		{
			case EOperator.Exponent:
				// TODO: implement this
				break;
			case EOperator.Multiply:
				return BinaryOperatorMultiply(aOperand0, aOperand1, aResultType);
			case EOperator.Divide:
			case EOperator.DivideInt:
				return BinaryOperatorDivide(aOperand0, aOperand1, aResultType);
			case EOperator.Modulus:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integral(aOperand0.Type,
								aOperand0.IntegralValue % aOperand1.IntegralValue);
					default:
						break;
				}
				break;
			case EOperator.Add:
				return BinaryOperatorAdd(aOperand0, aOperand1, aResultType);
			case EOperator.Subtract:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integral(aOperand0.Type,
								aOperand0.IntegralValue - aOperand1.IntegralValue);
					case EPrimitiveType.Single:
						return CValue.Single(aOperand0.SingleValue -
								aOperand1.SingleValue);
					default:
						break;
				}
				break;
			case EOperator.Concatenate:
				assert(aOperand0.Type == EPrimitiveType.String &&
						aOperand1.Type == EPrimitiveType.String);
				return CValue.String(aOperand0.StringValue ~ aOperand1.StringValue);
			case EOperator.BitwiseAnd:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
						return CValue.Integer(cast(short)aOperand0.IntegralValue &
								cast(short)aOperand1.IntegralValue);
					case EPrimitiveType.Long:
						return CValue.Long(aOperand0.IntegralValue & aOperand1.IntegralValue);
					default:
						break;
				}
				break;
			case EOperator.BitwiseOr:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
						return CValue.Integer(cast(short)aOperand0.IntegralValue |
								cast(short)aOperand1.IntegralValue);
					case EPrimitiveType.Long:
						return CValue.Long(aOperand0.IntegralValue | aOperand1.IntegralValue);
					default:
						break;
				}
				break;
			case EOperator.BitwiseXor:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
						return CValue.Integer(cast(short)aOperand0.IntegralValue ^
								cast(short)aOperand1.IntegralValue);
					case EPrimitiveType.Long:
						return CValue.Long(aOperand0.IntegralValue ^ aOperand1.IntegralValue);
					default:
						break;
				}
				break;
			case EOperator.CompareEqual:
				assert(aResultType == EPrimitiveType.Integer);
				return BinaryOperatorCompareEqual(aOperand0, aOperand1);
			case EOperator.CompareNotEqual:
				assert(aResultType == EPrimitiveType.Integer);
				return BinaryOperatorCompareNotEqual(aOperand0, aOperand1);
			case EOperator.CompareLess:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integer(aOperand0.IntegralValue <
								aOperand1.IntegralValue ? -1 : 0);
					case EPrimitiveType.Single:
						return CValue.Integer(aOperand0.SingleValue <
								aOperand1.SingleValue ? -1 : 0);
					default:
						break;
				}
				break;
			case EOperator.CompareLessOrEqual:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integer(aOperand0.IntegralValue <=
								aOperand1.IntegralValue ? -1 : 0);
					case EPrimitiveType.Single:
						return CValue.Integer(aOperand0.SingleValue <=
								aOperand1.SingleValue ? -1 : 0);
					default:
						break;
				}
				break;
			case EOperator.CompareGreater:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integer(aOperand0.IntegralValue >
								aOperand1.IntegralValue ? -1 : 0);
					case EPrimitiveType.Single:
						return CValue.Integer(aOperand0.SingleValue >
								aOperand1.SingleValue ? -1 : 0);
					default:
						break;
				}
				break;
			case EOperator.CompareGreaterOrEqual:
				switch (aOperand0.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integer(aOperand0.IntegralValue >=
								aOperand1.IntegralValue ? -1 : 0);
					case EPrimitiveType.Single:
						return CValue.Integer(aOperand0.SingleValue >=
								aOperand1.SingleValue ? -1 : 0);
					default:
						break;
				}
				break;
		}
		mEvaluatorState.SetError("Binary operator failed.");
		return null;
	}



	private CValue CheckInteger(int aValue)
	{
		if (aValue < -32768 || aValue > 32767)
		{
			mEvaluatorState.SetError("Integer value out of range.");
			return null;
		}
		return CValue.Integer(aValue);
	}



	private CValue CheckInteger(long aValue)
	{
		if (aValue < -32768 || aValue > 32767)
		{
			mEvaluatorState.SetError("Integer value out of range.");
			return null;
		}
		return CValue.Integer(cast(int)aValue);
	}



	private CValue CheckLong(long aValue)
	{
		if (aValue < -2147483648 || aValue > 2147483647)
		{
			mEvaluatorState.SetError("Long value out of range.");
			return null;
		}
		return CValue.Long(cast(int)aValue);
	}



	private CValue ConvertToInteger(CValue aValue)
	{
		switch (aValue.Type)
		{
			case EPrimitiveType.Integer:
				return aValue;
			case EPrimitiveType.Long:
				return CheckInteger(aValue.IntegralValue);
			case EPrimitiveType.Single:
				return CheckInteger(cast(long)round(aValue.SingleValue));
			default:
				mEvaluatorState.SetError("Unable to convert value to integer.");
				return null;
		}
	}



	private CValue ConvertToLong(CValue aValue)
	{
		switch (aValue.Type)
		{
			case EPrimitiveType.Integer:
				return CValue.Long(aValue.IntegralValue);
			case EPrimitiveType.Long:
				return aValue;
			case EPrimitiveType.Single:
				return CheckLong(cast(long)round(aValue.SingleValue));
			default:
				mEvaluatorState.SetError("Unable to convert value to long.");
				return null;
		}
	}



	private CValue ConvertToSingle(CValue aValue)
	{
		switch (aValue.Type)
		{
			case EPrimitiveType.Integer:
			case EPrimitiveType.Long:
				return CValue.Single(aValue.IntegralValue);
			case EPrimitiveType.Single:
				return aValue;
			default:
				mEvaluatorState.SetError("Unable to convert value to single.");
				return null;
		}
	}



	public CValue ConvertTo(CValue aValue, EPrimitiveType aTargetType)
	{
		switch (aTargetType)
		{
			case EPrimitiveType.Integer:
				return ConvertToInteger(aValue);
			case EPrimitiveType.Long:
				return ConvertToLong(aValue);
			case EPrimitiveType.Single:
				return ConvertToSingle(aValue);
			case EPrimitiveType.String:
				assert(aTargetType == EPrimitiveType.String);
				return aValue;
			default:
				mEvaluatorState.SetError("Unable to convert value to specified type.");
				return null;
		}
	}



	private CValue BinaryOperatorMultiply(CValue aOperand0, CValue aOperand1,
			EPrimitiveType aResultType)
	{
		CValue Operand0 = ConvertTo(aOperand0, aResultType);
		CValue Operand1 = ConvertTo(aOperand1, aResultType);
		if (Operand0 is null || Operand1 is null)
			return null;
		switch (aResultType)
		{
			case EPrimitiveType.Integer:
				return CheckInteger(Operand0.IntegralValue * Operand1.IntegralValue);
			case EPrimitiveType.Long:
				return CheckLong(cast(long)Operand0.IntegralValue * cast(long)Operand1.IntegralValue);
			case EPrimitiveType.Single:
				return CValue.Single(Operand0.SingleValue * Operand1.SingleValue);
			default:
				mEvaluatorState.SetError("Unable to convert multiply result to specified type.");
				return null;
		}
	}



	private CValue BinaryOperatorDivide(CValue aOperand0, CValue aOperand1,
			EPrimitiveType aResultType)
	{
		CValue Operand0 = ConvertTo(aOperand0, aResultType);
		CValue Operand1 = ConvertTo(aOperand1, aResultType);
		if (Operand0 is null || Operand1 is null)
			return null;
		switch (aResultType)
		{
			case EPrimitiveType.Integer:
				if (Operand1.IntegralValue == 0)
				{
					mEvaluatorState.SetError("Division by 0.");
					return null;
				}
				return CValue.Integer(Operand0.IntegralValue / Operand1.IntegralValue);
			case EPrimitiveType.Long:
				if (Operand1.IntegralValue == 0)
				{
					mEvaluatorState.SetError("Division by 0.");
					return null;
				}
				return CValue.Long(Operand0.IntegralValue / Operand1.IntegralValue);
			case EPrimitiveType.Single:
				if (Operand1.SingleValue == 0)
				{
					mEvaluatorState.SetError("Division by 0.");
					return null;
				}
				return CValue.Single(Operand0.SingleValue / Operand1.SingleValue);
			default:
				mEvaluatorState.SetError("Unable to convert divide result to specified type.");
				return null;
		}
	}



	private CValue BinaryOperatorAdd(CValue aOperand0, CValue aOperand1,
			EPrimitiveType aResultType)
	{
		CValue Operand0 = ConvertTo(aOperand0, aResultType);
		CValue Operand1 = ConvertTo(aOperand1, aResultType);
		if (Operand0 is null || Operand1 is null)
			return null;
		switch (aResultType)
		{
			case EPrimitiveType.Integer:
				return CheckInteger(Operand0.IntegralValue + Operand1.IntegralValue);
			case EPrimitiveType.Long:
				return CheckLong(cast(long)Operand0.IntegralValue + cast(long)Operand1.IntegralValue);
			case EPrimitiveType.Single:
				return CValue.Single(Operand0.SingleValue + Operand1.SingleValue);
			default:
				mEvaluatorState.SetError("Unable to convert add result to specified type.");
				return null;
		}
	}



	private CValue Bool(bool aValue)
	{
		return CValue.Integer(aValue ? -1 : 0);
	}



	private CValue BinaryOperatorCompareEqual(CValue aOperand0, CValue aOperand1)
	{
		if (aOperand0.Type == EPrimitiveType.String)
		{
			if (aOperand1.Type == EPrimitiveType.String)
				return Bool(equal(aOperand0.StringValue, aOperand1.StringValue));
		}
		else if (aOperand0.Type == EPrimitiveType.Single || aOperand1.Type == EPrimitiveType.Single)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Single);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Single);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.SingleValue == Operand1.SingleValue);
		}
		else if (aOperand0.Type == EPrimitiveType.Long || aOperand1.Type == EPrimitiveType.Long)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Long);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Long);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.IntegralValue == Operand1.IntegralValue);
		}
		else if (aOperand0.Type == EPrimitiveType.Integer || aOperand1.Type == EPrimitiveType.Integer)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Integer);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Integer);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.IntegralValue == Operand1.IntegralValue);
		}
		mEvaluatorState.SetError("Unable to compare the specified operand type.");
		return null;
	}



	private CValue BinaryOperatorCompareNotEqual(CValue aOperand0, CValue aOperand1)
	{
		if (aOperand0.Type == EPrimitiveType.String)
		{
			if (aOperand1.Type == EPrimitiveType.String)
				return Bool(!equal(aOperand0.StringValue, aOperand1.StringValue));
		}
		else if (aOperand0.Type == EPrimitiveType.Single || aOperand1.Type == EPrimitiveType.Single)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Single);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Single);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.SingleValue != Operand1.SingleValue);
		}
		else if (aOperand0.Type == EPrimitiveType.Long || aOperand1.Type == EPrimitiveType.Long)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Long);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Long);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.IntegralValue != Operand1.IntegralValue);
		}
		else if (aOperand0.Type == EPrimitiveType.Integer || aOperand1.Type == EPrimitiveType.Integer)
		{
			CValue Operand0 = ConvertTo(aOperand0, EPrimitiveType.Integer);
			CValue Operand1 = ConvertTo(aOperand1, EPrimitiveType.Integer);
			if (Operand0 is null || Operand1 is null)
				return null;
			return Bool(Operand0.IntegralValue != Operand1.IntegralValue);
		}
		mEvaluatorState.SetError("Unable to compare (<>) the specified operand type.");
		return null;
	}



	public CValue UnaryOperator(CValue aOperand, EUnaryOperator aOperator)
	{
		final switch (aOperator)
		{
			case EUnaryOperator.Negate:
				switch (aOperand.Type)
				{
					case EPrimitiveType.Integer:
					case EPrimitiveType.Long:
						return CValue.Integral(aOperand.Type,
								-aOperand.IntegralValue);
					case EPrimitiveType.Single:
						return CValue.Single(-aOperand.SingleValue);
					default:
						break;
				}
				break;
			case EUnaryOperator.BitwiseNot:
				switch (aOperand.Type)
				{
					case EPrimitiveType.Integer:
						return CValue.Integer(~cast(short)aOperand.IntegralValue);
					case EPrimitiveType.Long:
						return CValue.Long(~aOperand.IntegralValue);
					default:
						break;
				}
				break;
		}
		mEvaluatorState.SetError("Unary operator failed.");
		return null;
	}



	/**
	@returns True iff the function has executed completely. False if the function has
	partially executed.
	*/
	public bool BuiltinFunction(EBuiltinFunction aFunction, CValue[] aArguments,
			out CValue aResult)
	{
		final switch (aFunction)
		{
			case EBuiltinFunction.Asc:
				return FunctionAsc(aArguments[0], aResult);
			case EBuiltinFunction.Chr:
				aResult = FunctionChr(aArguments[0]);
				return true;
			case EBuiltinFunction.Command:
				return FunctionCommand(aResult);
			case EBuiltinFunction.EOF:
				return FunctionEOF(aArguments[0], aResult);
			case EBuiltinFunction.FreeFile:
				return FunctionFreeFile(aResult);
			case EBuiltinFunction.InKey:
				return FunctionInKey(aResult);
			case EBuiltinFunction.Int:
				aResult = FunctionInt(aArguments[0]);
				return true;
			case EBuiltinFunction.Len:
				return FunctionLen(aArguments[0], aResult);
			case EBuiltinFunction.LOF:
				return FunctionLOF(aArguments[0], aResult);
			case EBuiltinFunction.LTrim:
				aResult = FunctionLTrim(aArguments[0]);
				return true;
			case EBuiltinFunction.Mid:
				return FunctionMid(aArguments[0], aArguments[1], aArguments[2], aResult);
			case EBuiltinFunction.Rnd:
				aResult = FunctionRnd();
				return true;
			case EBuiltinFunction.RTrim:
				return FunctionRTrim(aArguments[0], aResult);
			case EBuiltinFunction.Sgn:
				return FunctionSgn(aArguments[0], aResult);
			case EBuiltinFunction.Space:
				return FunctionSpace(aArguments[0], aResult);
			case EBuiltinFunction.Str:
				aResult = FunctionStr(aArguments[0]);
				return true;
			case EBuiltinFunction.Val:
				return FunctionVal(aArguments[0], aResult);
			case EBuiltinFunction.VarSeg:
				return FunctionVarSeg(aArguments[0], aResult);
		}
	}



	private bool FunctionAsc(CValue aChar, out CValue aResult)
	{
		if (aChar.Type != EPrimitiveType.String || aChar.StringValue.length != 1)
		{
			mEvaluatorState.SetError("Expected string of length 1.");
			return false;
		}
		aResult = CValue.Integer(aChar.StringValue[0]);
		return true;
	}



	private CValue FunctionChr(CValue aCode)
	{
		int Code = aCode.IntegralValue;
		if (!(Code >= 0 && Code < 256))
			throw new Exception("Illegal function call");
		return CValue.String([ cast(ubyte)Code ]);
	}



	private bool FunctionCommand(out CValue aResult)
	{
		aResult = CValue.String(mEvaluatorState.CommandLine);
		return true;
	}



	private bool FunctionLen(CValue aString, out CValue aResult)
	{
		aResult = CValue.Integer(aString.StringValue.length);
		return true;
	}



	private bool FunctionLOF(CValue aFileNumber, out CValue aResult)
	{
		int LOF;
		if (!mEvaluatorState.FileSystem.LOF(aFileNumber.IntegralValue, LOF))
			return false;
		aResult = CValue.Integer(LOF);
		return true;
	}



	private bool FunctionEOF(CValue aFileNumber, out CValue aResult)
	{
		bool IsEOF;
		if (!mEvaluatorState.FileSystem.EOF(aFileNumber.IntegralValue, IsEOF))
			return false;
		aResult = CValue.Integer(IsEOF ? -1 : 0);
		return true;
	}



	private bool FunctionFreeFile(out CValue aResult)
	{
		aResult = CValue.Integer(mEvaluatorState.FileSystem.FreeFile());
		return true;
	}



	private bool FunctionInKey(out CValue aKey)
	{
		EKey Key;
		if (!mEvaluatorState.GetKey(Key))
			return false;
		final switch (Key)
		{
			case EKey.Left:
				aKey = CValue.String([ cast(ubyte)0, cast(ubyte)75 ]);
				break;
			case EKey.Right:
				aKey = CValue.String([ cast(ubyte)0, cast(ubyte)77 ]);
				break;
			case EKey.Up:
				aKey = CValue.String([ cast(ubyte)0, cast(ubyte)72 ]);
				break;
			case EKey.Down:
				aKey = CValue.String([ cast(ubyte)0, cast(ubyte)80 ]);
				break;
			case EKey.Enter:
				aKey = CValue.String([ cast(ubyte)13 ]);
				break;
			case EKey.Delete:
				aKey = CValue.String([ cast(ubyte)0, cast(ubyte)83 ]);
				break;
			case EKey.Escape:
				aKey = CValue.String([ cast(ubyte)27 ]);
				break;
		}
		return true;
	}



	private CValue FunctionInt(CValue aNumber)
	{
		if (canFind([ EPrimitiveType.Integer, EPrimitiveType.Long ], aNumber.Type))
			return aNumber;
		// TODO: return type?
		return CValue.Integer(cast(int)floor(aNumber.SingleValue));
	}



	private CValue FunctionLTrim(CValue aString)
	{
		auto Result = aString.StringValue;
		int Start = 0;
		foreach (int i, ubyte c; Result)
		{
			if (c != 32)
			{
				Start = i;
				break;
			}
		}
		return CValue.String(Result[Start .. $]);
	}



	private bool FunctionMid(CValue aString, CValue aStart, CValue aLength, out CValue aResult)
	{
		int Start = aStart.IntegralValue - 1;
		if (Start < 0 || Start + aLength.IntegralValue > aString.StringValue.length)
			return false;
		aResult = CValue.String(aString.StringValue[Start .. Start + aLength.IntegralValue]);
		return true;
	}



	private CValue FunctionRnd()
	{
		return CValue.Single(uniform(0.0f, 1.0f));
	}



	private bool FunctionRTrim(CValue aString, out CValue aResult)
	{
		auto Data = aString.StringValue;
		int End = Data.length;
		foreach_reverse (int i, ubyte c; Data)
		{
			if (c != 32)
				break;
			End--;
		}
		aResult = CValue.String(Data[0 .. End]);
		return true;
	}



	private bool FunctionSgn(CValue aValue, out CValue aResult)
	{
		int Value = aValue.IntegralValue;
		if (Value == 0)
			aResult = CValue.Integer(0);
		else
			aResult = CValue.Integer(Value > 0 ? 1 : -1);
		return true;
	}



	private bool FunctionSpace(CValue aLength, out CValue aResult)
	{
		if (aLength.IntegralValue < 0)
		{
			mEvaluatorState.SetError("Expected non-negative integer.");
			return false;
		}
		aResult = CValue.String(replicate([ cast(ubyte)32 ], aLength.IntegralValue));
		return true;
	}



	private CValue FunctionStr(CValue aNumber)
	{
		// TODO: implement this manually, instead of using to()
		string ResultStr = to!string(aNumber.IntegralValue);
		if (aNumber.IntegralValue >= 0)
			ResultStr = " " ~ ResultStr;
		ubyte[] Result = new ubyte[ResultStr.length];
		foreach (int i, char c; ResultStr)
			Result[i] = cast(ubyte)c;
		return CValue.String(Result.idup);
	}



	private bool FunctionVal(CValue aString, out CValue aResult)
	{
		int Result = 0;
		foreach (ubyte b; aString.StringValue)
		{
			if (b < 48 || b > 57)
			{
				mEvaluatorState.SetError("Invalid numeric character.");
				return false;
			}
			Result *= 10;
			Result += b - 48;
		}
		aResult = CValue.Integer(Result);
		return true;
	}



	private bool FunctionVarSeg(CValue aReference, out CValue aResult)
	{
		assert(aReference.Type == EPrimitiveType.Reference);
		aResult = aReference;
		return true;
	}
}