﻿module QBasic.Interpreter.CExpressionInterpreter;
import QBasic.Interpreter.CReference, QBasic.Interpreter.IInterpreterState,
		QBasic.Program.CType, QBasic.Program.CValue, QBasic.Program.EOperator,
		QBasic.Program.Expressions.CExpression, QBasic.Program.Expressions.Units;

class CExpressionInterpreter
{
	private class CStack
	{
		public this(CExpression aExpression)
		{
			Expression = aExpression;
			UnitIndex = Expression.Units.length - 1;
		}

		public CExpression Expression;
		public int UnitIndex;
		public bool FunctionGetResult = false;
		public CValue[] Values;
		public CValue[] Operands;
	}

	private CStack[] mStack;
	private CValue mResult;

	public this(CExpression aExpression)
	in
	{
		assert(aExpression !is null);
	}
	body
	{
		mStack = [ new CStack(aExpression) ];
	}



	public CValue GetResult()
	in
	{
		assert(mStack.length == 1 && mResult !is null);
	}
	body
	{
		return mResult;
	}



	public bool Execute(IInterpreterState aState)
	out (result)
	{
		assert(!result || mResult !is null);
	}
	body
	{
		while (true)
		{
			auto Stack = mStack[$ - 1];
			assert(Stack.Expression.Units.length != 0);
			while (Stack.UnitIndex >= 0)
			{
				CUnit Unit = Stack.Expression.Units[Stack.UnitIndex];
				final switch (Unit.UnitType)
				{
					case EUnitType.BinaryOperator:
						CValue Operand0 = Stack.Operands[$ - 1];
						CValue Operand1 = Stack.Operands[$ - 2];
						Stack.Operands.length = Stack.Operands.length - 2;
						auto BinaryOperator = Unit.AsBinaryOperator();
						auto Result = aState.GetEvaluator().BinaryOperator(Operand0,
								Operand1, BinaryOperator.Operator, BinaryOperator.ResultType);
						if (Result is null)
							return false;
						Stack.Operands ~= Result;
						break;
					case EUnitType.UnaryOperator:
						CValue Operand = Stack.Operands[$ - 1];
						Stack.Operands.length = Stack.Operands.length - 1;
						Stack.Operands ~= [ aState.GetEvaluator().UnaryOperator(Operand,
								Unit.AsUnaryOperator().Operator) ];
						break;
					case EUnitType.BuiltinFunction:
					case EUnitType.UserFunction:
						auto Function = Unit.AsFunction();
						bool Block;
						if (ExecuteFunctionUnit(aState, Stack, Function, Block))
						{
							Stack = mStack[$ - 1];
							continue;
						}
						if (Block)
							return false;
						break;
					case EUnitType.Value:
						Stack.Operands ~= [ Unit.AsValue().Value ];
						break;
					case EUnitType.Variable:
						auto Variable = Unit.AsVariable();
						if (ExecuteVariableNode(aState, Stack, Variable))
						{
							Stack = mStack[$ - 1];
							continue;
						}
						break;
				}
				Stack.UnitIndex--;
			}
			assert(Stack.Operands.length == 1);
			CValue Result = Stack.Operands[0];
			if (mStack.length == 1)
			{
				if (Result.Type == EPrimitiveType.ArrayReference || Result.Type ==
						EPrimitiveType.Reference)
					mResult = Result;
				else
					mResult = aState.GetEvaluator().ConvertTo(Result, Stack.Expression.ResultType);
				return mResult !is null;
			}
			mStack.length = mStack.length - 1;
			mStack[$ - 1].Values ~= [ Result ];
		}
		assert(0);
	}



	private bool ExecuteFunctionUnit(IInterpreterState aState, CStack aStack,
			CFunctionUnit aUnit, out bool aBlock)
	{
		if (aStack.FunctionGetResult)
		{
			aStack.Operands ~= [ aState.GetFunctionResult() ];
			aStack.FunctionGetResult = false;
			return false;
		}
		if (aStack.Values.length != aUnit.Arguments.length)
		{
			mStack ~= new CStack(aUnit.Arguments[aStack.Values.length]);
			return true;
		}
		if (aUnit.UnitType == EUnitType.BuiltinFunction)
		{
			auto BuiltinFunction = aUnit.AsBuiltinFunction();
			CValue Result;
			if (!aState.GetEvaluator.BuiltinFunction(BuiltinFunction.Function, aStack.Values,
					Result))
			{
				aBlock = true;
				return false;
			}
			aStack.Operands ~= Result;
			aStack.Values.length = 0;
			return false;
		}
		else
		{
			auto UserFunction = aUnit.AsUserFunction();
			aState.CallProcedure(UserFunction.Name, aStack.Values);
			aStack.FunctionGetResult = true;
			aStack.Values.length = 0;
			aBlock = true;
			return false;
		}
	}



	private bool ExecuteVariableNode(IInterpreterState aState, CStack aStack,
			CVariableUnit aUnit)
	{
		if (aStack.Values.length != aUnit.Subscripts.length)
		{
			mStack ~= new CStack(aUnit.Subscripts[aStack.Values.length]);
			return true;
		}
		int[] Subscripts = new int[aStack.Values.length];
		foreach (int i, CValue Value; aStack.Values)
			Subscripts[i] = Value.IntegralValue;
		if (aUnit.ArrayReference)
			aStack.Operands ~= CValue.ArrayReference(aState.TakeArrayReference(aUnit.Name,
					aUnit.Type));
		else if (aUnit.Reference)
		{
			auto Reference = aState.TakeReference(aUnit.Name, Subscripts.idup,
					aUnit.RecordOffset, aUnit.TargetType);
			auto Target = aState.ReadTarget(Reference);
			if (Target !is null && Target.Type == EPrimitiveType.Reference)
			{
				assert(Subscripts.length == 0 || Target.Reference.Subscripts.length == 0);
				auto NewSubscripts = Subscripts.length == 0 ? Target.Reference.Subscripts :
						Subscripts;
				Reference = new CReference(Target.Reference.Name, Target.Reference.StackFrame,
						NewSubscripts.idup, Target.Reference.RecordOffset + aUnit.RecordOffset,
						aUnit.Type);
			}
			aStack.Operands ~= CValue.Reference(Reference);
		}
		else
			aStack.Operands ~= aState.ReadVariable(aState.TakeReference(aUnit.Name,
					Subscripts.idup, aUnit.RecordOffset, aUnit.Type));
		aStack.Values.length = 0;
		return false;
	}
}