﻿module QBasic.Interpreter.CInterpreter;
import std.algorithm, std.string, std.typecons;
import QBasic.CCP437Encoder, QBasic.EKey, QBasic.IFileSystemExternal, QBasic.IGraphicsExternal,
		QBasic.IInputExternal, QBasic.Interpreter.CEvaluator,
		QBasic.Interpreter.CExpressionInterpreter, QBasic.Interpreter.CReference,
		QBasic.Interpreter.CStackFrame, QBasic.Interpreter.CVariable,
		QBasic.Interpreter.CVariableCollection, QBasic.Interpreter.Instructions.IInstructionInterpreter,
		QBasic.Interpreter.IInterpreterState, QBasic.Program.CArgument,
		QBasic.Program.CIdentifier, QBasic.Program.CName, QBasic.Program.CProcedure,
		QBasic.Program.CProgram, QBasic.Program.CType, QBasic.Program.CValue,
		QBasic.Program.EExtensionCommand, QBasic.Program.Expressions.CExpression,
		QBasic.Program.Instructions.IInstruction, QBasic.System.CFileSystem,
		QBasic.System.CGraphics, QBasic.System.IFileSystem, QBasic.System.IGraphics;

interface IInterpreterLog
{
	void TraceLog(string aLog);
}

class CInterpreter : IEvaluatorState, IInterpreterState
{
	private CEvaluator mEvaluator;
	private CFileSystem mFileSystem;
	private CGraphics mGraphics;
	private IInputExternal mInputExternal;
	private CProgram mProgram;
	private string mCommandLine;
	private IInterpreterLog mLog;
	private CStackFrame[] mStack;
	private int mNextInstruction = -1;
	private CValue mExpressionResult;
	private CVariableCollection mSharedVariables;
	private CIdentifier mCallProcedureName = null;
	private CValue[] mCallArguments;
	private bool mWaitingForKey = false;
	private Nullable!EKey mKey;
	private bool mError = false;
	private string mErrorMessage;
	private CReference mDefSegReference = null;
	private int[] mBreakpointLines;
	private CCP437Encoder mEncoder;

	public this(IFileSystemExternal aFileSystemExternal, IGraphicsExternal aGraphicsExternal,
			IInputExternal aInputExternal, CProgram aProgram, IInterpreterLog aLog,
			string aCommandLine)
	{
		mEvaluator = new CEvaluator(this);
		mFileSystem = new CFileSystem(aFileSystemExternal, this);
		mGraphics = new CGraphics(aGraphicsExternal, this);
		mInputExternal = aInputExternal;
		mProgram = aProgram;
		mLog = aLog;
		mCommandLine = aCommandLine;
		mSharedVariables = new CVariableCollection(mProgram.SharedVariables);
		mStack ~= new CStackFrame(mProgram.MainProcedure);
		mEncoder = new CCP437Encoder();
	}



	public bool Error()
	{
		return mError;
	}



	public string ErrorMessage()
	{
		return mErrorMessage;
	}



	public int ErrorLine()
	{
		return mStack[$ - 1].Procedure.Instructions[mStack[$ - 1].InstructionIndex].Line;
	}



	public int ErrorPos()
	{
		return mStack[$ - 1].Procedure.Instructions[mStack[$ - 1].InstructionIndex].Pos;
	}



	public CEvaluator GetEvaluator()
	{
		return mEvaluator;
	}



	public const(ubyte[]) ScreenData()
	{
		return mGraphics.ScreenData;
	}



	public int ScreenXSize()
	{
		return mGraphics.ScreenXSize;
	}



	public int ScreenYSize()
	{
		return mGraphics.ScreenYSize;
	}



	public void AddBreakpoint(int aLine)
	{
		mBreakpointLines ~= aLine;
	}



	private void BreakpointHit()
	{
		auto Frame = mStack[$ - 1];
		auto Instruction = Frame.Procedure.Instructions[Frame.InstructionIndex];
		mLog.TraceLog(format("Breakpoint at line %d", Instruction.Line + 1));
		mLog.TraceLog("Variables:");
		foreach (CIdentifier Identifier, const(CVariable) Variable; Frame.Variables.Variables)
		{
			mLog.TraceLog(format("%s: %s", Identifier.Identifier, ShowValue(Variable.Values[0])));
		}
	}



	private string ShowValue(const(CValue) aValue)
	{
		final switch (aValue.Type)
		{
			case EPrimitiveType.ArrayReference:
				return "arrayref to " ~ aValue.Reference.Name.Identifier.Identifier;
			case EPrimitiveType.Integer:
				return format("%d(%%)", aValue.IntegralValue);
			case EPrimitiveType.Long:
				return format("%d(&)", aValue.IntegralValue);
			case EPrimitiveType.Reference:
				return "ref to " ~ aValue.Reference.Name.Identifier.Identifier;
			case EPrimitiveType.Single:
				return format("%d(!)", aValue.SingleValue);
			case EPrimitiveType.String:
				return format("\"%s\"", mEncoder.Decode(aValue.StringValue));
		}
	}



	/**
	Partially executes the program.
	@return true if the program has finished executing, or false if it has made a blocking
	call.
	*/
	public bool Step()
	{
		while (true)
		{
			auto Frame = mStack[$ - 1];
			if (Frame.ExpressionInterpreter !is null)
			{
				bool Blocking;
				if (ExecuteExpression(Frame.ExpressionInterpreter, Blocking))
					Frame.ExpressionInterpreter = null;
				else
				{
					if (Blocking)   // blocking function call
						return false;
					if (mError)
						return true;
					Frame = mStack[$ - 1];
				}
			}
			auto Instructions = Frame.Procedure.Instructions;
			while (Frame.InstructionInterpreter !is null || Frame.InstructionIndex <
					Instructions.length)
			{
				if (Frame.InstructionInterpreter is null)
				{
					if (mBreakpointLines.canFind(Instructions[Frame.InstructionIndex].Line))
					{
						BreakpointHit();
						return false;
					}
					Frame.InstructionInterpreter =
							Instructions[Frame.InstructionIndex].CreateInterpreter();
				}
				bool Blocking;
				if (Frame.InstructionInterpreter.Execute(this, Blocking))
				{
					Frame.InstructionInterpreter = null;
					if (mNextInstruction == -1)
						Frame.InstructionIndex++;
					else
					{
						Frame.InstructionIndex = mNextInstruction;
						mNextInstruction = -1;
					}
					Frame.InstructionStage = 0;
					if (mCallProcedureName !is null)
					{
						auto Procedure = mProgram.Procedures[mCallProcedureName];
						auto StackFrame = new CStackFrame(Procedure);
						mStack ~= StackFrame;
						foreach (int i, CIdentifier Identifier; Procedure.ArgumentVariables)
							StackFrame.Variables.Set(Identifier, null, 0, mCallArguments[i]);
						mCallProcedureName = null;
						Frame = mStack[$ - 1];
						Instructions = Frame.Procedure.Instructions;
					}
				}
				else
				{
					if (Blocking)
						return false;
					if (mError)
						return true;
					Frame = mStack[$ - 1];
					Instructions = Frame.Procedure.Instructions;
				}
			}
			if (mStack.length == 1)
				return true;
			if (mStack[$ - 1].Procedure.Info.ReturnType !is null && mStack[$ - 2].ReturnValue is null)
			{
				CValue DefaultReturnValue;
				switch (mStack[$ - 1].Procedure.Info.ReturnType.PrimitiveType)
				{
					case EPrimitiveType.Integer:
						DefaultReturnValue = CValue.Integer(0);
						break;
					case EPrimitiveType.Long:
						DefaultReturnValue = CValue.Long(0);
						break;
					case EPrimitiveType.Single:
						DefaultReturnValue = CValue.Single(0);
						break;
					case EPrimitiveType.String:
						DefaultReturnValue = CValue.String([]);
						break;
					default:
						assert(false);
				}
				mStack[$ - 2].ReturnValue = DefaultReturnValue;
			}
			mStack.length = mStack.length - 1;
		}
	}



	/**
	@return true if the expression completed executing.
	*/
	private bool ExecuteExpression(CExpressionInterpreter aExpressionInterpreter,
			out bool aBlocking)
	{
		aBlocking = false;
		if (aExpressionInterpreter.Execute(this))
		{
			mExpressionResult = aExpressionInterpreter.GetResult();
			mStack[$ - 1].InstructionStage++;
			return true;
		}
		else
		{
			if (mCallProcedureName !is null)
			{
				CProcedure Procedure = mProgram.Procedures[mCallProcedureName];
				auto StackFrame = new CStackFrame(Procedure);
				mStack ~= StackFrame;
				foreach (int i, CArgument Argument; Procedure.Info.Arguments)
					StackFrame.Variables.Set(Argument.Identifier, null, 0, mCallArguments[i]);
				mCallProcedureName = null;
				return false;
			}
			else if (mError)
			{
				return false;
			}
			else   // blocking call
			{
				aBlocking = true;
				return false;
			}
		}
	}



	public IInstruction GetInstruction()
	{
		return mStack[$ - 1].Procedure.Instructions[mStack[$ - 1].InstructionIndex];
	}



	public int GetInstructionStage()
	{
		return mStack[$ - 1].InstructionStage;
	}



	public void NextInstruction(int aNextInstruction)
	in
	{
		assert(aNextInstruction >= 0 && aNextInstruction <= mStack[$ - 1].Procedure.
				Instructions.length);
	}
	body
	{
		mNextInstruction = aNextInstruction;
	}



	/**
	Partially (or fully) evaluates the expression.
	@param aBlocking Receives whether or not the expression made a blocking call if it
	was only partially executed.
	@return true if the expression was fully evaluated, or false if it was only
	partially evaluated.
	*/
	public bool Expression(CExpression aExpression, out bool aBlocking)
	in
	{
		assert(mStack[$ - 1].ExpressionInterpreter is null);
	}
	/*
	out (result)
	{
		bool n = mStack[$ - 1].ExpressionInterpreter is null;
		return result == n;
	}
	*/
	body
	{
		auto Frame = mStack[$ - 1];
		auto ExpressionInterpreter = new CExpressionInterpreter(aExpression);
		if (ExecuteExpression(ExpressionInterpreter, aBlocking))
			return true;
		Frame.ExpressionInterpreter = ExpressionInterpreter;
		return false;
	}



	/**
	Get the result of the last evaluated expression.
	*/
	public CValue GetExpressionResult()
	in
	{
		assert(mExpressionResult !is null);
	}
	body
	{
		return mExpressionResult;
	}



	/**
	Sets the specified procedure to be called when the currently executing operation returns.
	*/
	public void CallProcedure(CIdentifier aName, CValue[] aArguments)
	{
		mCallProcedureName = aName;
		mCallArguments = aArguments;
	}



	public CValue GetFunctionResult()
	out (result)
	{
		assert(result !is null);
	}
	body
	{
		return mStack[$ - 1].ReturnValue;
	}



	public void ExtensionCommand(EExtensionCommand aCommand, CValue[] aArguments)
	{
		final switch (aCommand)
		{
			case EExtensionCommand.KeyRate:
				mInputExternal.KeyRate(aArguments[0].IntegralValue);
				break;
			case EExtensionCommand.KeyStop:
				mInputExternal.KeyStop();
				break;
		}
	}



	public CValue ReadTarget(CReference aReference)
	{
		if (!aReference.Name.Shared && aReference.Name.Identifier == mStack[$ - 1].Procedure.Name)
		{
			assert(aReference.Subscripts.length == 0 && aReference.RecordOffset == 0);
			return null;
		}
		return *GetVariableReference(aReference.Name, aReference.StackFrame,
				aReference.Subscripts, aReference.RecordOffset);
	}



	public CValue ReadVariable(CReference aReference)
	in
	{
		assert(aReference.Type.Type == EType.Primitive);
	}
	body
	{
		CValue* Value = GetVariableReference(aReference.Name, aReference.StackFrame,
				aReference.Subscripts, aReference.RecordOffset);
		// if the value is a reference, read the referred variable
		if (Value.Type == EPrimitiveType.ArrayReference)
		{
			Value = GetVariableReference(Value.Reference.Name, Value.Reference.StackFrame,
					aReference.Subscripts, aReference.RecordOffset);
			assert(!canFind([ EPrimitiveType.ArrayReference, EPrimitiveType.Reference ],
					Value.Type));
		}
		else if (Value.Type == EPrimitiveType.Reference)
		{
			assert(Value.Reference.Subscripts.length == 0 || aReference.Subscripts.length == 0);
			auto Subscripts = Value.Reference.Subscripts.length == 0 ? aReference.Subscripts :
					Value.Reference.Subscripts;
			Value = GetVariableReference(Value.Reference.Name, Value.Reference.StackFrame,
					Subscripts, Value.Reference.RecordOffset + aReference.RecordOffset);
			assert(!canFind([ EPrimitiveType.ArrayReference, EPrimitiveType.Reference ],
					Value.Type));
		}
		return *Value;
	}



	public CReference TakeArrayReference(CName aName, CType aType)
	{
		int StackFrame = aName.Shared ? -1 : mStack.length - 1;
		return new CReference(aName, StackFrame, null, 0, aType);
	}



	public CReference TakeReference(CName aName, immutable(int[]) aSubscripts,
			int aRecordOffset, CType aType)
	{
		int StackFrame = aName.Shared ? -1 : mStack.length - 1;
		return new CReference(aName, StackFrame, aSubscripts, aRecordOffset, aType);
	}



	public void WriteVariable(CReference aTarget, CValue aValue)
	{
		// assign to the return value of a function
		if (!aTarget.Name.Shared && aTarget.Name.Identifier == mStack[$ - 1].Procedure.Name)
		{
			assert(aTarget.Subscripts.length == 0 && aTarget.RecordOffset == 0);
			mStack[$ - 2].ReturnValue = aValue;
			return;
		}
		CValue* Value = GetVariableReference(aTarget.Name, aTarget.StackFrame,
				aTarget.Subscripts, aTarget.RecordOffset);
		// if the value is a reference, write to the referred variable
		if (*Value !is null)
		{
			if (Value.Type == EPrimitiveType.ArrayReference)
				Value = GetVariableReference(Value.Reference.Name, Value.Reference.StackFrame,
						aTarget.Subscripts, aTarget.RecordOffset);
			else if (Value.Type == EPrimitiveType.Reference)
			{
				assert(Value.Reference.Subscripts.length == 0 || aTarget.Subscripts.length == 0);
				auto Subscripts = aTarget.Subscripts.length != 0 ? aTarget.Subscripts :
						Value.Reference.Subscripts;
				Value = GetVariableReference(Value.Reference.Name, Value.Reference.StackFrame,
						Subscripts, Value.Reference.RecordOffset + aTarget.RecordOffset);
			}
			assert(!canFind([ EPrimitiveType.ArrayReference, EPrimitiveType.Reference ],
					Value.Type));
		}
		assert(Value.Type == aValue.Type);
		*Value = aValue;
	}



	public void ResizeArray(CName aName, const(int[]) aNewDimensions)
	{
		if (aName.Shared)
			mSharedVariables.ResizeArray(aName.Identifier, aNewDimensions);
		else
			mStack[$ - 1].Variables.ResizeArray(aName.Identifier, aNewDimensions);
	}



	private CValue* GetVariableReference(CName aName, int aStackFrame,
			immutable(int[]) aSubscripts, int aRecordOffset)
	in
	{
		assert((aName.Shared && aStackFrame == -1) || (!aName.Shared && aStackFrame >= 0));
	}
	body
	{
		if (aName.Shared)
			return mSharedVariables.GetReference(aName.Identifier, aSubscripts, aRecordOffset);
		return mStack[aStackFrame].Variables.GetReference(aName.Identifier, aSubscripts,
				aRecordOffset);
	}



	public CVariableCollection SharedVariables()
	{
		return mSharedVariables;
	}



	public void SetStack(int aOffset, CValue aValue)
	{
		mStack[$ - 1].Stack[aOffset] = aValue;
	}



	public CValue GetStack(int aOffset)
	{
		return mStack[$ - 1].Stack[aOffset];
	}



	public bool GetKey(out EKey aKey)
	{
		if (!mKey.isNull)
		{
			aKey = mKey;
			mKey = Nullable!EKey();
			return true;
		}
		mWaitingForKey = true;
		return false;
	}



	public bool IsWaitingForKey()
	{
		return mWaitingForKey;
	}



	public void SetKey(EKey aKey)
	{
		mWaitingForKey = false;
		mKey = aKey;
	}



	public IFileSystem FileSystem()
	{
		return mFileSystem;
	}



	public IGraphics Graphics()
	{
		return mGraphics;
	}



	public void SetError(string aMessage)
	{
		mError = true;
		mErrorMessage = aMessage;
	}



	public void SetDefSeg(CReference aReference)
	{
		mDefSegReference = aReference;
	}



	public bool WriteData(const(ubyte[]) aData)
	{
		if (mDefSegReference is null)
		{
			SetError("Attempting to write but Def Seg is not set.");
			return false;
		}
		return WriteData(mDefSegReference, aData);
	}



	public bool WriteData(CReference aReference, const(ubyte[]) aData)
	{
		auto Variables = aReference.Name.Shared ? mSharedVariables :
				mStack[aReference.StackFrame].Variables;
		if (!Variables.WriteData(aReference.Name.Identifier, aReference.Subscripts,
				aReference.RecordOffset, aData))
		{
			SetError("Writing binary data to variable failed (buffer overflow).");
			return false;
		}
		return true;
	}



	public immutable(ubyte[]) ReadData(CReference aReference, int aByteSize)
	{
		auto Variables = aReference.Name.Shared ? mSharedVariables :
				mStack[aReference.StackFrame].Variables;
		if (Variables.IsArrayReference(aReference.Name.Identifier))
		{
			auto Value = Variables.GetReference(aReference.Name.Identifier, [], 0);
			aReference = new CReference(Value.Reference.Name, Value.Reference.StackFrame,
					aReference.Subscripts, aReference.RecordOffset, aReference.Type);
			Variables = aReference.Name.Shared ? mSharedVariables :
				mStack[aReference.StackFrame].Variables;
		}
		auto Data = Variables.ReadData(aReference.Name.Identifier, aReference.Subscripts,
				aReference.RecordOffset, aByteSize);
		if (Data is null)
			SetError("Reading binary data from variable failed (buffer overflow).");
		return Data;
	}



	public int ValueSize(CReference aTarget)
	in
	{
		assert(aTarget.Type.Type == EType.Primitive || aTarget.Type.Type == EType.Record);   // TODO: support arrays?
	}
	body
	{
		if (aTarget.Type.Type == EType.Primitive && aTarget.Type.PrimitiveType ==
				EPrimitiveType.String)
		{
			auto Value = ReadVariable(aTarget);
			return Value.StringValue.length;
		}
		return aTarget.Type.ByteSize();
	}



	public immutable(ubyte[]) CommandLine()
	{
		return mEncoder.Encode(mCommandLine);
	}
}