﻿module QBasic.Interpreter.CVariable;
import std.algorithm.iteration;
import QBasic.Program.CType, QBasic.Program.CValue, QBasic.Program.CVariableInfo;

class CVariable
{
	private CVariableInfo mInfo;
	private int[] mDimensions;
	private CValue[] mValues;

	public this(CVariableInfo aInfo)
	{
		mInfo = aInfo;
		if (mInfo.Type.Type == EType.Array)
		{
			mDimensions = mInfo.InitialDimensions;
			mValues = DefaultArrayValues();
		}
		else if (mInfo.Type.Type == EType.ArrayReference)
			mValues = [ cast(CValue)null ];
		else
			mValues = DefaultValuesForType(mInfo.Type);
	}



	private CValue[] DefaultValuesForType(CType aType)
	{
		final switch (aType.Type)
		{
			case EType.Array:
			case EType.ArrayReference:
				assert(false);
			case EType.Primitive:
				return [ DefaultValueForPrimitiveType(aType.PrimitiveType) ];
			case EType.Record:
				return DefaultRecordValues(aType);
		}
	}



	private CValue[] DefaultArrayValues()
	{
		int Size = 1;
		foreach (int Dimension; mInfo.InitialDimensions)
			Size *= Dimension;
		int TypeSize = TypeSize(mInfo.Type.SubType);
		auto Values = new CValue[Size * TypeSize];
		for (int i = 0; i < Values.length; i += TypeSize)
			Values[i .. i + TypeSize] = DefaultValuesForType(mInfo.Type.SubType)[];
		return Values;
	}



	private CValue[] DefaultRecordValues(CType aType)
	{
		auto Values = new CValue[aType.RecordSize];
		for (int i = 0; i < aType.RecordSize; i++)
		{
			auto FieldType = aType.GetRecordFieldType(i);
			Values[i] = DefaultValueForPrimitiveType(FieldType.PrimitiveType);
		}
		return Values;
	}



	private CValue DefaultValueForPrimitiveType(EPrimitiveType aType)
	{
		final switch (aType)
		{
			case EPrimitiveType.ArrayReference:
				assert(false);
			case EPrimitiveType.Integer:
			case EPrimitiveType.Long:
				return CValue.Integral(aType, 0);
			case EPrimitiveType.Reference:
				assert(false);
			case EPrimitiveType.Single:
				return CValue.Single(0f);
			case EPrimitiveType.String:
				return CValue.String([]);
		}
	}



	private int TypeSize(CType aType)
	{
		final switch (aType.Type)
		{
			case EType.Array:
			case EType.ArrayReference:
				assert(false);
			case EType.Primitive:
				return 1;
			case EType.Record:
				return aType.RecordSize;
		}
	}



	public const(CValue[]) Values() const
	{
		return mValues;
	}



	public void Set(immutable(int[]) aSubscripts, int aRecordOffset, CValue aValue)
	in
	{
		assert(mInfo.Type.Type != EType.ArrayReference || (aSubscripts is null &&
				aRecordOffset == 0 && aValue.Type == EPrimitiveType.ArrayReference));
	}
	body
	{
		int Offset = GetOffset(aSubscripts, aRecordOffset);
		assert((mValues[Offset] is null && (aValue.Type == EPrimitiveType.ArrayReference ||
				aValue.Type == EPrimitiveType.Reference)) ||
				mValues[Offset].Type == aValue.Type ||
				aValue.Type == EPrimitiveType.Reference);
		mValues[Offset] = aValue;
	}



	public CValue* GetReference(immutable(int[]) aSubscripts, int aRecordOffset)
	out(rResult)
	{
		assert(*rResult !is null);
	}
	body
	{
		if (mInfo.Type.Type == EType.ArrayReference || mValues[0].Type == EPrimitiveType.Reference)
			return &mValues[0];
		return &mValues[GetOffset(aSubscripts, aRecordOffset)];
	}



	private int GetOffset(const(int[]) aSubscripts, int aRecordOffset)
	{
		final switch (mInfo.Type.Type)
		{
			case EType.ArrayReference:
			case EType.Primitive:
				assert(aSubscripts.length == 0 && aRecordOffset == 0);
				return 0;
			case EType.Array:
				assert(aSubscripts.length == mDimensions.length);
				int Offset = 0;
				foreach (int i, int Subscript; aSubscripts)
				{
					assert(Subscript >= 0 && Subscript < mDimensions[i]);
					if (i == 0)
						Offset += Subscript;
					else
						Offset += Subscript * mDimensions[i - 1];
				}
				return Offset * TypeSize(mInfo.Type.SubType) + aRecordOffset;
			case EType.Record:
				assert(aSubscripts.length == 0 && aRecordOffset >= 0 && aRecordOffset <
						mInfo.Type.RecordSize);
				return aRecordOffset;
		}
	}



	public void ResizeArray(const(int[]) aNewDimensions)
	in
	{
		assert(aNewDimensions.length != 0 && aNewDimensions.length == mDimensions.length);
	}
	body
	{
		int TypeSize = TypeSize(mInfo.Type.SubType);
		auto NewValues = new CValue[reduce!((a, b) => a * b)(aNewDimensions) * TypeSize];
		auto Iterator = new int[mDimensions.length];
		for (int i = 0; i < NewValues.length; i += TypeSize)
		{
			bool KeepOld = true;
			for (int j = 0; j < mDimensions.length; j++)
			{
				if (Iterator[j] >= mDimensions[j])
				{
					KeepOld = false;
					break;
				}
			}
			if (KeepOld)
			{
				int Offset = GetOffset(Iterator, 0);
				NewValues[i .. i + TypeSize] = mValues[Offset .. Offset + TypeSize];
			}
			else
				NewValues[i .. i + TypeSize] = DefaultValuesForType(mInfo.Type.SubType)[];
			for (int j = 0; j < mDimensions.length; j++)
			{
				if (Iterator[j] == aNewDimensions[j] - 1)
					Iterator[j] = 0;
				else
				{
					Iterator[j]++;
					break;
				}
			}
		}
		mDimensions[] = aNewDimensions[];
		mValues = NewValues;
	}



	public bool WriteData(const(int[]) aSubscripts, int aRecordOffset, const(ubyte[]) aData)
	{
		int Offset = GetOffset(aSubscripts, aRecordOffset);
		if (mValues[Offset].Type == EPrimitiveType.String)
		{
			assert(mValues[Offset].StringValue.length == aData.length);
			mValues[Offset] = CValue.String(aData.idup);
			return true;
		}
		for (int DataIndex = 0; DataIndex < aData.length; Offset++)
		{
			if (Offset == mValues.length)
				return false;
			WriteDataPrimitive(Offset, aData, DataIndex);
		}
		return true;
	}



	private void WriteDataPrimitive(int aOffset, const(ubyte[]) aData, ref int aDataIndex)
	{
		int TypeSize = PrimitiveTypeByteSize(mValues[aOffset].Type);
		assert(aDataIndex + TypeSize <= aData.length);   // TODO: might not be true
		uint Value = 0;
		for (int i = 0; i < TypeSize; i++)
			Value |= aData[aDataIndex + i] << 8 * i;
		aDataIndex += TypeSize;
		switch (mValues[aOffset].Type)
		{
			case EPrimitiveType.Integer:
				mValues[aOffset] = CValue.Integral(EPrimitiveType.Integer, *cast(short*)&Value);
				break;
			case EPrimitiveType.Long:
				mValues[aOffset] = CValue.Integral(EPrimitiveType.Long, *cast(int*)&Value);
				break;
			case EPrimitiveType.Single:
				mValues[aOffset] = CValue.Single(*cast(float*)&Value);
				break;
			default:
				assert(false);
		}
	}



	public bool IsArrayReference()
	{
		return mInfo.Type.Type == EType.ArrayReference;
	}



	public immutable(ubyte[]) ReadData(const(int[]) aSubscripts, int aRecordOffset,
			int aByteSize)
	{
		int Offset = GetOffset(aSubscripts, aRecordOffset);
		if (Offset >= mValues.length)
			return null;
		if (mValues[Offset].Type == EPrimitiveType.String)
			return mValues[Offset].StringValue;
		ubyte[] Data = new ubyte[aByteSize];
		for (int DataIndex; DataIndex < Data.length; Offset++)
		{
			if (Offset == mValues.length)
				return null;
			ReadDataPrimitive(Offset, Data, DataIndex);
		}
		return Data.idup;
	}



	public void ReadDataPrimitive(int aOffset, ubyte[] aData, ref int aDataIndex)
	{
		int TypeSize = PrimitiveTypeByteSize(mValues[aOffset].Type);
		assert(aDataIndex + TypeSize <= aData.length);   // TODO: might not be true
		uint Value = 0;
		switch (mValues[aOffset].Type)
		{
			case EPrimitiveType.Integer:
			case EPrimitiveType.Long:
				int i = mValues[aOffset].IntegralValue;
				Value = *cast(uint*)&i;
				break;
			case EPrimitiveType.Single:
				float f = mValues[aOffset].SingleValue;
				Value = *cast(uint*)&f;
				break;
			default:
				assert(false);
		}
		for (int i = 0; i < TypeSize; i++)
			aData[aDataIndex + i] = (Value >> 8 * i) & 0xff;
		aDataIndex += TypeSize;
	}
}