﻿module QBasic.Parser.CExpressionParser;
import std.algorithm, std.array, std.typecons;
import QBasic.Interpreter.CEvaluator, QBasic.Parser.Expressions.CExpressionBuilder,
		QBasic.Program.CIdentifier, QBasic.Program.CName, QBasic.Program.CType,
		QBasic.Program.CValue, QBasic.Program.EBuiltinFunction, QBasic.Program.EOperator,
		QBasic.Program.Expressions.CExpression;

class CExpressionParser
{
	private CEvaluator mEvaluator;
	private Nullable!(EPrimitiveType) mResultType;
	private CExpressionBuilder[] mStates;

	public this(CEvaluator aEvaluator, Nullable!(EPrimitiveType) aResultType)
	{
		mEvaluator = aEvaluator;
		mResultType = aResultType;
		mStates = [ new CExpressionBuilder(mEvaluator, mResultType) ];
	}



	public bool IsValid()
	{
		return mStates.length == 1 && mStates[0].IsValid();
	}



	public CExpression GetExpression(bool aReference)
	in
	{
		assert(IsValid());
	}
	body
	{
		return mStates[0].GetExpression(aReference);
	}



	public bool Function(CIdentifier aName, CType aType, EPrimitiveType[] aArgumentTypes)
	{
		return mStates[$ - 1].UserFunction(aName, aType, aArgumentTypes);
	}



	public bool Symbol(string aSymbol)
	{
		switch (aSymbol)
		{
			case "(":
				return OpenBracket();
			case ")":
				return CloseBracket();
			case "^":
				return mStates[$ - 1].BinaryOperator(EOperator.Exponent);
			case "*":
				return mStates[$ - 1].BinaryOperator(EOperator.Multiply);
			case "/":
				return mStates[$ - 1].BinaryOperator(EOperator.Divide);
			case "\\":
				return mStates[$ - 1].BinaryOperator(EOperator.DivideInt);
			case "+":
				return mStates[$ - 1].Plus();
			case "-":
				return mStates[$ - 1].Minus();
			case "=":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareEqual);
			case "<>":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareNotEqual);
			case "<":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareLess);
			case "<=":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareLessOrEqual);
			case ">":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareGreater);
			case ">=":
				return mStates[$ - 1].BinaryOperator(EOperator.CompareGreaterOrEqual);
			case ",":
				return Comma();
			case "NOT":
				return mStates[$ - 1].UnaryOperator(EUnaryOperator.BitwiseNot);
			case "AND":
				return mStates[$ - 1].BinaryOperator(EOperator.BitwiseAnd);
			case "OR":
				return mStates[$ - 1].BinaryOperator(EOperator.BitwiseOr);
			case "XOR":
				return mStates[$ - 1].BinaryOperator(EOperator.BitwiseXor);
			case "MOD":
				return mStates[$ - 1].BinaryOperator(EOperator.Modulus);
			case "ASC":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Asc, [ EPrimitiveType.String ]);
			case "CHR$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Chr, [ EPrimitiveType.Integer ]);
			case "COMMAND$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Command, []);
			case "EOF":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.EOF, [ EPrimitiveType.Integer ]);
			case "FREEFILE":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.FreeFile, []);
			case "INKEY$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.InKey, []);
			case "INT":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Int, [ EPrimitiveType.Single ]);
			case "LEN":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Len, [ EPrimitiveType.String ]);
			case "LOF":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.LOF, [ EPrimitiveType.Integer ]);
			case "LTRIM$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.LTrim, [ EPrimitiveType.String ]);
			case "MID$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Mid, [ EPrimitiveType.String, EPrimitiveType.Integer, EPrimitiveType.Integer ]);
			case "RND":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Rnd, []);
			case "RTRIM$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.RTrim, [ EPrimitiveType.String ]);
			case "SGN":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Sgn, [ EPrimitiveType.Integer ]);   // TODO: allow any numeric type
			case "SPACE$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Space, [ EPrimitiveType.Integer ]);
			case "STR$":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Str, [ EPrimitiveType.Integer ]);
			case "TIMER":
				return mStates[$ - 1].Value(CValue.Single(0.0f));
			case "VAL":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.Val, [ EPrimitiveType.String ]);
			case "VARSEG":
				return mStates[$ - 1].BuiltinFunction(EBuiltinFunction.VarSeg, [ EPrimitiveType.Reference ], true);
			default:
				return false;
		}
	}



	private bool OpenBracket()
	{
		Nullable!(EPrimitiveType) NewExpressionType;
		if (!mStates[$ - 1].OpenBracket(NewExpressionType))
			return false;
		if (!NewExpressionType.isNull)
			mStates ~= [ new CExpressionBuilder(mEvaluator, NewExpressionType) ];
		return true;
	}



	private bool CloseBracket()
	{
		if (!mStates[$ - 1].CloseBracket())
		{
			if (mStates.length == 1 || !(mStates[$ - 1].IsEmpty() || mStates[$ - 1].IsValid()))
				return false;
			if (mStates[$ - 1].IsEmpty())
				mStates.length = mStates.length - 1;
			else
			{
				auto ExpressionState = mStates[$ - 1];
				auto PreviousState = mStates[$ - 2];
				CExpression Expression = ExpressionState.GetExpression(PreviousState.TakeArgumentReference());
				if (Expression is null)
					return false;
				mStates.length = mStates.length - 1;
				Nullable!(EPrimitiveType) NextExpressionType;
				if (!PreviousState.Expression(Expression, NextExpressionType))
					return false;
				assert(NextExpressionType.isNull);
			}
			return mStates[$ - 1].CloseBracket();
		}
		return true;
	}



	private bool Comma()
	{
		if (mStates.length == 1 || !mStates[$ - 1].IsValid())
			return false;
		CExpression Expression = mStates[$ - 1].GetExpression(false);
		if (Expression is null)
			return false;
		mStates.length = mStates.length - 1;
		Nullable!(EPrimitiveType) NextExpressionType;
		if (!mStates[$ - 1].Expression(Expression, NextExpressionType))
			return false;
		mStates ~= new CExpressionBuilder(mEvaluator, NextExpressionType);
		return true;
	}



	public bool Value(CValue aValue)
	{
		return mStates[$ - 1].Value(aValue);
	}



	public bool IsVariableValid()
	{
		return mStates[$ - 1].IsVariableValid();
	}



	public bool Variable(CName aName, CType aType)
	in
	{
		assert(aName !is null && aType !is null);
	}
	body
	{
		return mStates[$ - 1].Variable(aName, aType);
	}



	public void RecordOffset(int aOffset, CType aFieldType)
	in
	{
		assert(aOffset >= 0);
	}
	body
	{
		mStates[$ - 1].RecordOffset(aOffset, aFieldType);
	}



	public CType GetPreviousVariableType()
	{
		return mStates[$ - 1].GetPreviousVariableType();
	}
}