﻿module QBasic.Parser.CParser;
import std.algorithm, std.array, std.container, std.stream, std.string, std.typecons;
import QBasic.CCP437Encoder, QBasic.EKey, QBasic.Interpreter.CEvaluator,
		QBasic.Parser.CExpressionParser, QBasic.Parser.CToken, QBasic.Parser.CTokeniser,
		QBasic.Parser.Expressions.CExpressionBuilder, QBasic.Parser.Expressions.CVariableExpressionNode,
		QBasic.Program.CArgument, QBasic.Program.CIdentifier, QBasic.Program.CName,
		QBasic.Program.CProcedureInfo, QBasic.Program.CProgram, QBasic.Program.CType,
		QBasic.Program.CValue, QBasic.Program.CVariableInfo, QBasic.Program.EExtensionCommand,
		QBasic.Program.EFileOpenMode, QBasic.Program.EGraphicsOperation,
		QBasic.Program.ELineType, QBasic.Program.Expressions.CExpression,
		QBasic.Program.Instructions.CAssignmentInstruction, QBasic.Program.Instructions.CBLoadInstruction,
		QBasic.Program.Instructions.CBranchInstruction, QBasic.Program.Instructions.CDefSegInstruction,
		QBasic.Program.Instructions.CExtensionCommandInstruction, QBasic.Program.Instructions.CFileCloseInstruction,
		QBasic.Program.Instructions.CFileGetInstruction, QBasic.Program.Instructions.CFileLineInputInstruction,
		QBasic.Program.Instructions.CFileOpenInstruction, QBasic.Program.Instructions.CFilePutInstruction,
		QBasic.Program.Instructions.CFileSeekInstruction, QBasic.Program.Instructions.CForInitInstruction,
		QBasic.Program.Instructions.CForInstruction, QBasic.Program.Instructions.CForNextInstruction,
		QBasic.Program.Instructions.CGraphicsLineInstruction, QBasic.Program.Instructions.CGraphicsPutInstruction,
		QBasic.Program.Instructions.CIfInstruction, QBasic.Program.Instructions.CInstruction,
		QBasic.Program.Instructions.CLoopStartInstruction, QBasic.Program.Instructions.CMidInstruction,
		QBasic.Program.Instructions.COutInstruction, QBasic.Program.Instructions.CRedimInstruction,
		QBasic.Program.Instructions.CSelectCase, QBasic.Program.Instructions.CSelectCaseInstruction,
		QBasic.Program.Instructions.CSelectInstruction, QBasic.Program.Instructions.CSubCallInstruction,
		QBasic.Program.Instructions.IBranchInstruction, QBasic.Program.Instructions.IInstruction,
		QBasic.System.IFileSystem;

enum EParseError { ExpectedStatement, ExpectedEndStatement, ExpectedKeyword, ExpectedType,
		ExpectedIdentifier, ExpectedSymbol, ExpectedExpression, ExpectedConstant,
		DuplicateDefinition, InvalidExpression, IdentifierWithPeriod, InvalidNumber,
		NotFirstStatement, ExpectedRelationalOperator, UndeclaredProcedure, ArgumentsMismatch,
		InvalidDeclaration, InvalidSubName, UndeclaredVariable, ExpressionNotConstant,
		InvalidBlockEnd, InvalidBlockExit, InvalidFieldName, Unsupported,
		ExpectedNumericVariable, TypeMismatch, InvalidAssignmentTarget };

enum ENameType { Const, Procedure, Variable };

interface IParserCallback
{
	void ReadToken(CToken aToken);
}

class CParser : IEvaluatorState
{
	private struct STypeFieldInfo
	{
		this(string aName, CType aType, int aOffset)
		{
			Name = aName;
			Type = aType;
			Offset = aOffset;
		}

		string Name;
		CType Type;
		int Offset;
	}

	private struct STypeInfo
	{
		this(STypeFieldInfo[] aFields, int aSize)
		{
			TypeList.length = aFields.length;
			foreach (int i, STypeFieldInfo Field; aFields)
			{
				TypeList[i] = Field.Type;
				Types[Field.Name] = Field.Type;
				Offsets[Field.Name] = Field.Offset;
			}
			Size = aSize;
		}

		CType[] TypeList;
		CType[string] Types;
		int[string] Offsets;
		int Size;
	}

	private enum EContextType { Else, ElseIf, For, If, Loop, Select, SelectCase,
			SelectCaseElse };
	private struct SContext
	{
		this(EContextType aType, int aInstructionIndex)
		{
			Type = aType;
			InstructionIndex = aInstructionIndex;
			SelectExpressionType = Nullable!(EPrimitiveType)();
		}

		EContextType Type;
		int InstructionIndex;
		int[] ExitInstructions;
		private Nullable!(EPrimitiveType) SelectExpressionType;

		EPrimitiveType GetSelectExpressionType()
		{
			assert(!SelectExpressionType.isNull);
			return SelectExpressionType;
		}

		void SetSelectExpressionType(EPrimitiveType aSelectExpressionType)
		{
			SelectExpressionType = aSelectExpressionType;
		}
	}

	private CCP437Encoder mCP437Encoder;
	private CTokeniser mTokeniser;
	private CEvaluator mEvaluator;
	private bool mSuccess;
	private EParseError mError;
	private CToken mErrorToken;
	private IParserCallback mParserCallback = null;

	private CToken mToken;
	private bool mFirstStatement;
	private bool mStatementStarted = false;
	private int mInstructionLine;
	private int mInstructionPos;

	private bool mLineIfStatement;
	private bool mLineIfFirst;

	private CProgram mProgram;
	private STypeInfo[string] mTypes;
	private CValue[CIdentifier] mConstants;
	private CProcedureInfo[CIdentifier] mProcedures;
	private CVariableInfo[CIdentifier] mSharedVariables;
	private bool[string] mSharedVariablePrefixes;

	private CIdentifier mCurrentProcedure = null;
	private CVariableInfo[CIdentifier] mVariables;
	private CIdentifier[] mArgumentVariables;
	private bool[string] mVariablePrefixes;
	private IInstruction[] mInstructions;
	private int mStackSize = 0;
	private int mCurrentStackSize = 0;
	private SContext[] mContext;
	private int[] mExitProcedureInstructions;

	public this(CCP437Encoder aCP437Encoder, Stream aStream)
	{
		mCP437Encoder = aCP437Encoder;
		mTokeniser = new CTokeniser(aStream);
		mEvaluator = new CEvaluator(this);
	}



	public void SetParserCallback(IParserCallback aParserCallback)
	{
		mParserCallback = aParserCallback;
	}



	public bool Success()
	{
		return mSuccess;
	}



	public CProgram GetProgram()
	in
	{
		assert(mSuccess);
	}
	body
	{
		return mProgram;
	}



	public EParseError GetError()
	in
	{
		assert(!mSuccess);
	}
	body
	{
		return mError;
	}



	public int GetErrorLine()
	in
	{
		assert(!mSuccess);
	}
	body
	{
		return mErrorToken.Line;
	}



	public int GetErrorPos()
	in
	{
		assert(!mSuccess);
	}
	body
	{
		return mErrorToken.Pos;
	}



	public void Parse()
	{
		mFirstStatement = true;
		mLineIfStatement = false;
		mProgram = new CProgram();

		ReadToken();
		mSuccess = Parse1();
		assert(mSuccess || mErrorToken !is null);
	}



	private bool Parse1()
	{
		while (true)
		{
			if (mToken is null)
			{
				if (mCurrentProcedure is null)
				{
					mProgram.AddProcedure(null, null, mVariables, null, mInstructions, mStackSize);
					mProgram.SetSharedVariables(mSharedVariables);
				}
				return true;
			}
			if (mToken.Type == ETokenType.Newline)   // ignore blank lines
			{
				Consume();
				mFirstStatement = true;
				continue;
			}
			if (!ReadStatement())
				return false;
			if (!mLineIfStatement)
			{
				if (IsNewline())
					mFirstStatement = true;
				else if (IsSymbol(":"))
					mFirstStatement = false;
				else
				{
					ParseError(EParseError.ExpectedEndStatement);
					return false;
				}
				Consume();
			}
			else
			{
				if (IsSymbol(":"))
					Consume();
				else if (IsKeyword("ELSE"))
					ReadElseStatement();
				else if (mLineIfFirst)
					mLineIfFirst = false;
				else if (IsNewline())
				{
					Consume();
					ReadEndIfStatement();
					mFirstStatement = true;
					mLineIfStatement = false;
				}
				else
				{
					ParseError(EParseError.ExpectedEndStatement);
					return false;
				}
			}
		}
	}



	private bool ReadStatement()
	{
		switch (mToken.Type)
		{
			case ETokenType.Keyword:
				switch (mToken.Text)
				{
					case "BLOAD":
						if (!ReadBLoadStatement())
							return false;
						break;
					case "CONST":
						Consume();
						if (!ReadConstStatement())
							return false;
						break;
					case "DECLARE":
						Consume();
						if (!ReadDeclareStatement())
							return false;
						break;
					case "DEF":
						if (!ReadDefStatement())
							return false;
						break;
					case "DEFINT":
						Consume();
						while (true)
						{
							if (IsEndStatement())
								break;
							Consume();
						}
						break;
					case "DIM":
						Consume();
						if (!ReadDimStatement())
							return false;
						break;
					case "REDIM":
						if (!ReadRedimStatement())
							return false;
						break;
					case "DO":
						if (!ReadDoStatement())
							return false;
						break;
					case "LOOP":
						if (!ReadLoopStatement())
							return false;
						break;
					case "END":
						if (!ReadEndStatement())
							return false;
						break;
					case "EXIT":
						if (!ReadExitStatement())
							return false;
						break;
					case "FOR":
						if (!ReadForStatement())
							return false;
						break;
					case "NEXT":
						if (!NextStatement())
							return false;
						break;
					case "FUNCTION":
					case "SUB":
						if (mCurrentProcedure is null)
						{
							mProgram.AddProcedure(null, null, mVariables, null, mInstructions,
									mStackSize);
							mProgram.SetSharedVariables(mSharedVariables);
						}
						bool Sub = mToken.Text == "SUB";
						Consume();
						if (!ReadProcedureStatement(Sub))
							return false;
						break;
					case "GET":
						if (!ReadGetFileStatement())
							return false;
						break;
					case "IF":
						if (!ReadIfStatement())
							return false;
						break;
					case "ELSEIF":
						if (!ReadElseIfStatement())
							return false;
						break;
					case "ELSE":
						if (!ReadElseStatement())
							return false;
						break;
					case "LINE":
						StartStatement();
						Consume();
						if (IsKeyword("INPUT"))
						{
							Consume();
							if (!ReadLineInputStatement())
								return false;
						}
						else
						{
							if (!ReadLineGraphicsStatement())
								return false;
						}
						break;
					case "MID$":
						if (!ReadMidStatement())
							return false;
						break;
					case "OPEN":
						if (!ReadOpenStatement())
							return false;
						break;
					case "CLOSE":
						if (!ReadCloseStatement())
							return false;
						break;
					case "OUT":
						if (!ReadOutStatement())
							return false;
						break;
					case "PUT":
						StartStatement();
						Consume();
						if (IsSymbol("("))
						{
							if (!ReadPutGraphicsStatement())
								return false;
						}
						else
						{
							if (!ReadPutFileStatement())
								return false;
						}
						break;
					case "RANDOMIZE":
					case "SCREEN":
					case "WIDTH":
						// these commands are ignored
						Consume();
						CExpression Expression;
						if (!ParseExpression(Expression, EPrimitiveType.Integer))
							return false;
						break;
					case "REM":
						mTokeniser.SkipLine();
						Consume();
						break;
					case "SEEK":
						if (!ReadSeekStatement())
							return false;
						break;
					case "SELECT":
						if (!ReadSelectStatement())
							return false;
						break;
					case "CASE":
						if (!ReadCaseStatement())
							return false;
						break;
					case "TYPE":
						Consume();
						if (!ReadTypeBlock())
							return false;
						break;
					case "QB.KEYRATE":
					case "QB.KEYSTOP":
						if (!ReadExtensionCommandStatement())
							return false;
						break;
					default:
						ParseError(EParseError.ExpectedStatement);
						return false;
				}
				break;
			case ETokenType.Identifier:
				auto Identifier = new CIdentifier(mToken.Text);
				if (IsSub(Identifier))
				{
					if (!ReadSubCallStatement(Identifier))
						return false;
				}
				else if (!ReadAssignmentStatement())
					return false;
				break;
			default:
				ParseError(EParseError.ExpectedStatement);
				return false;
		}
		return true;
	}



	private bool ReadBLoadStatement()
	{
		StartStatement();
		Consume();
		CExpression FileNameExpression;
		if (!ParseExpression(FileNameExpression, EPrimitiveType.String))
			return false;
		if (IsSymbol(","))
		{
			Consume();
			CExpression OffsetExpression;
			if (!ParseExpression(OffsetExpression, EPrimitiveType.Integer))
				return false;
		}
		Emit(new CBLoadInstruction(FileNameExpression));
		return true;
	}



	private bool ReadConstStatement()
	{
		if (mToken.Type != ETokenType.Identifier)
		{
			ParseError(EParseError.ExpectedIdentifier);
			return false;
		}
		auto Identifier = new CIdentifier(mToken.Text);
		if (IsNameDefined(Identifier))
		{
			ParseError(EParseError.DuplicateDefinition);
			return false;
		}
		Consume();
		CValue Value;
		if (!(ExpectSymbol("=") && ReadConstExpression(PrimitiveTypeForConstName(Identifier),
				Value)))
			return false;
		mConstants[Identifier] = Value;
		return true;
	}



	private Nullable!(EPrimitiveType) PrimitiveTypeForConstName(CIdentifier aIdentifier)
	{
		final switch (aIdentifier.Suffix)
		{
			case ESuffix.None:
				return Nullable!(EPrimitiveType)();
			case ESuffix.Integer:
				return Nullable!(EPrimitiveType)(EPrimitiveType.Integer);
			case ESuffix.Long:
				return Nullable!(EPrimitiveType)(EPrimitiveType.Long);
			case ESuffix.Single:
				return Nullable!(EPrimitiveType)(EPrimitiveType.Single);
			case ESuffix.String:
				return Nullable!(EPrimitiveType)(EPrimitiveType.String);
		}
	}



	private bool ReadConstExpression(EPrimitiveType aType, out CValue aResult)
	{
		return ReadConstExpression(Nullable!(EPrimitiveType)(aType), aResult);
	}

	private bool ReadConstExpression(Nullable!(EPrimitiveType) aType, out CValue aResult)
	{
		CExpression Expression;
		if (!ParseExpression(Expression, aType))
			return false;
		if (Expression.Value is null)
		{
			ParseError(EParseError.ExpressionNotConstant);
			return false;
		}
		if (!aType.isNull && Expression.Value.Type != aType)
		{
			ParseError(EParseError.TypeMismatch);
			return false;
		}
		aResult = Expression.Value;
		return true;
	}



	private bool ReadDeclareStatement()
	{
		if (!IsKeyword("SUB") && !IsKeyword("FUNCTION"))
		{
			ParseError(EParseError.ExpectedKeyword);
			return false;
		}
		bool IsSub = IsKeyword("SUB");
		Consume();
		if (mToken.Type != ETokenType.Identifier)
		{
			ParseError(EParseError.ExpectedIdentifier);
			return false;
		}
		auto Identifier = new CIdentifier(mToken.Text);
		if (IsSub && Identifier.Suffix != ESuffix.None)
		{
			ParseError(EParseError.InvalidSubName);
			return false;
		}
		if (IsNameDefined(Identifier))
		{
			ParseError(EParseError.DuplicateDefinition);
			return false;
		}
		Consume();
		CArgument[] Arguments = [];
		bool[CIdentifier] ArgumentIdentifiers;
		if (!ExpectSymbol("("))
			return false;
		if (IsSymbol(")"))
			Consume();
		else
		{
			while (true)
			{
				CArgument Argument;
				if (!ReadArgument(Argument))
					return false;
				if (Argument.Identifier in ArgumentIdentifiers)
				{
					ParseError(EParseError.DuplicateDefinition);
					return false;
				}
				Arguments ~= [ Argument ];
				ArgumentIdentifiers[Argument.Identifier] = true;
				if (IsSymbol(")"))
				{
					Consume();
					break;
				}
				if (!ExpectSymbol(","))
					return false;
			}
		}
		if (IsSub)
			mProcedures[Identifier] = new CProcedureInfo(Arguments);
		else
			mProcedures[Identifier] = new CProcedureInfo(GetIdentifierType(Identifier, true),
					Arguments);
		return true;
	}



	private bool ReadArgument(out CArgument aArgument)
	out (result)
	{
		assert(!result || aArgument !is null);
	}
	body
	{
		if (mToken.Type != ETokenType.Identifier)
		{
			ParseError(EParseError.ExpectedIdentifier);
			return false;
		}
		auto Identifier = new CIdentifier(mToken.Text);
		Consume();
		CType Type;
		if (IsKeyword("AS"))
		{
			Type = GetIdentifierType(Identifier, false);
			if (Type !is null)
			{
				ParseError(EParseError.InvalidDeclaration);
				return false;
			}
			Consume();
			if (!ExpectType(Type))
				return false;
			if (Type.Type == EType.Record && canFind(Identifier.Name, '.'))
			{
				ParseError(EParseError.IdentifierWithPeriod);
				return false;
			}
		}
		else if (IsSymbol("("))
		{
			Consume();
			if (!ExpectSymbol(")"))
				return false;
			Type = CType.ArrayReference(GetIdentifierType(Identifier, true));
			// TODO: allow specifying type with AS
		}
		else
			Type = GetIdentifierType(Identifier, true);
		aArgument = new CArgument(Identifier, Type);
		return true;
	}



	private bool ReadDefStatement()
	{
		StartStatement();
		Consume();
		if (!ExpectKeyword("SEG"))
			return false;
		if (IsEndStatement())
		{
			CancelStatement();
			return true;
		}
		if (!ExpectSymbol("="))
			return false;
		CExpression AddressExpression;
		if (!ParseExpression(AddressExpression, EPrimitiveType.Reference))
			return false;
		Emit(new CDefSegInstruction(AddressExpression));
		return true;
	}



	private bool ReadDimStatement()
	{
		bool IsShared = IsKeyword("SHARED");
		if (IsShared)
			Consume();
		while (true)
		{
			if (mToken.Type != ETokenType.Identifier)
			{
				ParseError(EParseError.ExpectedIdentifier);
				return false;
			}
			auto Identifier = new CIdentifier(mToken.Text);
			if (IsNameDefined(Identifier))
			{
				ParseError(EParseError.DuplicateDefinition);
				return false;
			}
			Consume();
			CType Type = GetIdentifierType(Identifier, false);
			int[] Dimensions = null;
			if (IsSymbol("("))
			{
				Consume();
				// read subscripts
				while (true)
				{
					CValue Value;
					if (!ReadConstExpression(EPrimitiveType.Integer, Value))
						return false;
					Dimensions ~= [ Value.IntegralValue + 1 ];
					if (IsSymbol(")"))
					{
						Consume();
						break;
					}
					ExpectSymbol(",");
				}
			}
			if (IsKeyword("AS"))
			{
				if (Type !is null)
				{
					ParseError(EParseError.InvalidDeclaration);
					return false;
				}
				Consume();
				if (!ExpectType(Type))
					return false;
				if (Type.Type == EType.Record)
				{
					if (canFind(Identifier.Name, '.'))
					{
						ParseError(EParseError.IdentifierWithPeriod);
						return false;
					}
					if (Identifier.Name in mSharedVariablePrefixes ||
							Identifier.Name in mVariablePrefixes)
					{
						ParseError(EParseError.DuplicateDefinition);
						return false;
					}
				}
			}
			else if (Type is null)
				Type = CType.Primitive(EPrimitiveType.Integer);   // default type
			if (Dimensions !is null)   // if it is an array
				Type = CType.Array(Dimensions.length, Type);
			(IsShared ? mSharedVariables : mVariables)[Identifier] = new CVariableInfo(Type,
					Dimensions);
			int DotIndex = Identifier.Name.countUntil('.');
			if (DotIndex != -1)
				(IsShared ? mSharedVariablePrefixes : mVariablePrefixes)[Identifier.
						Name[0 .. DotIndex]] = true;

			if (IsSymbol(","))
			{
				Consume();
				continue;
			}
			return true;
		}
	}



	private bool ReadRedimStatement()
	{
		StartStatement();
		Consume();
		while (true)
		{
			if (mToken.Type != ETokenType.Identifier)
			{
				ParseError(EParseError.ExpectedIdentifier);
				return false;
			}
			auto Identifier = new CIdentifier(mToken.Text);
			Consume();
			CExpression SubscriptExpression;
			if (!(ExpectSymbol("(") && ParseExpression(SubscriptExpression,
					EPrimitiveType.Integer) && ExpectSymbol(")")))
				return false;
			CName Name;
			CType Type = GetVariableInfo(Identifier, true, Name);
			if (Type is null)
				ParseError(EParseError.UndeclaredVariable);   // TODO: can also be ambiguous variable
			Emit(new CRedimInstruction(Name, SubscriptExpression));
			if (!IsSymbol(","))
				return true;
			Consume();
		}
	}



	private bool ReadDoStatement()
	{
		StartStatement();
		Consume();
		bool Until = IsKeyword("UNTIL");
		CExpression ConditionExpression = null;
		if (Until || IsKeyword("WHILE"))
		{
			Consume();
			if (!ParseExpression(ConditionExpression, EPrimitiveType.Integer))
				return false;
		}
		mContext ~= [ SContext(EContextType.Loop, mInstructions.length) ];
		Emit(new CLoopStartInstruction(Until, ConditionExpression));
		return true;
	}



	private bool ReadLoopStatement()
	{
		StartStatement();
		Consume();
		SContext Context;
		if (!PopContext(EContextType.Loop, Context))
			return false;
		auto LoopStartInstruction = cast(CLoopStartInstruction)mInstructions[Context.
				InstructionIndex];
		bool Until = IsKeyword("UNTIL");
		CExpression ConditionExpression = null;
		if (Until || IsKeyword("WHILE"))
		{
			if (LoopStartInstruction.GetExpression() !is null)
			{
				ParseError(EParseError.InvalidBlockEnd);
				return false;
			}
			Consume();
			if (!ParseExpression(ConditionExpression, EPrimitiveType.Integer))
				return false;
		}
		// use magic to make the loop end instruction
		Emit(new CLoopStartInstruction(!Until, ConditionExpression,
				Context.InstructionIndex));
		LoopStartInstruction.SetNextInstruction(mInstructions.length);
		foreach (int Index; Context.ExitInstructions)
			(cast(CBranchInstruction)mInstructions[Index]).SetNextInstruction(mInstructions.length);
		return true;
	}



	private bool ReadEndStatement()
	{
		Consume();
		if (IsEndStatement())
		{
			ParseError(EParseError.Unsupported);
			return false;
		}
		if (mToken.Type == ETokenType.Keyword && canFind([ "FUNCTION", "SUB" ], mToken.Text))
		{
			foreach (int Index; mExitProcedureInstructions)
				(cast(CBranchInstruction)mInstructions[Index]).
						SetNextInstruction(mInstructions.length);
			mProgram.AddProcedure(mCurrentProcedure, mProcedures[mCurrentProcedure],
					mVariables, mArgumentVariables, mInstructions, mStackSize);
			Consume();
			return true;
		}
		else if (IsKeyword("IF"))
		{
			Consume();
			return ReadEndIfStatement();
		}
		else if (IsKeyword("SELECT"))
		{
			Consume();
			return ReadEndSelectStatement();
		}
		ParseError(EParseError.ExpectedKeyword);
		return false;
	}



	private bool ReadExitStatement()
	{
		StartStatement();
		Consume();
		if (mToken.Type == ETokenType.Keyword)
		{
			switch (mToken.Text)
			{
				case "FUNCTION":
				case "SUB":
					if (mCurrentProcedure !is null && (mToken.Text == "SUB") ==
						(mProcedures[mCurrentProcedure].ReturnType is null))
					{
						Consume();
						mExitProcedureInstructions ~= [ mInstructions.length ];
						Emit(new CBranchInstruction());
						return true;
					}
					else
					{
						ParseError(EParseError.InvalidBlockExit);
						return false;
					}
				case "DO":
				case "FOR":
					EContextType Type;
					final switch (mToken.Text)
					{
						case "DO":
							Type = EContextType.Loop;
							break;
						case "FOR":
							Type = EContextType.For;
							break;
					}
					foreach_reverse (ref SContext Context; mContext)
					{
						if (Context.Type == Type)
						{
							Consume();
							Context.ExitInstructions ~= [ mInstructions.length ];
							Emit(new CBranchInstruction());
							return true;
						}
					}
					ParseError(EParseError.InvalidBlockExit);
					return false;
				default:
					break;
			}
		}
		ParseError(EParseError.ExpectedKeyword);
		return false;
	}



	private bool ReadForStatement()
	{
		StartStatement();
		Consume();
		CExpression TargetExpression;
		CType Type;
		if (!ReadAssignableExpression(TargetExpression, Type))
			return false;
		if (Type.Type != EType.Primitive || !IsNumericType(Type.PrimitiveType))
		{
			ParseError(EParseError.ExpectedNumericVariable);
			return false;
		}
		CExpression StartExpression;
		CExpression EndExpression;
		if (!(ExpectSymbol("=") && ParseExpression(StartExpression, Type.PrimitiveType) &&
				ExpectKeyword("TO") && ParseExpression(EndExpression, Type.PrimitiveType)))
			return false;
		CExpression StepExpression;
		if (IsKeyword("STEP"))
		{
			Consume();
			if (!ParseExpression(StepExpression, Type.PrimitiveType))
				return false;
		}
		Emit(new CForInitInstruction(mCurrentStackSize, TargetExpression,
				StartExpression, EndExpression, StepExpression));
		mContext ~= [ SContext(EContextType.For, mInstructions.length) ];
		mStatementStarted = true;   // TODO: do this better
		Emit(new CForInstruction(mCurrentStackSize));
		AddStack(3);
		return true;
	}



	private bool NextStatement()
	{
		StartStatement();
		Consume();
		SContext Context;
		if (!PopContext(EContextType.For, Context))
			return false;
		auto ForInstruction = cast(CForInstruction)mInstructions[Context.InstructionIndex];
		mCurrentStackSize -= 3;
		Emit(new CForNextInstruction(mCurrentStackSize, Context.InstructionIndex));
		ForInstruction.SetNextInstruction(mInstructions.length);
		foreach (int Index; Context.ExitInstructions)
			(cast(CBranchInstruction)mInstructions[Index]).SetNextInstruction(mInstructions.length);
		return true;
	}



	private bool ReadIfStatement()
	{
		StartStatement();
		Consume();
		CExpression ConditionExpression;
		if (!ParseExpression(ConditionExpression, EPrimitiveType.Integer) || !ExpectKeyword("THEN"))
			return false;
		if (IsNewline())
		{
			if (!mFirstStatement)
			{
				ParseError(EParseError.NotFirstStatement);
				return false;
			}
		}
		else
		{
			mLineIfStatement = true;
			mLineIfFirst = true;
		}
		mContext ~= [ SContext(EContextType.If, mInstructions.length) ];
		Emit(new CIfInstruction(ConditionExpression));
		return true;
	}



	private bool ReadElseStatement()
	{
		StartStatement();
		Consume();
		SContext Context = mContext[$ - 1];
		if (Context.Type != EContextType.ElseIf && Context.Type != EContextType.If)
		{
			ParseError(EParseError.InvalidBlockEnd);
			return false;
		}
		mContext[$ - 1].InstructionIndex = mInstructions.length;
		mContext ~= [ SContext(EContextType.Else, mInstructions.length) ];
		Emit(new CBranchInstruction());
		auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
		Instruction.SetNextInstruction(mInstructions.length);
		return true;
	}



	private bool ReadElseIfStatement()
	{
		StartStatement();
		Consume();
		CExpression ConditionExpression;
		if (!(ParseExpression(ConditionExpression, EPrimitiveType.Integer) && ExpectKeyword("THEN")))
			return false;
		SContext Context = mContext[$ - 1];
		if (Context.Type != EContextType.ElseIf && Context.Type != EContextType.If)
		{
			ParseError(EParseError.InvalidBlockEnd);
			return false;
		}
		mContext[$ - 1].InstructionIndex = mInstructions.length;
		mInstructions ~= [ new CBranchInstruction() ];
		auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
		Instruction.SetNextInstruction(mInstructions.length);

		mContext ~= [ SContext(EContextType.ElseIf, mInstructions.length) ];
		Emit(new CIfInstruction(ConditionExpression));
		return true;
	}



	public bool ReadEndIfStatement()
	{
		while (true)
		{
			SContext Context = mContext[$ - 1];
			if (Context.Type == EContextType.Else)
			{
				mContext.length = mContext.length - 1;
				continue;
			}
			if (Context.Type != EContextType.ElseIf && Context.Type != EContextType.If)
			{
				ParseError(EParseError.InvalidBlockEnd);
				return false;
			}
			auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
			Instruction.SetNextInstruction(mInstructions.length);
			mContext.length = mContext.length - 1;
			if (Context.Type == EContextType.If)
				return true;
		}
	}



	private bool ReadLineInputStatement()
	{
		CExpression FileNumberExpression;
		CExpression TargetExpression;
		CType TargetType;
		if (!(ExpectSymbol("#") && ParseExpression(FileNumberExpression, EPrimitiveType.Integer) &&
				ExpectSymbol(",") && ReadAssignableExpression(TargetExpression, TargetType)))
			return false;
		if (TargetType.Type != EType.Primitive && TargetType.PrimitiveType != EPrimitiveType.String)
		{
			ParseError(EParseError.TypeMismatch);
			return false;
		}
		Emit(new CFileLineInputInstruction(FileNumberExpression, TargetExpression));
		return true;
	}



	private bool ReadLineGraphicsStatement()
	{
		CExpression X0Expression;
		CExpression Y0Expression;
		CExpression X1Expression;
		CExpression Y1Expression;
		CExpression ColourExpression;
		if (!(ExpectSymbol("(") && ParseExpression(X0Expression, EPrimitiveType.Integer) &&
				ExpectSymbol(",") && ParseExpression(Y0Expression, EPrimitiveType.Integer) &&
				ExpectSymbol(")") && ExpectSymbol("-") && ExpectSymbol("(") &&
				ParseExpression(X1Expression, EPrimitiveType.Integer) && ExpectSymbol(",") &&
				ParseExpression(Y1Expression, EPrimitiveType.Integer) && ExpectSymbol(")") &&
				ExpectSymbol(",") && ParseExpression(ColourExpression, EPrimitiveType.Integer)))
			return false;
		ELineType Type = ELineType.Line;
		if (IsSymbol(","))
		{
			Consume();
			if (!(mToken.Type == ETokenType.Identifier && (mToken.Text == "B" ||
					mToken.Text == "BF")))
			{
				ParseError(EParseError.ExpectedIdentifier);
				return false;
			}
			Consume();
			Type = mToken.Text == "B" ? ELineType.Rectangle : ELineType.FilledRectangle;
		}
		Emit(new CGraphicsLineInstruction(X0Expression, Y0Expression, X1Expression,
				Y1Expression, ColourExpression, Type));
		return true;
	}



	private bool ReadMidStatement()
	{
		StartStatement();
		Consume();
		CExpression TargetExpression;
		CType TargetType;
		CExpression StartExpression;
		//CExpression LengthExpression;
		CExpression Expression;
		if (!(ExpectSymbol("(") && ReadAssignableExpression(TargetExpression, TargetType) &&
				ExpectSymbol(",") && ParseExpression(StartExpression, EPrimitiveType.Integer) &&
				ExpectSymbol(")") && ExpectSymbol("=") && ParseExpression(Expression,
				EPrimitiveType.String)))
			return false;
		if (TargetType.Type != EType.Primitive && TargetType.PrimitiveType != EPrimitiveType.String)
		{
			ParseError(EParseError.TypeMismatch);
			return false;
		}
		Emit(new CMidInstruction(TargetExpression, StartExpression, Expression));
		return true;
	}



	private bool ReadOpenStatement()
	{
		StartStatement();
		Consume();
		CExpression PathExpression;
		if (!(ParseExpression(PathExpression, EPrimitiveType.String) && ExpectKeyword("FOR")))
			return false;
		if (mToken.Type != ETokenType.Keyword || !canFind([ "APPEND", "BINARY", "INPUT",
				"OUTPUT", "RANDOM" ], mToken.Text))
		{
			ParseError(EParseError.ExpectedKeyword);
			return false;
		}
		EFileOpenMode Mode;
		switch (mToken.Text)
		{
			case "BINARY":
				Mode = EFileOpenMode.Binary;
				break;
			case "INPUT":
				Mode = EFileOpenMode.Input;
				break;
			default:
				ParseError(EParseError.Unsupported);
				return false;
		}
		Consume();
		if (!ExpectKeyword("AS"))
			return false;
		if (IsSymbol("#"))
			Consume();
		CExpression FileNumberExpression;
		if (!ParseExpression(FileNumberExpression, EPrimitiveType.Integer))
			return false;
		Emit(new CFileOpenInstruction(PathExpression, Mode, FileNumberExpression));
		return true;
	}



	private bool ReadCloseStatement()
	{
		StartStatement();
		Consume();
		while (true)
		{
			if (IsSymbol("#"))
				Consume();
			CExpression FileNumberExpression;
			if (!ParseExpression(FileNumberExpression, EPrimitiveType.Integer))
				return false;
			Emit(new CFileCloseInstruction(FileNumberExpression));
			if (!IsSymbol(","))
				return true;
			mStatementStarted = true;   // TODO: do this a better way
			Consume();
		}
	}



	private bool ReadOutStatement()
	{
		StartStatement();
		Consume();
		CExpression PortExpression;
		CExpression DataExpression;
		if (!(ParseExpression(PortExpression, EPrimitiveType.Integer) && ExpectSymbol(",") &&
				ParseExpression(DataExpression, EPrimitiveType.Integer)))
			return false;
		Emit(new COutInstruction(PortExpression, DataExpression));
		return true;
	}



	private bool ReadPutFileStatement()
	{
		if (IsSymbol("#"))
			Consume();
		CExpression FileNumberExpression;
		if (!ParseExpression(FileNumberExpression, EPrimitiveType.Integer))
			return false;
		if (!(ExpectSymbol(",") && ExpectSymbol(",")))
			return false;
		CExpression SourceExpression;
		CType SourceType;
		if (!ReadAssignableExpression(SourceExpression, SourceType))
			return false;
		Emit(new CFilePutInstruction(FileNumberExpression, SourceExpression));
		return true;
	}



	private bool ReadPutGraphicsStatement()
	{
		CExpression XExpression;
		CExpression YExpression;
		if (!(ExpectSymbol("(") && ParseExpression(XExpression, EPrimitiveType.Integer) &&
				ExpectSymbol(",") && ParseExpression(YExpression, EPrimitiveType.Integer) &&
				ExpectSymbol(")") && ExpectSymbol(",")))
			return false;
		CExpression ImageExpression;
		if (!ParseExpression(ImageExpression, EPrimitiveType.Reference, true))
			return false;
		if (!ExpectSymbol(","))
			return false;
		if (mToken.Type != ETokenType.Keyword || !canFind([ "AND", "OR", "PSET" ],
				mToken.Text))
		{
			ParseError(EParseError.ExpectedKeyword);
			return false;
		}
		EGraphicsOperation Operation;
		final switch (mToken.Text)
		{
			case "AND":
				Operation = EGraphicsOperation.And;
				break;
			case "OR":
				Operation = EGraphicsOperation.Or;
				break;
			case "PSET":
				Operation = EGraphicsOperation.PSet;
				break;
		}
		Consume();
		Emit(new CGraphicsPutInstruction(XExpression, YExpression, ImageExpression, Operation));
		return true;
	}



	private bool ReadSeekStatement()
	{
		StartStatement();
		Consume();
		if (IsSymbol("#"))
			Consume();
		CExpression FileNumberExpression;
		CExpression PositionExpression;
		if (!(ParseExpression(FileNumberExpression, EPrimitiveType.Integer) && ExpectSymbol(",") &&
				ParseExpression(PositionExpression, EPrimitiveType.Integer)))
			return false;
		Emit(new CFileSeekInstruction(FileNumberExpression, PositionExpression));
		return true;
	}



	private bool ReadSelectStatement()
	{
		StartStatement();
		Consume();
		CExpression Expression;
		if (!(ExpectKeyword("CASE") && ParseExpression(Expression, Nullable!(EPrimitiveType)())))
			return false;
		mContext ~= [ SContext(EContextType.Select, mInstructions.length) ];
		mContext[$ - 1].SetSelectExpressionType(Expression.ResultType);
		Emit(new CSelectInstruction(mCurrentStackSize, Expression));
		AddStack(1);
		return true;
	}



	private bool ReadEndSelectStatement()
	{
		bool First = true;
		while (true)
		{
			SContext Context = mContext[$ - 1];
			if (!((!First && Context.Type == EContextType.Select) || Context.Type ==
					EContextType.SelectCase || Context.Type == EContextType.SelectCaseElse))
			{
				ParseError(EParseError.InvalidBlockEnd);
				return false;
			}
			mContext.length = mContext.length - 1;
			if (Context.Type == EContextType.Select)
				break;
			if (Context.Type == EContextType.SelectCase)
			{
				auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
				Instruction.SetNextInstruction(mInstructions.length);
			}
			if (First)
				First = false;
		}
		mCurrentStackSize--;
		return true;
	}



	private bool ReadCaseStatement()
	{
		StartStatement();
		Consume();
		SContext Context = mContext[$ - 1];
		mContext[$ - 1].InstructionIndex = mInstructions.length;
		if (IsKeyword("ELSE"))
		{
			Consume();
			if (Context.Type != EContextType.SelectCase)
			{
				ParseError(EParseError.InvalidBlockEnd);
				return false;
			}
			Emit(new CBranchInstruction());
			if (Context.Type == EContextType.SelectCase)
			{
				auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
				Instruction.SetNextInstruction(mInstructions.length);
			}
			mContext ~= [ SContext(EContextType.SelectCaseElse, mInstructions.length) ];
			// no instructions needed for this
		}
		else
		{
			if (Context.Type != EContextType.Select && Context.Type != EContextType.SelectCase)
			{
				ParseError(EParseError.InvalidBlockEnd);
				return false;
			}
			if (Context.Type == EContextType.SelectCase)
			{
				mInstructions ~= [ new CBranchInstruction() ];
				auto Instruction = cast(IBranchInstruction)mInstructions[Context.InstructionIndex];
				Instruction.SetNextInstruction(mInstructions.length);
			}
			CSelectCase[] Cases = [];
			while (true)
			{
				if (IsKeyword("IS"))
				{
					Consume();
					if (mToken.Type != ETokenType.Symbol || !canFind([ "<", "<=", ">",
							">=", "=", "<>" ], mToken.Text))
					{
						ParseError(EParseError.ExpectedRelationalOperator);
						return false;
					}
					ECaseIsOperator Operator;
					final switch (mToken.Text)
					{
						case "<":
							Operator = ECaseIsOperator.Less;
							break;
						case "<=":
							Operator = ECaseIsOperator.LessOrEqual;
							break;
						case ">":
							Operator = ECaseIsOperator.Greater;
							break;
						case ">=":
							Operator = ECaseIsOperator.GreaterOrEqual;
							break;
						case "=":
							Operator = ECaseIsOperator.Equal;
							break;
						case "<>":
							Operator = ECaseIsOperator.NotEqual;
							break;
					}
					Consume();
					CExpression Expression;
					if (!ParseExpression(Expression, Context.SelectExpressionType))
						return false;
					Cases ~= [ new CSelectCase(Operator, Expression) ];
				}
				else
				{
					CExpression Expression;
					if (!ParseExpression(Expression, Context.SelectExpressionType))
						return false;
					if (IsKeyword("TO"))
					{
						Consume();
						CExpression ToExpression;
						if (!ParseExpression(ToExpression, Context.SelectExpressionType))
							return false;
						Cases ~= [ new CSelectCase(Expression, ToExpression) ];
					}
					else
						Cases ~= [ new CSelectCase(Expression) ];
				}
				if (IsSymbol(","))
					Consume();
				else
					break;
			}
			mContext ~= [ SContext(EContextType.SelectCase, mInstructions.length) ];
			mContext[$ - 1].SetSelectExpressionType(Context.SelectExpressionType);
			Emit(new CSelectCaseInstruction(mCurrentStackSize - 1, Cases));
		}
		return true;
	}



	private bool ReadProcedureStatement(bool sub)
	{
		if (mToken.Type != ETokenType.Identifier)
		{
			ParseError(EParseError.ExpectedIdentifier);
			return false;
		}
		auto Identifier = new CIdentifier(mToken.Text);
		if (Identifier !in mProcedures || (sub != IsSub(Identifier)))
		{
			ParseError(EParseError.UndeclaredProcedure);
			return false;
		}
		mCurrentProcedure = Identifier;
		mArgumentVariables = [];
		mInstructions = [];
		mExitProcedureInstructions = [];
		mCurrentStackSize = 0;
		mStackSize = 0;
		Consume();
		CArgument[] Arguments = [];
		if (IsSymbol("("))
		{
			Consume();
			while (true)
			{
				CArgument Argument;
				if (!ReadArgument(Argument))
					return false;
				Arguments ~= [ Argument ];
				mArgumentVariables ~= Argument.Identifier;
				if (IsSymbol(")"))
				{
					Consume();
					break;
				}
				if (!ExpectSymbol(","))
					return false;
			}
		}
		if (!mProcedures[mCurrentProcedure].ArgumentsMatch(Arguments))
		{
			ParseError(EParseError.ArgumentsMismatch);
			return false;
		}
		mVariables = null;
		mVariablePrefixes = null;
		mInstructions = [];
		foreach (CArgument Argument; Arguments)
			mVariables[Argument.Identifier] = new CVariableInfo(Argument.Type, null);
		return true;
	}



	private bool ReadGetFileStatement()
	{
		StartStatement();
		Consume();
		if (IsSymbol("#"))
			Consume();
		CExpression FileNumberExpression;
		if (!ParseExpression(FileNumberExpression, EPrimitiveType.Integer))
			return false;
		if (!(ExpectSymbol(",") && ExpectSymbol(",")))
		{
			// TODO: support record expression
			ParseError(EParseError.Unsupported);
			return false;
		}
		CExpression TargetExpression;
		CType TargetType;
		if (!ReadAssignableExpression(TargetExpression, TargetType))
			return false;
		Emit(new CFileGetInstruction(FileNumberExpression, TargetExpression));
		return true;
	}



	private bool ReadTypeBlock()
	{
		if (mToken.Type != ETokenType.Identifier)
		{
			ParseError(EParseError.ExpectedIdentifier);
			return false;
		}
		if (mToken.Text in mTypes)
		{
			ParseError(EParseError.DuplicateDefinition);
			return false;
		}
		string TypeName = mToken.Text;
		Consume();
		if (!ExpectEndStatement())
			return false;
		STypeFieldInfo[] Fields;
		int Offset = 0;
		while (true)
		{
			if (IsKeyword("END"))
			{
				Consume();
				break;
			}
			if (mToken.Type != ETokenType.Identifier)
			{
				ParseError(EParseError.ExpectedIdentifier);
				return false;
			}
			if (canFind(mToken.Text, '.'))
			{
				ParseError(EParseError.IdentifierWithPeriod);
				return false;
			}
			string Name = mToken.Text;
			Consume();
			CType Type;
			if (!ExpectKeyword("AS") || !ExpectType(Type) || !ExpectEndStatement())
				return false;
			Fields ~= STypeFieldInfo(Name, Type, Offset);
			switch (Type.Type)
			{
				case EType.Primitive:
					Offset += 1;
					break;
				case EType.Record:
					Offset += Type.RecordSize;
					break;
				default:
					ParseError(EParseError.InvalidDeclaration);
					return false;
			}
		}
		if (!ExpectKeyword("TYPE"))
			return false;
		if (Fields.length == 0)
		{
			ParseError(EParseError.InvalidDeclaration);
			return false;
		}
		mTypes[TypeName] = STypeInfo(Fields, Offset);
		return true;
	}



	private bool ReadSubCallStatement(CIdentifier aIdentifier)
	{
		StartStatement();
		Consume();
		CExpression[] Arguments;
		foreach (int i, CArgument Argument; mProcedures[aIdentifier].Arguments)
		{
			if (i != 0)
			{
				if (!ExpectSymbol(","))
					return false;
			}
			CExpression ArgumentExpression;
			EPrimitiveType ArgumentType;
			switch (Argument.Type.Type)
			{
				case EType.ArrayReference:
					ArgumentType = EPrimitiveType.ArrayReference;
					break;
				case EType.Primitive:
					ArgumentType = Argument.Type.PrimitiveType;
					break;
				case EType.Record:
					ArgumentType = EPrimitiveType.Reference;
					break;
				default:
					assert(false);
			}
			if (!ParseExpression(ArgumentExpression, ArgumentType, true))
				return false;
			Arguments ~= [ ArgumentExpression ];
		}
		Emit(new CSubCallInstruction(aIdentifier, Arguments));
		return true;
	}



	private bool ReadExtensionCommandStatement()
	{
		StartStatement();
		if (mToken.Type != ETokenType.Keyword)
		{
			ParseError(EParseError.ExpectedKeyword);
			return false;
		}
		EExtensionCommand Command;
		EPrimitiveType[] ArgumentTypes;
		switch (mToken.Text)
		{
			case "QB.KEYRATE":
				Command = EExtensionCommand.KeyRate;
				ArgumentTypes = [ EPrimitiveType.Integer ];
				break;
			case "QB.KEYSTOP":
				Command = EExtensionCommand.KeyStop;
				break;
			default:
				assert(false);
		}
		Consume();
		CExpression[] Arguments;
		foreach (EPrimitiveType Type; ArgumentTypes)
		{
			if (Arguments.length != 0 && !ExpectSymbol(","))
				return false;
			CExpression ArgumentExpression;
			if (!ParseExpression(ArgumentExpression, Type, false))
				return false;
			Arguments ~= ArgumentExpression;
		}
		Emit(new CExtensionCommandInstruction(Command, Arguments));
		return true;
	}



	private bool ReadAssignmentStatement()
	{
		StartStatement();
		CExpression TargetExpression;
		CType TargetType;
		if (!ReadAssignableExpression(TargetExpression, TargetType))
			return false;
		EPrimitiveType TargetPrimitiveType;
		switch (TargetType.Type)
		{
			case EType.Primitive:
				TargetPrimitiveType = TargetType.PrimitiveType;
				break;
			case EType.Record:
				TargetPrimitiveType = EPrimitiveType.Reference;
				break;
			default:
				ParseError(EParseError.InvalidAssignmentTarget);
				return false;
		}
		CExpression Expression;
		if (!(ExpectSymbol("=") && ParseExpression(Expression, TargetPrimitiveType,
				TargetPrimitiveType == EPrimitiveType.Reference)))
			return false;
		Emit(new CAssignmentInstruction(TargetExpression, Expression,
				TargetPrimitiveType == EPrimitiveType.Reference));
		return true;
	}



	/**
	@param aExpression Receives an expression that evaluates to a reference to the
	variable being assigned to.
	@param aTargetType Receives the type of the target being assigned to.
	*/
	private bool ReadAssignableExpression(out CExpression aExpression,
			out CType aTargetType)
	out(rResult)
	{
		assert(!rResult || aExpression !is null);
	}
	body
	{
		auto Identifier = new CIdentifier(mToken.Text);
		int DotIndex = Identifier.Name.countUntil('.');
		string[] RecordPath;
		if (DotIndex != -1)
		{
			auto Prefix = new CIdentifier(Identifier.Name[0 .. DotIndex]);
			if (IsRecordVariable(Prefix))
			{
				RecordPath = Split(Identifier.Name[DotIndex + 1 .. $], '.');
				Identifier = Prefix;
			}
		}
		if (!(!IsNameDefined(Identifier) || IsVariable(Identifier) ||
				(Identifier == mCurrentProcedure && IsFunction(Identifier))))
		{
			ParseError(EParseError.DuplicateDefinition);
			return false;
		}
		Consume();
		CName Name;
		CType Type;
		CExpression[] Subexpressions;
		int RecordOffset;
		CType RecordType = null;
		if (IsSymbol("("))
		{
			if (IsNameDefined(Identifier))
				Type = GetVariableInfo(Identifier, true, Name);
			else
			{
				Name = new CName(Identifier, false);
				Type = CType.Array(1, GetIdentifierType(Identifier, true));
				mVariables[Identifier] = new CVariableInfo(Type, [ 10 ]);
			}
			Consume();
			while (true)
			{
				CExpression Subexpression;
				if (!ParseExpression(Subexpression, EPrimitiveType.Integer))
					return false;
				Subexpressions ~= Subexpression;
				if (IsSymbol(","))
					Consume();
				else
					break;
			}
			ExpectSymbol(")");
			if (IsSymbol("."))
			{
				Consume();
				if (mToken.Type != ETokenType.Identifier)
				{
					ParseError(EParseError.ExpectedIdentifier);
					return false;
				}
				RecordPath = Split(mToken.Text, '.');
				Consume();
				RecordType = Type.SubType;
			}
		}
		else if (!IsNameDefined(Identifier))
		{
			Name = new CName(Identifier, false);
			Type = GetIdentifierType(Identifier, true);
			mVariables[Identifier] = new CVariableInfo(Type, null);
		}
		else if (Identifier == mCurrentProcedure && IsFunction(Identifier))
		{
			Name = new CName(Identifier, false);
			Type = mProcedures[Identifier].ReturnType;
		}
		else
		{
			Type = GetVariableInfo(Identifier, false, Name);
			if (Type.Type == EType.Record)
				RecordType = Type;
		}
		CType FieldType = Type;
		if (RecordPath.length == 0)
			RecordOffset = 0;
		else
		{
			assert(RecordType !is null);
			RecordOffset = GetRecordOffset(RecordPath, RecordType, FieldType);
			if (RecordOffset == -1)
				return false;
		}
		auto ExpressionBuilder = new CExpressionBuilder(mEvaluator, Nullable!(EPrimitiveType)());
		ExpressionBuilder.Variable(Name, Type);
		aTargetType = Type;
		if (Subexpressions.length != 0)
		{
			Nullable!(EPrimitiveType) NextExpressionType;
			ExpressionBuilder.OpenBracket(NextExpressionType);
			foreach (CExpression Subexpression; Subexpressions)
				ExpressionBuilder.Subscript(Subexpression, NextExpressionType);
			ExpressionBuilder.CloseBracket();
			aTargetType = aTargetType.SubType;
		}
		if (RecordType !is null)
		{
			ExpressionBuilder.RecordOffset(RecordOffset, FieldType);
			aTargetType = FieldType;
		}
		aExpression = ExpressionBuilder.GetExpression(true);
		assert(aExpression !is null);
		return true;
	}



	private int GetRecordOffset(string[] aRecordPath, CType aType, out CType aFieldType)
	{
		STypeInfo TypeInfo = mTypes[aType.RecordName];
		int RecordOffset;
		foreach (int i, string FieldName; aRecordPath)
		{
			if (FieldName !in TypeInfo.Offsets)
			{
				ParseError(EParseError.InvalidFieldName);
				return false;
			}
			RecordOffset += TypeInfo.Offsets[FieldName];
			if (i == aRecordPath.length - 1)
				aFieldType = TypeInfo.Types[FieldName];
			else
			{
				if (TypeInfo.Types[FieldName].Type != EType.Record)   // if there is a dot after a non-record type field
				{
					ParseError(EParseError.IdentifierWithPeriod);
					return -1;
				}
				TypeInfo = mTypes[TypeInfo.Types[FieldName].RecordName];
			}
		}
		return RecordOffset;
	}



	private string[] Split(string aString, char aSplit)
	{
		int Start = 0;
		string[] Result = [];
		int End;
		for (End = 0; End < aString.length; End++)
		{
			if (aString[End] == aSplit)
			{
				Result ~= [ aString[Start .. End] ];
				Start = End + 1;
			}
		}
		Result ~= [ aString[Start .. End] ];
		return Result;
	}



	private bool ParseExpression(out CExpression aExpression, EPrimitiveType aResultType,
			bool aReference = false)
	{
		return ParseExpression(aExpression, Nullable!(EPrimitiveType)(aResultType), aReference);
	}

	private bool ParseExpression(out CExpression aExpression,
			Nullable!(EPrimitiveType) aResultType, bool aReference = false)
	{
		CExpressionParser ExpressionParser = new CExpressionParser(mEvaluator, aResultType);
		while (true)
		{
			bool Result;
			if (!ProcessExpressionToken(Result, ExpressionParser))
				return false;
			if (!Result)
			{
				if (!ExpressionParser.IsValid())
				{
					ParseError(EParseError.InvalidExpression);
					return false;
				}
				aExpression = ExpressionParser.GetExpression(aReference);
				if (aExpression is null)
				{
					ParseError(EParseError.TypeMismatch);
					return false;
				}
				return true;
			}
			Consume();
		}
	}



	private bool ProcessExpressionToken(out bool aResult, CExpressionParser aExpressionParser)
	{
		switch (mToken.Type)
		{
			case ETokenType.Keyword:
				aResult = aExpressionParser.Symbol(mToken.Text);
				break;
			case ETokenType.Identifier:
				auto Identifier = new CIdentifier(mToken.Text);
				int DotIndex = Identifier.Name.countUntil('.');
				string[] RecordPath;
				if (DotIndex != -1)
				{
					auto Prefix = new CIdentifier(Identifier.Name[0 .. DotIndex]);
					if (IsRecordVariable(Prefix))
					{
						RecordPath = Split(Identifier.Name[DotIndex + 1 .. $], '.');
						Identifier = Prefix;
					}
				}
				if (!IsNameDefined(Identifier))
				{
					ParseError(EParseError.UndeclaredVariable);
					return false;
				}
				else if (IsVariable(Identifier))
				{
					if (!aExpressionParser.IsVariableValid())
						aResult = false;
					else
					{
						Consume();
						bool IsArray = mToken.Type == ETokenType.Symbol && mToken.Text == "(";
						CName Name;
						CType Type = GetVariableInfo(Identifier, IsArray, Name);
						if (Type is null)
							ParseError(EParseError.UndeclaredVariable);   // TODO: can also be ambiguous variable
						aResult = aExpressionParser.Variable(Name, Type);
						if (RecordPath.length != 0)
						{
							CType FieldType;
							int RecordOffset = GetRecordOffset(RecordPath, Type, FieldType);
							if (RecordOffset == -1)
								return false;
							aExpressionParser.RecordOffset(RecordOffset, FieldType);
						}
						if (!ProcessExpressionToken(aResult, aExpressionParser))
							return false;
					}
				}
				else if (Identifier in mConstants)
					aResult = aExpressionParser.Value(mConstants[Identifier]);
				else if (IsFunction(Identifier))
				{
					auto Procedure = mProcedures[Identifier];
					EPrimitiveType[] ArgumentTypes = new EPrimitiveType[Procedure.Arguments.length];
					foreach (int i, CArgument Argument; Procedure.Arguments)
					{
						switch (Argument.Type.Type)
						{
							case EType.Primitive:
								ArgumentTypes[i] = Argument.Type.PrimitiveType;
								break;
							default:
								assert(false);
						}
					}
					aResult = aExpressionParser.Function(Identifier, Procedure.ReturnType,
							ArgumentTypes);
				}
				else
					assert(false);
				break;
			case ETokenType.Symbol:
				if (mToken.Text == ":")
					return true;
				if (mToken.Text == ".")
				{
					Consume();
					if (mToken.Type != ETokenType.Identifier)
					{
						ParseError(EParseError.ExpectedIdentifier);
						return false;
					}
					auto Identifier = new CIdentifier(mToken.Text);
					string[] RecordPath = Split(Identifier.Name, '.');
					CType FieldType;
					int RecordOffset = GetRecordOffset(RecordPath,
							aExpressionParser.GetPreviousVariableType().SubType, FieldType);
					if (RecordOffset == -1)
						return false;
					aExpressionParser.RecordOffset(RecordOffset, FieldType);
					aResult = true;
				}
				else
					aResult = aExpressionParser.Symbol(mToken.Text);
				break;
			case ETokenType.Number:
			case ETokenType.NumberFixed:
			case ETokenType.StringLiteral:
				CValue Value = GetTokenValue();
				if (Value is null)
					return false;
				aResult = aExpressionParser.Value(Value);
				break;
			case ETokenType.Newline:
				aResult = false;
				break;
			default:
				break;
		}
		return true;
	}



	private CType GetIdentifierType(CIdentifier aIdentifier, bool aDefault)
	{
		final switch (aIdentifier.Suffix)
		{
			case ESuffix.None:
				return aDefault ? CType.Primitive(EPrimitiveType.Integer) : null;
			case ESuffix.Integer:
				return CType.Primitive(EPrimitiveType.Integer);
			case ESuffix.Long:
				return CType.Primitive(EPrimitiveType.Long);
			case ESuffix.Single:
				return CType.Primitive(EPrimitiveType.Single);
			case ESuffix.String:
				return CType.Primitive(EPrimitiveType.String);
		}
	}



	private bool ExpectKeyword(string aKeyword)
	{
		if (IsKeyword(aKeyword))
		{
			Consume();
			return true;
		}
		ParseError(EParseError.ExpectedKeyword);
		return false;
	}



	private bool ExpectType(out CType aType)
	{
		if (mToken.Type == ETokenType.Keyword)
		{
			switch (mToken.Text)
			{
				case "INTEGER":
					Consume();
					aType = CType.Primitive(EPrimitiveType.Integer);
					return true;
				case "LONG":
					Consume();
					aType = CType.Primitive(EPrimitiveType.Long);
					return true;
				case "STRING":
					Consume();
					aType = CType.Primitive(EPrimitiveType.String);
					return true;
				default:
					break;
			}
		}
		else if (mToken.Type == ETokenType.Identifier)
		{
			STypeInfo* TypeInfo = mToken.Text in mTypes;
			if (TypeInfo != null)
			{
				aType = CType.Record(mToken.Text, TypeInfo.Size, TypeInfo.TypeList);
				Consume();
				return true;
			}
		}
		ParseError(EParseError.ExpectedType);
		return false;
	}



	private bool ExpectSymbol(string aSymbol)
	{
		if (mToken.Type == ETokenType.Symbol && mToken.Text == aSymbol)
		{
			Consume();
			return true;
		}
		ParseError(EParseError.ExpectedSymbol);
		return false;
	}



	private bool ExpectEndStatement()
	{
		if (IsEndStatement())
		{
			Consume();
			return true;
		}
		ParseError(EParseError.ExpectedEndStatement);
		return false;
	}



	private bool PopContext(EContextType aExpectedType, out SContext aContext)
	{
		if (mContext.length == 0 || mContext[$ - 1].Type != aExpectedType)
		{
			ParseError(EParseError.InvalidBlockEnd);
			return false;
		}
		aContext = mContext[$ - 1];
		mContext.length = mContext.length - 1;
		return true;
	}



	private bool IsNameDefined(CIdentifier aName)
	{
		if (aName in mConstants || aName in mProcedures || aName in mSharedVariables ||
				aName in mVariables)
			return true;
		int DotIndex = aName.Name.countUntil('.');
		if (DotIndex == -1)
			return false;
		return IsRecordVariable(new CIdentifier(aName.Name[0 .. DotIndex]));
	}



	private bool IsFunction(CIdentifier aIdentifier)
	{
		return aIdentifier in mProcedures && mProcedures[aIdentifier].ReturnType !is null;
	}



	private bool IsSub(CIdentifier aIdentifier)
	{
		return aIdentifier in mProcedures && mProcedures[aIdentifier].ReturnType is null;
	}



	private bool IsVariable(CIdentifier aName)
	{
		return aName in mSharedVariables || aName in mVariables;
	}



	private bool IsRecordVariable(CIdentifier aName)
	{
		auto SharedInfo = aName in mSharedVariables;
		if (SharedInfo !is null && SharedInfo.Type.Type == EType.Record)
			return true;
		auto LocalInfo = aName in mVariables;
		return LocalInfo !is null && LocalInfo.Type.Type == EType.Record;
	}



	private CType GetVariableInfo(CIdentifier aIdentifier, bool aIsArray, out CName aName)
	out(rResult)
	{
		assert((rResult is null) == (aName is null));
	}
	body
	{
		auto LocalInfo = aIdentifier in mVariables;
		if (LocalInfo !is null && canFind([ EType.Array, EType.ArrayReference ],
				LocalInfo.Type.Type) != aIsArray)
			LocalInfo = null;
		auto SharedInfo = aIdentifier in mSharedVariables;
		if (SharedInfo !is null && canFind([ EType.Array, EType.ArrayReference ],
				SharedInfo.Type.Type) != aIsArray)
			SharedInfo = null;
		if (LocalInfo !is null && SharedInfo !is null)
			return null;
		if (LocalInfo !is null)
		{
			aName = new CName(aIdentifier, false);
			return LocalInfo.Type;
		}
		if (SharedInfo !is null)
		{
			aName = new CName(aIdentifier, true);
			return SharedInfo.Type;
		}
		return null;
	}



	private bool IsKeyword(string aKeyword)
	{
		return mToken.Type == ETokenType.Keyword && mToken.Text == aKeyword;
	}



	private bool IsSymbol(string aSymbol)
	{
		return mToken.Type == ETokenType.Symbol && mToken.Text == aSymbol;
	}



	private bool IsEndStatement()
	{
		return IsNewline() || IsSymbol(":");
	}



	private bool IsNewline()
	{
		return mToken.Type == ETokenType.Newline;
	}



	private CValue GetTokenValue()
	{
		CValue Value = mToken.GetValue(mCP437Encoder);
		if (Value is null)
			ParseError(EParseError.InvalidNumber);
		return Value;
	}



	private void ReadToken()
	{
		mToken = mTokeniser.ReadToken();
		if (mParserCallback !is null)
			mParserCallback.ReadToken(mToken);
	}



	private void Consume()
	{
		ReadToken();
	}



	private void ParseError(EParseError aError)
	{
		mError = aError;
		mErrorToken = mToken;
	}



	private void AddStack(int aDelta)
	{
		mCurrentStackSize += aDelta;
		if (mCurrentStackSize > mStackSize)
			mStackSize = mCurrentStackSize;
	}



	private void StartStatement()
	in
	{
		assert(!mStatementStarted);
	}
	body
	{
		mStatementStarted = true;
		mInstructionLine = mToken.Line;
		mInstructionPos = mToken.Pos;
	}



	private void CancelStatement()
	in
	{
		assert(mStatementStarted);
	}
	body
	{
		mStatementStarted = false;
	}



	private void Emit(CInstruction aInstruction)
	in
	{
		assert(mStatementStarted);
	}
	body
	{
		aInstruction.SetSourceInfo(mInstructionLine, mInstructionPos);
		mInstructions ~= aInstruction;
		mStatementStarted = false;
	}



	public void SetError(string aMessage)
	{
	}



	public IFileSystem FileSystem()
	{
		return null;
	}



	public bool GetKey(out EKey aKey)
	{
		return false;
	}



	public immutable(ubyte[]) CommandLine()
	{
		return null;
	}
}