﻿module QBasic.Parser.CToken;
import std.algorithm, std.conv;
import QBasic.CCP437Encoder, QBasic.Program.CType, QBasic.Program.CValue;

enum ETokenType { Symbol, Keyword, Identifier, Number, NumberFixed, StringLiteral, Newline }

class CToken
{
	public immutable ETokenType Type;
	public immutable string Text;
	public immutable int Line;
	public immutable int Pos;
	private CValue mValue;

	invariant()
	{
		assert(Type == ETokenType.StringLiteral || Type == ETokenType.Newline ||
				Text.length != 0);
	}



	public this(ETokenType aType, string aText, int aLine, int aPos)
	{
		Type = aType;
		Text = aText;
		Line = aLine;
		Pos = aPos;
	}



	public CValue GetValue(CCP437Encoder aCP437Encoder)
	in
	{
		assert(canFind([ ETokenType.Number, ETokenType.NumberFixed,
				ETokenType.StringLiteral ], Type));
	}
	body
	{
		if (mValue is null)
		{
			switch (Type)
			{
				case ETokenType.StringLiteral:
					mValue = CValue.String(aCP437Encoder.Encode(Text));
					break;
				case ETokenType.Number:
					bool Negate = Text[0] == '-';
					int StartIndex = Text[0] == '-' ? 1 : 0;
					int Radix;
					if (Text[0] == '&')
					{
						StartIndex += 2;
						if (Text[1] == 'H')
							Radix = 16;
						else
							Radix = 8;
					}
					else
						Radix = 10;
					switch (Text[$ - 1])
					{
						case '%':
							int Number = ParseInt(Text[StartIndex .. $ - 1], Radix);
							if (Negate)
								Number = -Number;
							if (Number >= -32768 && Number <= 32767)
								mValue = CValue.Integer(Number);
							break;
						case '&':
							int Number = ParseInt(Text[StartIndex .. $ - 1], Radix);
							if (Negate)
								Number = -Number;
							mValue = CValue.Long(Number);
							break;
						default:
							int Number = ParseInt(Text[StartIndex .. $], Radix);
							if (Negate)
								Number = -Number;
							mValue = Number >= -32768 && Number <= 32767 ?
									CValue.Integer(Number) : CValue.Long(Number);
							break;
					}
					break;
				case ETokenType.NumberFixed:
				{
					float Number = to!float(Text);
					mValue = CValue.Single(Number);
					break;
				}
				default:
					assert(false);
			}
		}
		return mValue;
	}



	private int ParseInt(string aText, int aRadix)
	in
	{
		assert(aRadix >= 2 && aRadix <= 36);
	}
	body
	{
		int Value = 0;
		foreach (char c; aText)
		{
			int DigitValue;
			if (c >= '0' && c <= '9')
				DigitValue = c - '0';
			else if (c >= 'A' && c <= 'Z')
				DigitValue = 10 + (c - 'A');
			else if (c >= 'a' && c <= 'z')
				DigitValue = 10 + (c - 'a');
			else
				throw new Exception("Invalid integer character");   // TODO: give a compile error in this case
			if (DigitValue >= aRadix)
				throw new Exception("Digit out of range");   // TODO:
			Value *= aRadix;
			Value += DigitValue;
		}
		return Value;
	}
}