﻿module QBasic.Parser.CTokeniser;
import std.algorithm, std.stream, std.string;
import DUtil.CStringBuilder, DUtil.CTextFileReader;
import QBasic.Parser.CToken;

class CTokeniser
{
	private CTextFileReader mReader;
	private bool mEOF = false;
	private dchar mChar;
	private int mLine = 0;
	private int mPos = 0;

	public this(Stream aStream)
	{
		mReader = new CTextFileReader(aStream);
		ReadChar();
	}



	public CToken ReadToken()
	{
		while (!mEOF && canFind([ '\t', ' ' ], mChar))
			Consume();
		if (mEOF)
			return null;
		int Line = mLine;
		int Pos = mPos;
		if (mChar == '\n')   // newline
		{
			Consume();
			return new CToken(ETokenType.Newline, "", Line, Pos);
		}
		if ((mChar >= 'A' && mChar <= 'Z') || (mChar >= 'a' && mChar <= 'z'))   // keyword/identifier
		{
			CStringBuilder Token = new CStringBuilder();
			Token.AppendChar(mChar);
			Consume();
			while (!mEOF)
			{
				if ((mChar >= 'A' && mChar <= 'Z') || (mChar >= 'a' && mChar <= 'z') ||
						(mChar >= '0' && mChar <= '9') || mChar == '.')
				{
					Token.AppendChar(mChar);
					Consume();
				}
				else if (canFind([ '$', '%', '&', '!', '#' ], mChar) && !IsKeyword(Token.GetString()))
				{
					Token.AppendChar(mChar);
					Consume();
					break;
				}
				else
					break;
			}
			string TokenStr = Token.GetString();
			if (IsKeyword(TokenStr))
				return new CToken(ETokenType.Keyword, toUpper(TokenStr), Line, Pos);
			return new CToken(ETokenType.Identifier, TokenStr, Line, Pos);
		}
		CStringBuilder Number = new CStringBuilder();
		bool HasDot = false;
		if (mChar == '-')
		{
			Number.AppendChar(mChar);
			Consume();
		}
		if (mChar == '&')
		{
			Consume();
			dchar Type = mChar;
			if (Type == 'H' || Type == 'O')
			{
				Consume();
				Number.AppendChar('&');
				Number.AppendChar(Type);
				while (!mEOF)
				{
					if (Type == 'H')
					{
						if ((mChar >= '0' && mChar <= '9') || (mChar >= 'A' && mChar <= 'Z') ||
								(mChar >= 'a' && mChar <= 'z'))
						{
							Number.AppendChar(mChar);
							Consume();
						}
						else
							break;
					}
					else
					{
						if (mChar >= 0 && mChar <= '7')
						{
							Number.AppendChar(mChar);
							Consume();
						}
						else
							break;
					}
				}
				return new CToken(ETokenType.Number, Number.GetString(), Line, Pos);
			}
			else
				return new CToken(ETokenType.Symbol, [ '&' ], Line, Pos);
		}
		else if (mChar == '.')
		{
			Number.AppendChar(mChar);
			HasDot = true;
			Consume();
		}
		if (mChar >= '0' && mChar <= '9')   // number
		{
			Number.AppendChar(mChar);
			Consume();
			while (!mEOF)
			{
				if (mChar >= '0' && mChar <= '9')
				{
					Number.AppendChar(mChar);
					Consume();
				}
				else if (canFind([ '%', '&' ], mChar))
				{
					Number.AppendChar(mChar);
					Consume();
					break;
				}
				else if (!HasDot && mChar == '.')
				{
					Number.AppendChar(mChar);
					HasDot = true;
					Consume();
				}
				else
					break;
			}
			return new CToken(HasDot ? ETokenType.NumberFixed : ETokenType.Number,
					Number.GetString(), Line, Pos);
		}
		else if (Number.GetString() != "")
			return new CToken(ETokenType.Symbol, Number.GetString(), Line, Pos);
		if (mChar == '"')   // string literal
		{
			Consume();
			CStringBuilder Literal = new CStringBuilder();
			while (true)
			{
				if (mEOF || mChar == '\n')
					break;   // implicit end of literal
				if (mChar == '"')
				{
					Consume();
					break;
				}
				Literal.AppendChar(mChar);
				Consume();
			}
			return new CToken(ETokenType.StringLiteral, Literal.GetString(), Line, Pos);
		}
		if (mChar == '\'')   // comment
		{
			Consume();
			CStringBuilder Comment = new CStringBuilder();
			while (!mEOF)
			{
				if (mChar == '\n')
				{
					Consume();
					break;
				}
				Comment.AppendChar(mChar);
				Consume();
			}
			return new CToken(ETokenType.Newline, Comment.GetString(), Line, Pos);
		}
		dchar SymbolChar = mChar;
		CStringBuilder Symbol = new CStringBuilder();
		Symbol.AppendChar(SymbolChar);
		Consume();
		if (((SymbolChar == '>' || SymbolChar == '<') && mChar == '=') ||
				(SymbolChar == '<' && mChar == '>'))
		{
			Symbol.AppendChar(mChar);
			Consume();
		}
		return new CToken(ETokenType.Symbol, Symbol.GetString(), Line, Pos);
	}



	public void SkipLine()
	{
		while (!mEOF && mChar != '\n')
			Consume();
	}



	private void ReadChar()
	{
		if (!mReader.ReadChar(mChar))
			mEOF = true;
	}



	private void Consume()
	{
		if (mChar == '\n')
		{
			mLine++;
			mPos = 0;
		}
		else
			mPos++;
		ReadChar();
	}



	private bool IsKeyword(string aToken)
	{
		return canFind([ "AND", "APPEND", "AS", "ASC", "BINARY", "BLOAD", "CASE", "CHR$",
				"CLOSE", "COMMAND$", "CONST", "DECLARE", "DEF", "DEFINT", "DIM", "DO", "ELSE",
				"ELSEIF", "END", "EOF", "EXIT", "FOR", "FREEFILE", "FUNCTION", "GET", "IF",
				"INKEY$", "INPUT", "INT", "INTEGER", "IS", "LEN", "LINE", "LOF", "LONG",
				"LOOP", "LTRIM$", "MID$", "MOD", "NEXT", "NOT", "OPEN", "OR", "OUT", "OUTPUT",
				"PSET", "PUT", "RANDOM", "RANDOMIZE", "REDIM", "REM", "RND", "RTRIM$",
				"SCREEN", "SEEK", "SEG", "SELECT", "SGN", "SHARED", "SPACE$", "STEP", "STR$",
				"STRING", "SUB", "THEN", "TIMER", "TO", "TYPE", "UNTIL", "VAL", "VARSEG",
				"WIDTH", "WHILE", "XOR", "QB.KEYRATE", "QB.KEYSTOP" ], toUpper(aToken));
	}
}