﻿module QBasic.Parser.Expressions.CExpressionBuilder;
import std.array, std.typecons;
import QBasic.Interpreter.CEvaluator, QBasic.Parser.Expressions.Nodes,
		QBasic.Program.CIdentifier, QBasic.Program.CName, QBasic.Program.CType,
		QBasic.Program.CValue, QBasic.Program.EBuiltinFunction, QBasic.Program.EOperator,
		QBasic.Program.Expressions.CExpression, QBasic.Program.Expressions.Units;

class CExpressionBuilder
{
	private enum EMode { ArrayReference, ExpectOperand, ExpectOperator,
			ExpectArgumentsOpenBracket, ExpectArgument, ExpectArrayOpenBracket,
			ExpectSubscript };

	private CEvaluator mEvaluator;
	private immutable Nullable!(EPrimitiveType) mResultType;
	private EMode mMode = EMode.ExpectOperand;
	private CExpressionNode[] mNodes;
	private int[] mFirstNode;
	private CFunctionExpressionNode mCurrentFunction;
	private CVariableExpressionNode mCurrentVariable;
	private CVariableExpressionNode mPreviousVariable = null;
	private EPrimitiveType[] mArgumentTypes;
	private int mNumSubscripts;
	private bool mTakeArgumentReference = false;
	private bool mTypeError = false;

	invariant()
	{
		assert(mFirstNode.length != 0);
	}

	public this(CEvaluator aEvaluator, Nullable!(EPrimitiveType) aResultType)
	{
		mEvaluator = aEvaluator;
		mResultType = aResultType;
		mFirstNode = [ 0 ];
	}



	public bool IsEmpty()
	{
		return mNodes.length == 0 && mMode == EMode.ExpectOperand;
	}



	public bool IsValid()
	{
		return (mNodes.length != 0 && mMode == EMode.ExpectOperator) ||
				mMode == EMode.ArrayReference;
	}



	public bool TakeArgumentReference()
	{
		return mTakeArgumentReference;
	}



	public CExpression GetExpression(bool aReference)
	in
	{
		assert(IsValid());
	}
	body
	{
		MarkNodeComplete(0);
		if (aReference && mNodes.length == 1 && mNodes[0].NodeType ==
				EExpressionNodeType.Variable)
		{
			auto Node = mNodes[0].AsVariable();
			if (!Node.ArrayReference)
				Node.SetReference();
			// TODO: should this function fail if node is an ArrayReference but aReference
			// is false?
		}
		CUnit[] Units;
		foreach (CExpressionNode Node; mNodes)
			Units ~= GetUnitForNode(Node);
		EPrimitiveType ResultType = mResultType.isNull ? mNodes[0].Type : mResultType;
		if (!IsTypeConvertibleTo(mNodes[0].Type, ResultType))
			return null;
		return new CExpression(Units, ResultType);
	}



	private bool IsTypeConvertibleTo(EPrimitiveType aSource, EPrimitiveType aTarget)
	{
		if (IsNumericType(aTarget))
			return IsNumericType(aSource);
		if (aTarget == EPrimitiveType.String)
			return aSource == EPrimitiveType.String;
		if (aTarget == EPrimitiveType.ArrayReference || aTarget == EPrimitiveType.Reference)
			return true;
		assert(false);
	}



	private CUnit GetUnitForNode(CExpressionNode aNode)
	{
		final switch (aNode.NodeType)
		{
			case EExpressionNodeType.BinaryOperator:
				auto BinaryOperator = aNode.AsBinaryOperator();
				return new CBinaryOperatorUnit(BinaryOperator.Operator, BinaryOperator.Type);
			case EExpressionNodeType.BuiltinFunction:
				auto BuiltinFunction = aNode.AsBuiltinFunction();
				return new CBuiltinFunctionUnit(BuiltinFunction.Function, BuiltinFunction.Arguments);
			case EExpressionNodeType.UnaryOperator:
				auto UnaryOperator = aNode.AsUnaryOperator();
				return new CUnaryOperatorUnit(UnaryOperator.Operator, UnaryOperator.Type);
			case EExpressionNodeType.UserFunction:
				auto UserFunction = aNode.AsUserFunction();
				return new CUserFunctionUnit(UserFunction.Name, UserFunction.Arguments);
			case EExpressionNodeType.Value:
				return new CValueUnit(aNode.AsValue().Value);
			case EExpressionNodeType.Variable:
				auto Variable = aNode.AsVariable();
				return new CVariableUnit(Variable.Name, Variable.VariableType,
						Variable.TargetType, Variable.Subscripts, Variable.RecordOffset,
						Variable.ArrayReference, Variable.Reference);
		}
	}



	public bool BuiltinFunction(EBuiltinFunction aFunction, EPrimitiveType[] aArgumentTypes,
			bool aReference = false)
	in
	{
		assert(!aReference || aArgumentTypes.length == 1);
	}
	body
	{
		mTakeArgumentReference = aReference;
		return Function(new CBuiltinFunctionExpressionNode(aFunction), aArgumentTypes);
	}



	public bool UserFunction(CIdentifier aName, CType aType, EPrimitiveType[] aArgumentTypes)
	{
		mTakeArgumentReference = true;
		return Function(new CUserFunctionExpressionNode(aName, aType.PrimitiveType),
				aArgumentTypes);
	}



	private bool Function(CFunctionExpressionNode aFunctionNode, EPrimitiveType[] aArgumentTypes)
	{
		if (mMode != EMode.ExpectOperand)
			return false;
		AppendOperand(aFunctionNode);
		if (aArgumentTypes.length != 0)
		{
			mMode = EMode.ExpectArgumentsOpenBracket;
			mCurrentFunction = aFunctionNode;
			mArgumentTypes = aArgumentTypes;
		}
		return true;
	}



	public bool Expression(CExpression aExpression, out Nullable!(EPrimitiveType) aNextExpressionType)
	{
		switch (mMode)
		{
			case EMode.ExpectArgument:
				return Argument(aExpression, aNextExpressionType);
			case EMode.ExpectSubscript:
				return Subscript(aExpression, aNextExpressionType);
			default:
				return false;
		}
	}



	public bool Argument(CExpression aExpression, out Nullable!(EPrimitiveType) aNextExpressionType)
	in
	{
		assert(mMode == EMode.ExpectArgument);
	}
	body
	{
		if (mCurrentFunction.Arguments.length == mArgumentTypes.length)
			return false;
		mCurrentFunction.AddArgument(aExpression);
		if (mCurrentFunction.Arguments.length < mArgumentTypes.length)
			aNextExpressionType = mArgumentTypes[mCurrentFunction.Arguments.length];
		return true;
	}



	public bool Subscript(CExpression aExpression, out Nullable!(EPrimitiveType) aNextExpressionType)
	in
	{
		assert(mMode == EMode.ExpectSubscript);
	}
	body
	{
		if (mNumSubscripts != 0 && mCurrentVariable.Subscripts.length == mNumSubscripts)
			return false;
		mCurrentVariable.AddSubscript(aExpression);
		if (mCurrentVariable.Subscripts.length < mNumSubscripts)
			aNextExpressionType = EPrimitiveType.Integer;
		return true;
	}



	public bool OpenBracket(out Nullable!(EPrimitiveType) aNewExpressionType)
	{
		switch (mMode)
		{
			case EMode.ExpectArgumentsOpenBracket:
				mMode = EMode.ExpectArgument;
				aNewExpressionType = mArgumentTypes[0];
				return true;
			case EMode.ExpectArrayOpenBracket:
				mMode = EMode.ExpectSubscript;
				aNewExpressionType = EPrimitiveType.Integer;
				return true;
			case EMode.ExpectOperand:
				mFirstNode ~= mNodes.length;
				return true;
			default:
				return false;
		}
	}



	/**
	Returns true if the expression is valid after applying the close-bracket or false if
	it is not valid to apply the close-bracket.
	*/
	public bool CloseBracket()
	{
		switch (mMode)
		{
			case EMode.ExpectArgument:
				if (mCurrentFunction.Arguments.length != mArgumentTypes.length)
					return false;
				mMode = EMode.ExpectOperator;
				mCurrentFunction = null;
				return true;
			case EMode.ExpectSubscript:
				if (mCurrentVariable.Subscripts.length == 0)
					return ArrayReference();
				if (mNumSubscripts != 0 && mCurrentVariable.Subscripts.length != mNumSubscripts)
					return false;
				if (mCurrentVariable.VariableType.SubType.Type == EType.Record)
					mPreviousVariable = mCurrentVariable;
				mCurrentVariable = null;
				mMode = EMode.ExpectOperator;
				return true;
			case EMode.ExpectOperator:
				if (mFirstNode.length == 1)
					return false;
				MarkNodeComplete(mFirstNode[$ - 1]);
				mFirstNode.length = mFirstNode.length - 1;
				return true;
			default:
				return false;
		}
	}



	public bool Value(CValue aValue)
	{
		return AppendOperand(new CValueExpressionNode(aValue));
	}



	public bool IsVariableValid()
	{
		return mMode == EMode.ExpectOperand;
	}



	public bool Variable(CName aName, CType aType)
	in
	{
		assert(aName !is null && aType !is null);
	}
	body
	{
		CVariableExpressionNode VariableNode = new CVariableExpressionNode(aName, aType);
		if (!AppendOperand(VariableNode))
			return false;
		final switch (aType.Type)
		{
			case EType.Array:
			case EType.ArrayReference:
				mMode = EMode.ExpectArrayOpenBracket;
				mCurrentVariable = VariableNode;
				// TODO: can array references reference multi-dimensional arrays?
				mNumSubscripts = aType.Type == EType.Array ? aType.NumDimensions : 1;
				break;
			case EType.Primitive:
				break;
			case EType.Record:
				mPreviousVariable = VariableNode;
				break;
		}
		return true;
	}



	private bool ArrayReference()
	{
		if (!(mNodes.length == 1 && mNodes[0].NodeType == EExpressionNodeType.Variable))
			return false;
		mNodes[0].AsVariable().SetArrayReference();
		mMode = EMode.ArrayReference;
		return true;
	}



	public void RecordOffset(int aOffset, CType aFieldType)
	in
	{
		assert(aOffset >= 0);
	}
	body
	{
		mPreviousVariable.SetRecordOffset(aOffset, aFieldType);
		mPreviousVariable = null;
	}



	public CType GetPreviousVariableType()
	{
		return mPreviousVariable.VariableType;
	}



	public bool AppendOperand(CExpressionNode aNode)
	{
		if (mMode != EMode.ExpectOperand)
			return false;
		mMode = EMode.ExpectOperator;
		mNodes ~= aNode;
		aNode.SetComplete(1);
		return true;
	}



	public bool Minus()
	{
		if (mMode == EMode.ExpectOperator)
			return BinaryOperator(EOperator.Subtract);
		return UnaryOperator(EUnaryOperator.Negate);
	}



	public bool Plus()
	{
		if (mMode != EMode.ExpectOperator)
			return false;
		if (mNodes[$ - 1].Type == EPrimitiveType.String)
			return BinaryOperator(EOperator.Concatenate);
		return BinaryOperator(EOperator.Add);
	}



	public bool BinaryOperator(EOperator aOperator)
	{
		if (mMode != EMode.ExpectOperator)
			return false;
		mMode = EMode.ExpectOperand;
		int Index = mFirstNode[$ - 1];
		while (Index < mNodes.length)
		{
			auto Node = mNodes[Index];
			if (mNodes[Index].Complete)
				break;
			bool HasHigherPrecedence;
			if (Node.NodeType == EExpressionNodeType.UnaryOperator)
				HasHigherPrecedence = false;
			else
				HasHigherPrecedence = HigherPrecedence(aOperator,
						Node.AsBinaryOperator().Operator);
			if (!HasHigherPrecedence)
				break;
			Index += 1 + mNodes[Index + 1].SubtreeSize;
		}
		auto Node = new CBinaryOperatorExpressionNode(aOperator);
		mNodes.insertInPlace(Index, [ Node ]);
		MarkNodeComplete(Index + 1);
		return !mTypeError;
	}



	private Nullable!EPrimitiveType BinaryOperatorResultType(EOperator aOperator,
			EPrimitiveType aLeft, EPrimitiveType aRight)
	{
		final switch (aOperator)
		{
			case EOperator.Add:
			case EOperator.Exponent:
			case EOperator.Subtract:
			case EOperator.Multiply:
				if (NumericTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(GreatestPrecisionType(aLeft, aRight));
				return Nullable!EPrimitiveType();
			case EOperator.Divide:
				if (NumericTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(EPrimitiveType.Single);
				return Nullable!EPrimitiveType();
			case EOperator.DivideInt:
			case EOperator.Modulus:
				if (NumericTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(
							aLeft != EPrimitiveType.Integer || aRight != EPrimitiveType.Integer
							? EPrimitiveType.Long : EPrimitiveType.Integer);
				return Nullable!EPrimitiveType();
			case EOperator.Concatenate:
				if (aLeft == EPrimitiveType.String || aRight == EPrimitiveType.String)
					return Nullable!EPrimitiveType(EPrimitiveType.String);
				return Nullable!EPrimitiveType();
			case EOperator.BitwiseAnd:
			case EOperator.BitwiseOr:
			case EOperator.BitwiseXor:
				if (IntegralTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(GreatestPrecisionType(aLeft, aRight));
				return Nullable!EPrimitiveType();
			case EOperator.CompareEqual:
			case EOperator.CompareNotEqual:
				if ((aLeft == EPrimitiveType.String && aRight == EPrimitiveType.String) ||
						NumericTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(EPrimitiveType.Integer);
				return Nullable!EPrimitiveType();
			case EOperator.CompareLess:
			case EOperator.CompareLessOrEqual:
			case EOperator.CompareGreater:
			case EOperator.CompareGreaterOrEqual:
				if (NumericTypes(aLeft, aRight))
					return Nullable!EPrimitiveType(EPrimitiveType.Integer);
				return Nullable!EPrimitiveType();
		}
	}



	private bool IntegralTypes(EPrimitiveType aType0, EPrimitiveType aType1)
	{
		return IntegralType(aType0) && IntegralType(aType1);
	}



	private bool IntegralType(EPrimitiveType aType)
	{
		switch (aType)
		{
			case EPrimitiveType.Integer:
			case EPrimitiveType.Long:
				return true;
			default:
				return false;
		}
	}



	private bool NumericTypes(EPrimitiveType aType0, EPrimitiveType aType1)
	{
		return IsNumericType(aType0) && IsNumericType(aType1);
	}



	private EPrimitiveType GreatestPrecisionType(EPrimitiveType aType0, EPrimitiveType aType1)
	{
		return TypePrecisionLevel(aType0) >= TypePrecisionLevel(aType1) ? aType0 : aType1;
	}



	private int TypePrecisionLevel(EPrimitiveType aType)
	{
		switch (aType)
		{
			case EPrimitiveType.Integer:
				return 0;
			case EPrimitiveType.Long:
				return 1;
			case EPrimitiveType.Single:
				return 2;
			default:
				assert(false);
		}
	}



	private int MarkNodeComplete(int aIndex)
	{
		// TODO: rewrite to be not recursive
		auto Node = mNodes[aIndex];
		if (Node.Complete)
			return Node.SubtreeSize;
		int SubtreeSize;
		switch (Node.NodeType)
		{
			case EExpressionNodeType.BinaryOperator:
				auto LeftNode = mNodes[aIndex + 1];
				assert(LeftNode.Complete);
				SubtreeSize = 1 + LeftNode.SubtreeSize +
						MarkNodeComplete(aIndex + 1 + LeftNode.SubtreeSize);
				if (SimplifyBinaryOperator(aIndex))
					SubtreeSize = 1;
				else
				{
					auto BinaryOperator = Node.AsBinaryOperator();
					auto RightNode = mNodes[aIndex + 1 + LeftNode.SubtreeSize];
					auto Type = BinaryOperatorResultType(BinaryOperator.Operator,
							LeftNode.Type, RightNode.Type);
					if (Type.isNull)
					{
						mTypeError = true;
						return SubtreeSize;
					}
					BinaryOperator.SetType(Type);
				}
				break;
			case EExpressionNodeType.UnaryOperator:
				SubtreeSize = 1 + MarkNodeComplete(aIndex + 1);
				if (SimplifyUnaryOperator(aIndex))
					SubtreeSize = 1;
				else
				{
					auto UnaryOperator = Node.AsUnaryOperator();
					auto ChildNode = mNodes[aIndex + 1];
					auto Type = UnaryOperatorResultType(UnaryOperator.Operator, ChildNode.Type);
					if (Type.isNull)
					{
						mTypeError = true;
						return SubtreeSize;
					}
					UnaryOperator.SetType(Type);
				}
				break;
			default:
				assert(false);
		}
		Node.SetComplete(SubtreeSize);
		return SubtreeSize;
	}



	public bool UnaryOperator(EUnaryOperator aOperator)
	{
		if (mMode != EMode.ExpectOperand)
			return false;
		auto Node = new CUnaryOperatorExpressionNode(aOperator);
		mNodes ~= Node;
		return true;
	}



	public Nullable!EPrimitiveType UnaryOperatorResultType(EUnaryOperator aOperator, EPrimitiveType aType)
	{
		final switch (aOperator)
		{
			case EUnaryOperator.Negate:
				if (IsNumericType(aType))
					return Nullable!EPrimitiveType(aType);
				return Nullable!EPrimitiveType();
			case EUnaryOperator.BitwiseNot:
				if (IntegralType(aType))
					return Nullable!EPrimitiveType(aType);
				return Nullable!EPrimitiveType();
		}
	}



	private bool HigherPrecedence(EOperator aOperator0, EOperator aOperator1)
	{
		return PrecedenceValue(aOperator0) > PrecedenceValue(aOperator1);
	}



	private int PrecedenceValue(EOperator aOperator)
	{
		final switch (aOperator)
		{
			case EOperator.Exponent:
				return 6;
			case EOperator.Multiply:
			case EOperator.Divide:
				return 5;
			case EOperator.DivideInt:
				return 4;
			case EOperator.Modulus:
				return 3;
			case EOperator.Add:
			case EOperator.Concatenate:
			case EOperator.Subtract:
				return 2;
			case EOperator.CompareEqual:
			case EOperator.CompareNotEqual:
			case EOperator.CompareLess:
			case EOperator.CompareLessOrEqual:
			case EOperator.CompareGreater:
			case EOperator.CompareGreaterOrEqual:
				return 1;
			case EOperator.BitwiseAnd:
			case EOperator.BitwiseOr:
			case EOperator.BitwiseXor:
				return 0;
		}
	}



	/**
	Attempts to simplify the binary operator at the specified index.
	@return true if the operator was simplified to a value node, false if it was not.
	*/
	public bool SimplifyBinaryOperator(int aIndex)
	{
		auto OperatorNode = mNodes[aIndex].AsBinaryOperator();
		if (mNodes[aIndex + 1].NodeType != EExpressionNodeType.Value ||
				mNodes[aIndex + 2].NodeType != EExpressionNodeType.Value)
			return false;
		EOperator Operator = OperatorNode.Operator;
		CValue Value0 = mNodes[aIndex + 1].AsValue().Value;
		CValue Value1 = mNodes[aIndex + 2].AsValue().Value;
		auto Type = BinaryOperatorResultType(Operator, Value0.Type, Value1.Type);
		if (Type.isNull)
			return false;
		auto Result = mEvaluator.BinaryOperator(Value0, Value1, Operator, Type);
		if (Result is null)
			return false;
		CExpressionNode NewNode = new CValueExpressionNode(Result);
		NewNode.SetComplete(1);
		mNodes.replaceInPlace(aIndex, aIndex + 3, [ NewNode ]);
		return true;
	}



	public bool SimplifyUnaryOperator(int aIndex)
	{
		auto OperatorNode = mNodes[aIndex].AsUnaryOperator();
		if (mNodes[aIndex + 1].NodeType != EExpressionNodeType.Value)
			return false;
		EUnaryOperator Operator = OperatorNode.Operator;
		CValue Value = mNodes[aIndex + 1].AsValue().Value;
		auto Result = mEvaluator.UnaryOperator(Value, Operator);
		if (Result is null)
			return false;
		CExpressionNode NewNode = new CValueExpressionNode(Result);
		NewNode.SetComplete(1);
		mNodes.replaceInPlace(aIndex, aIndex + 2, [ NewNode ]);
		return true;
	}
}