﻿module QBasic.Program.CType;
import std.algorithm, std.typecons;
import QBasic.Program.CValue;

enum EType { Array, ArrayReference, Primitive, Record };
enum EPrimitiveType { ArrayReference, Integer, Long, Reference, Single, String };

bool IsNumericType(EPrimitiveType aType)
{
	final switch (aType)
	{
		case EPrimitiveType.Integer:
		case EPrimitiveType.Long:
		case EPrimitiveType.Single:
			return true;
		case EPrimitiveType.ArrayReference:
		case EPrimitiveType.Reference:
		case EPrimitiveType.String:
			return false;
	}
}



int PrimitiveTypeByteSize(EPrimitiveType aType)
{
	final switch (aType)
	{
		case EPrimitiveType.Integer:
			return 2;
		case EPrimitiveType.Long:
		case EPrimitiveType.Single:
			return 4;
		case EPrimitiveType.String:
		case EPrimitiveType.ArrayReference:
		case EPrimitiveType.Reference:
			assert(false);
	}
}



class CType
{
	public immutable EType Type;
	private immutable Nullable!(EPrimitiveType) mPrimitiveType;
	private immutable int mNumDimensions;
	private CType mSubType;
	private immutable string mRecordName;
	private immutable int mRecordSize;
	private CType[] mRecordFieldTypes;

	public static CType Array(int aNumDimensions, CType aSubType)
	{
		return new CType(EType.Array, Nullable!(EPrimitiveType)(), aNumDimensions, aSubType,
				null, 0, null);
	}



	public static CType ArrayReference(CType aSubType)
	{
		return new CType(EType.ArrayReference, Nullable!(EPrimitiveType)(), 0, aSubType,
				null, 0, null);
	}



	public static CType Primitive(EPrimitiveType aPrimitiveType)
	{
		return new CType(EType.Primitive, Nullable!(EPrimitiveType)(aPrimitiveType), 0, null,
				null, 0, null);
	}



	public static CType Record(string aName, int aSize, CType[] aFieldTypes)
	{
		return new CType(EType.Record, Nullable!(EPrimitiveType)(), 0, null, aName, aSize,
				aFieldTypes);
	}



	private this(EType aType, Nullable!(EPrimitiveType) aPrimitiveType, int aNumDimensions,
			CType aSubType, string aRecordName, int aRecordSize, CType[] aRecordFieldTypes)
	in
	{
		assert(aType == EType.Primitive ||
				(aType == EType.Array && aNumDimensions != 0 && aSubType !is null &&
				canFind([ EType.Primitive, EType.Record ], aSubType.Type)) ||
				(aType == EType.ArrayReference && aSubType !is null) ||
				(aType == EType.Record && aRecordName != "" && aRecordSize >= 1 &&
					aRecordFieldTypes.length >= 1));
	}
	body
	{
		Type = aType;
		mPrimitiveType = aPrimitiveType;
		mNumDimensions = aNumDimensions;
		mSubType = aSubType;
		mRecordName = aRecordName;
		mRecordSize = aRecordSize;
		mRecordFieldTypes = aRecordFieldTypes;
	}



	public EPrimitiveType PrimitiveType() const
	in
	{
		assert(Type == EType.Primitive);
	}
	body
	{
		return mPrimitiveType;
	}



	public string RecordName() const
	in
	{
		assert(Type == EType.Record);
	}
	body
	{
		return mRecordName;
	}



	public int RecordSize() const
	in
	{
		assert(Type == EType.Record);
	}
	body
	{
		return mRecordSize;
	}



	public CType GetRecordFieldType(int aOffset)
	{
		return GetRecordFieldTypeImpl(aOffset);
	}

	public const(CType) GetRecordFieldType(int aOffset) const
	{
		return (cast(CType)this).GetRecordFieldTypeImpl(aOffset);
	}

	private CType GetRecordFieldTypeImpl(int aOffset)
	in
	{
		assert(Type == EType.Record && aOffset >= 0 && aOffset < mRecordSize);
	}
	out(result)
	{
		assert(result.Type == EType.Primitive);
	}
	body
	{
		auto Type = this;
		int Index = 0;
		int Offset = 0;
		while (true)
		{
			CType FieldType = Type.mRecordFieldTypes[Index];
			if (Offset == aOffset && FieldType.Type == EType.Primitive)
				return FieldType;
			assert(Offset <= aOffset);
			final switch (FieldType.Type)
			{
				case EType.Array:
				case EType.ArrayReference:
					assert(false);
				case EType.Primitive:
					Index++;
					Offset++;
					break;
				case EType.Record:
					if (Offset + FieldType.RecordSize <= aOffset)
					{
						Index++;
						Offset += FieldType.RecordSize;
					}
					else
					{
						Type = FieldType;
						Index = 0;
					}
					break;
			}
		}
	}



	public int NumDimensions() const
	in
	{
		assert(Type == EType.Array);
	}
	body
	{
		return mNumDimensions;
	}



	public CType SubType()
	in
	{
		assert(canFind([ EType.Array, EType.ArrayReference ], Type));
	}
	body
	{
		return mSubType;
	}

	public const(CType) SubType() const
	in
	{
		assert(canFind([ EType.Array, EType.ArrayReference ], Type));
	}
	body
	{
		return mSubType;
	}



	public bool Equals(const(CType) aType) const
	{
		if (Type != aType.Type)
			return false;
		final switch (Type)
		{
			case EType.Array:
				return mNumDimensions == aType.mNumDimensions && mSubType.Equals(aType.mSubType);
			case EType.ArrayReference:
				return mSubType.Equals(aType.mSubType);
			case EType.Primitive:
				return mPrimitiveType == aType.mPrimitiveType;
			case EType.Record:
				return mRecordName == aType.mRecordName;
		}
	}



	public int ByteSize()
	{
		assert(Type == EType.Primitive || Type == EType.Record);
		if (Type == EType.Record)
		{
			int Size;
			foreach (CType FieldType; mRecordFieldTypes)
				Size += FieldType.ByteSize();
			return Size;
		}
		return PrimitiveTypeByteSize(PrimitiveType);
	}
}