﻿module QBasic.System.CFileSystem;
import std.stream;
import QBasic.IFileSystemExternal, QBasic.Interpreter.IInterpreterState,
		QBasic.Program.EFileOpenMode, QBasic.System.IFileSystem;

class CFileInfo
{
	public EFileOpenMode mMode;
	public Stream mDataStream;

	public this(EFileOpenMode aMode, Stream aDataStream)
	{
		mMode = aMode;
		mDataStream = aDataStream;
	}

	public EFileOpenMode Mode()
	{
		return mMode;
	}

	public Stream DataStream()
	{
		return mDataStream;
	}
}

class CFileSystem : IFileSystem
{
	private IFileSystemExternal mExternal;
	private IInterpreterState mInterpreterState;
	private CFileInfo[int] mFiles;

	public this(IFileSystemExternal aExternal, IInterpreterState aInterpreterState)
	{
		mExternal = aExternal;
		mInterpreterState = aInterpreterState;
	}



	public bool Open(int aNumber, EFileOpenMode aMode, string aPath)
	{
		if (aNumber in mFiles)
		{
			mInterpreterState.SetError("File number already in use.");
			return false;
		}
		auto DataStream = mExternal.Open(aPath);
		if (DataStream is null)
		{
			mInterpreterState.SetError("Failed to open file.");
			return false;
		}
		mFiles[aNumber] = new CFileInfo(aMode, DataStream);
		return true;
	}



	public bool Close(int aNumber)
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		mFiles.remove(aNumber);
		mExternal.Close(File.DataStream);
		return true;
	}



	public int FreeFile()
	{
		for (int i = 1; ; i++)
		{
			if (!(i in mFiles))
				return i;
		}
	}



	public bool ReadLine(int aNumber, out ubyte[] aLine)
	{
		auto File = aNumber in mFiles;
		if (File is null || File.Mode != EFileOpenMode.Input || File.DataStream.eof)
		{
			mInterpreterState.SetError("File not open for input.");
			return false;
		}
		while (!File.DataStream.eof)
		{
			ubyte b;
			File.DataStream.read(b);
			if (b == 0xa)
				break;
			if (b == 0xd)
			{
				if (!File.DataStream.eof)
				{
					File.DataStream.read(b);
					if (b != 0xa)
						File.DataStream.seek(-1, SeekPos.Current);
				}
				break;
			}
			aLine ~= b;
		}
		return true;
	}



	public bool EOF(int aNumber, out bool aIsEOF)
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		aIsEOF = File.DataStream.eof;
		return true;
	}



	public bool LOF(int aNumber, out int aLOF)
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		auto Size = File.DataStream.size;
		if (Size >= int.max)
		{
			mInterpreterState.SetError("File too large.");
			return false;
		}
		aLOF = cast(int)Size;
		return true;
	}



	public bool Seek(int aNumber, int aPosition)
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		File.DataStream.seek(aPosition - 1, SeekPos.Set);
		return true;
	}



	public bool BLoad(string aPath, out ubyte[] aData)
	{
		auto DataStream = mExternal.Open(aPath);
		if (DataStream is null)
		{
			mInterpreterState.SetError("Failed to open file.");
			return false;
		}
		enum int HeaderSize = 7;
		ulong Size = DataStream.size;
		if (Size <= HeaderSize)
		{
			mInterpreterState.SetError("File too small.");
			return false;
		}
		if (Size > 1000000)
		{
			mInterpreterState.SetError("File too large.");
			return false;
		}
		aData = new ubyte[cast(int)Size - HeaderSize];
		DataStream.seek(HeaderSize, SeekPos.Set);   // skip header
		DataStream.readExact(aData.ptr, aData.length);
		mExternal.Close(DataStream);
		return true;
	}



	public bool Get(int aNumber, int aSize, out ubyte[] aData)
	in
	{
		assert(aSize > 0);
	}
	body
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		aData = new ubyte[aSize];
		if (File.DataStream.read(aData) != aSize)
		{
			mInterpreterState.SetError("End-of-file reached.");
			return false;
		}
		return true;
	}



	public bool Put(int aNumber, immutable(ubyte[]) aData)
	in
	{
		assert(aData.length > 0);
	}
	body
	{
		auto File = aNumber in mFiles;
		if (File is null)
		{
			mInterpreterState.SetError("File number not open.");
			return false;
		}
		if (File.DataStream.write(aData) != aData.length)
		{
			mInterpreterState.SetError("Failed to write data to file.");
			return false;
		}
		return true;
	}
}