﻿module QBasic.System.CGraphics;
import std.algorithm;
import QBasic.IGraphicsExternal, QBasic.Interpreter.CReference, QBasic.Interpreter.IInterpreterState,
		QBasic.Program.EGraphicsOperation, QBasic.Program.ELineType, QBasic.System.IGraphics;

class CGraphics : IGraphics
{
	private IGraphicsExternal mGraphicsExternal;
	private IInterpreterState mInterpreterState;
	private int mCurrentColourIndex = 0;
	private int mCurrentComponent = 0;
	private int mR, mG, mB;
	private enum int eScreenXSize = 320;
	private enum int eScreenYSize = 192;
	private ubyte[] mScreenData;

	invariant
	{
		assert(mCurrentColourIndex >= 0 && mCurrentColourIndex <= 256 &&
				mCurrentComponent >= 0 && mCurrentComponent < 3);
	}

	public this(IGraphicsExternal aGraphicsExternal, IInterpreterState aInterpreterState)
	{
		mGraphicsExternal = aGraphicsExternal;
		mInterpreterState = aInterpreterState;
		mScreenData = new ubyte[eScreenXSize * eScreenYSize];
	}



	public const(ubyte[]) ScreenData()
	{
		return mScreenData;
	}



	public int ScreenXSize()
	{
		return eScreenXSize;
	}



	public int ScreenYSize()
	{
		return eScreenYSize;
	}



	public void SetCurrentColourIndex(int aIndex)
	in
	{
		assert(aIndex >= 0 && aIndex < 256);
	}
	body
	{
		mCurrentColourIndex = aIndex;
		mCurrentComponent = 0;
	}



	public bool SetCurrentColourElement(int aValue)
	{
		if (aValue < 0 || aValue > 63)
			return false;
		switch (mCurrentComponent)
		{
			case 0:
				mR = aValue;
				break;
			case 1:
				mG = aValue;
				break;
			case 2:
				mB = aValue;
				break;
			default:
				assert(false);
		}
		if (++mCurrentComponent == 3)
		{
			mGraphicsExternal.SetColour(mCurrentColourIndex, mR, mG, mB);
			mCurrentColourIndex++;
			mCurrentComponent = 0;
		}
		return true;
	}



	public bool Line(int aX0, int aY0, int aX1, int aY1, int aColour, ELineType aType)
	{
		if (aColour < 0 || aColour >= 255)
		{
			mInterpreterState.SetError("Invalid colour value");
			return false;
		}
		final switch (aType)
		{
			case ELineType.Line:
				return DrawLine(aX0, aY0, aX1, aY1, cast(ubyte)aColour);
			case ELineType.Rectangle:
				return DrawRectangle(aX0, aY0, aX1, aY1, cast(ubyte)aColour);
			case ELineType.FilledRectangle:
				return DrawFilledRectangle(aX0, aY0, aX1, aY1, cast(ubyte)aColour);
		}
	}



	private bool DrawLine(int aX0, int aY0, int aX1, int aY1, ubyte aColour)
	{
		if (aX0 != aX1 && aY0 != aY1)
		{
			mInterpreterState.SetError("Only horizontal and vertical lines are supported");
			return false;
		}
		if (aY0 == aY1)
		{
			int X0 = min(aX0, aX1);
			int X1 = max(aX0, aX1);
			if (X0 < 0 || X1 >= eScreenXSize || aY0 < 0 || aY0 >= eScreenYSize)
			{
				mInterpreterState.SetError("Invalid LINE coordinates");
				return false;
			}
			for (int x = X0; x <= X1; x++)
				mScreenData[x + aY0 * eScreenXSize] = aColour;
		}
		else
		{
			int Y0 = min(aY0, aY1);
			int Y1 = max(aY0, aY1);
			if (aX0 < 0 || aX1 >= eScreenXSize || Y0 < 0 || Y0 >= eScreenYSize)
			{
				mInterpreterState.SetError("Invalid LINE coordinates");
				return false;
			}
			for (int y = Y0; y <= Y1; y++)
				mScreenData[aX0 + y * eScreenXSize] = aColour;
		}
		return true;
	}



	private bool DrawRectangle(int aX0, int aY0, int aX1, int aY1, ubyte aColour)
	{
		int X0 = min(aX0, aX1);
		int X1 = max(aX0, aX1);
		int Y0 = min(aY0, aY1);
		int Y1 = max(aY0, aY1);
		if (X0 == X1)
			return DrawLine(X0, Y0, X0, Y1, aColour);
		if (!DrawLine(X0, Y0, X1, Y0, aColour))
			return false;
		if (Y0 == Y1)
			return true;
		if (!DrawLine(X0, Y1, X1, Y1, aColour))
			return false;
		if (Y1 - Y0 == 1)
			return true;
		if (!DrawLine(X0, Y0 + 1, X0, Y1 - 1, aColour))
			return false;
		return DrawLine(X1, Y0 + 1, X1, Y1 - 1, aColour);
	}



	private bool DrawFilledRectangle(int aX0, int aY0, int aX1, int aY1, ubyte aColour)
	{
		int X0 = min(aX0, aX1);
		int X1 = max(aX0, aX1);
		int Y0 = min(aY0, aY1);
		int Y1 = max(aY0, aY1);
		if (X0 < 0 || X1 >= eScreenXSize || Y0 < 0 || Y1 >= eScreenYSize)
		{
			mInterpreterState.SetError("Invalid LINE coordinates");
			return false;
		}
		for (int y = Y0; y <= Y1; y++)
		{
			for (int x = X0; x <= X1; x++)
				mScreenData[x + y * eScreenXSize] = aColour;
		}
		return true;
	}



	public bool Put(int aX, int aY, CReference aImage, EGraphicsOperation aOperation)
	{
		auto SizeData = mInterpreterState.ReadData(aImage, 4);
		if (SizeData is null)
			return false;
		int XSize = SizeData[0] | (cast(int)SizeData[1] << 8);
		if (XSize % 8 != 0)
		{
			mInterpreterState.SetError("Invalid image header.");
			return false;
		}
		XSize /= 8;
		int YSize = SizeData[2] | (cast(int)SizeData[3] << 8);
		if (aX < 0 || aX + XSize > eScreenXSize || aY < 0 || aY + YSize > eScreenYSize)
		{
			mInterpreterState.SetError("Invalid PUT coordinates.");
			return false;
		}
		auto Data = mInterpreterState.ReadData(aImage, 4 + XSize * YSize);
		if (Data is null)
			return false;
		int ScreenOffset = aY * eScreenXSize + aX;
		int ImageOffset = 4;
		for (int y = 0; y < YSize; y++)
		{
			for (int x = 0; x < XSize; x++)
			{
				ubyte Old = mScreenData[ScreenOffset];
				ubyte Datum = Data[ImageOffset++];
				ubyte New;
				final switch (aOperation)
				{
					case EGraphicsOperation.And:
						New = Old & Datum;
						break;
					case EGraphicsOperation.Or:
						New = Old | Datum;
						break;
					case EGraphicsOperation.PSet:
						New = Datum;
						break;
				}
				mScreenData[ScreenOffset++] = New;
			}
			ScreenOffset += eScreenXSize - XSize;
		}
		return true;
	}
}