﻿module Core.CCoreModule;
import std.algorithm, std.datetime, std.file, std.path, std.range, std.stream, std.typecons;
import Tractor.Core;
import Engine.CEngine, Engine.CTractorException, Engine.IDisplay, Engine.IFileSystem,
		Engine.ILogger;

class CCoreModule : ICoreModule
{
	private IDisplay mDisplay;
	private CEngine mEngine;
	private IFileSystem mFileSystem;
	private ILogger mLogger;

	public this(IDisplay aDisplay, CEngine aEngine, IFileSystem aFileSystem, ILogger aLogger)
	{
		mDisplay = aDisplay;
		mEngine = aEngine;
		mFileSystem = aFileSystem;
		mLogger = aLogger;
	}



	public Stream OpenDataFile(string aFileName)
	{
		File f;
		try
		{
			f = new File(mFileSystem.GetDataFilePath(aFileName), FileMode.In);
		}
		catch (OpenException oe)
		{
			mLogger.Log("Data file not found: " ~ aFileName);
			throw new CTractorException();
		}
		return f;
	}



	public Stream OpenLocalFile(string aFileName)
	{
		File f;
		try
		{
			f = new File(mFileSystem.GetLocalFilePath(aFileName), FileMode.In | FileMode.Out);
		}
		catch (OpenException oe)
		{
			mLogger.Log("Unable to open local file: " ~ aFileName);
			throw new CTractorException();
		}
		return f;
	}



	public void CloseFile(Stream aStream)
	{
		(cast(File)aStream).close();
	}



	public DirEntry[string] GetLocalFiles()
	{
		return dirEntries(mFileSystem.GetLocalDirectoryPath(), SpanMode.shallow, false)
				.filter!(e => e.isFile)
				.map!(f => tuple(baseName(f.name), f))
				.assocArray();
	}



	public void Log(string aMessage)
	{
		mLogger.Log(aMessage);
	}



	public void Redraw()
	{
		mDisplay.Redraw();
	}



	public void StartTimer()
	{
		mEngine.StartTimer();
	}



	public ITimer CreateTimer(void delegate() aCallback)
	{
		return new CTimer(mEngine, aCallback);
	}



	public string FormatDateTime(SysTime aTime)
	{
		return mEngine.FormatDateTime(aTime);
	}



	public void Quit()
	{
		mEngine.Quit();
	}
}



interface ITimerFactory
{
	void RegisterTimer(CTimer aTimer);
	void TimerStart(CTimer aTimer, int aPeriod);
	void TimerStop(CTimer aTimer);
}



class CTimer : ITimer
{
	private ITimerFactory mTimerFactory;
	private void delegate() mCallback;
	private bool mStarted = false;

	public this(ITimerFactory aTimerFactory, void delegate() aCallback)
	{
		mTimerFactory = aTimerFactory;
		mCallback = aCallback;
		mTimerFactory.RegisterTimer(this);
	}



	public void Start(int aPeriod)
	in
	{
		assert(!mStarted);
	}
	body
	{
		mStarted = true;
		mTimerFactory.TimerStart(this, aPeriod);
	}



	public void Stop()
	in
	{
		assert(mStarted);
	}
	body
	{
		mStarted = false;
		mTimerFactory.TimerStop(this);
	}



	public bool Started()
	{
		return mStarted;
	}



	public void OnTick()
	{
		mCallback();
	}
}