﻿module Engine.CConsoleWindow;
import std.conv, std.datetime, std.math, std.string, std.utf;
import win32.windows;
import Engine.CEngine, Engine.ILogger;

struct SDrawLine
{
	public this(int aLine, int aStart, int aLength)
	{
		Line = aLine;
		Start = aStart;
		Length = aLength;
	}

	int Line;
	int Start;
	int Length;
}

class CConsoleWindow : ILogger
{
	private CEngine mEngine;
	private HWND mWindow;
	private HFONT mFont;
	private int mFontHeight;
	private Date mLastDate = Date.min();
	private string[] mLines;
	private SDrawLine[] mDrawLines;
	private bool mRewrapLines = false;
	private int mViewLine;
	private int mMaxLines;
	private int mMouseWheelDelta = 0;

	public this(CEngine aEngine)
	{
		mEngine = aEngine;
	}



	public void Create()
	{
		WNDCLASSEX ClassInfo;
		ClassInfo.cbSize = WNDCLASSEX.sizeof;
		ClassInfo.style = CS_OWNDC;
		ClassInfo.lpfnWndProc = &ConsoleWndProc;
		ClassInfo.hInstance = mEngine.Instance();
		ClassInfo.hIcon = cast(HICON)LoadImage(mEngine.Instance, MAKEINTRESOURCE(100),
				IMAGE_ICON, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
		ClassInfo.hCursor = LoadCursor(null, IDC_ARROW);
		assert(ClassInfo.hCursor != NULL);
		ClassInfo.hbrBackground = cast(HBRUSH)GetStockObject(WHITE_BRUSH);
		ClassInfo.lpszClassName = cast(wchar*)"TractorConsoleWindow"w;
		ATOM RCResult = RegisterClassEx(&ClassInfo);
		assert(RCResult != 0);

		int X;
		int Y;
		int XSize;
		int YSize;
		DWORD Style;
		RECT r;
		bool SRResult = 0 != SetRect(&r, 0, 0, 640, 480);
		assert(SRResult);
		Style = WS_OVERLAPPEDWINDOW | WS_VSCROLL;
		bool AWRResult = 0 != AdjustWindowRect(&r, Style, false);
		assert(AWRResult);
		XSize = r.right - r.left;
		X = (GetSystemMetrics(SM_CXSCREEN) - XSize) / 2;
		YSize = r.bottom - r.top;
		Y = (GetSystemMetrics(SM_CYSCREEN) - r.bottom) / 2;
		mWindow = CreateWindowEx(WS_EX_APPWINDOW, toUTF16z("TractorConsoleWindow"),
				toUTF16z(mEngine.Title() ~ " \u2013 Console"), Style, X, Y, XSize, YSize, null, null, null,
				cast(LPVOID)this);
		assert(mWindow != NULL && GetLastError() == 0);

		HDC DC = GetDC(mWindow);
		int Height = -MulDiv(12, GetDeviceCaps(DC, LOGPIXELSY), 72);
		mFont = CreateFont(Height, 0, 0, 0, FW_NORMAL, false, false, false, ANSI_CHARSET,
				OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH |
				FF_DONTCARE, toUTF16z("Calibri"));
		assert(mFont !is null);
		HGDIOBJ LastFont = SelectObject(DC, cast(HGDIOBJ)mFont);
		TEXTMETRIC TextMetrics;
		bool GTMResult = 0 != GetTextMetrics(DC, &TextMetrics);
		assert(GTMResult);
		mFontHeight = TextMetrics.tmHeight;
		SelectObject(DC, LastFont);
	}



	private void OnWindowDestroyed()
	{
		DeleteObject(cast(HGDIOBJ)mFont);
	}



	public void Show()
	{
		bool SWPResult = 0 != SetWindowPos(mWindow, HWND_BOTTOM, 0, 0, 0, 0,
				SWP_SHOWWINDOW | SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
		assert(SWPResult);
		Redraw(true);
	}



	public void Close()
	{
		//DestroyWindow(mWindow);
	}



	public void BringToFront()
	{
		bool SWPResult = 0 != SetWindowPos(mWindow, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		assert(SWPResult);
	}



	public void Log(string aMessage)
	{
		auto Time = cast(DateTime)Clock.currTime();
		// TODO: use Windows formatting instead of hardcoding formats
		if (Time.date() != mLastDate)
		{
			mLines ~= format("%02d/%02d/%02d", Time.day(), Time.month(), Time.year());
			mLastDate = Time.date();
		}
		bool First = true;
		int StartIndex = 0;
		while (true)
		{
			auto Index = aMessage.indexOf('\n', StartIndex);
			bool Finished = Index == -1;
			if (Index == -1)
				Index = aMessage.length;
			string Line = aMessage[StartIndex .. Index];
			StartIndex = Index + 1;
			if (First)
			{
				First = false;
				mLines ~= format("[%02d:%02d:%02d] %s", Time.hour(), Time.minute(),
						Time.second(), Line);
			}
			else
				mLines ~= Line;
			if (Finished)
				break;
		}
		Redraw(true);
	}



	private void Redraw(bool aRewrapLines)
	{
		bool IRResult = 0 != InvalidateRect(mWindow, null, false);
		assert(IRResult);
		mRewrapLines = aRewrapLines;
	}



	public bool OnPaint()
	{
		PAINTSTRUCT Paint;
		HDC DC = BeginPaint(mWindow, &Paint);
		assert(DC !is null);
		RECT ClientRect;
		GetClientRect(mWindow, &ClientRect);
		int ClientXSize = ClientRect.right - ClientRect.left;
		int ClientYSize = ClientRect.bottom - ClientRect.top;
		HGDIOBJ LastFont = SelectObject(DC, cast(HGDIOBJ)mFont);
		if (mRewrapLines)
		{
			mRewrapLines = false;
			RewrapLines(DC, ClientXSize);
			mMaxLines = (ClientRect.bottom - ClientRect.top) / mFontHeight;
			mViewLine = std.algorithm.max(cast(int)mDrawLines.length - mMaxLines, 0);
			UpdateScroll();
		}
		SetTextColor(DC, RGB(0, 0, 0));
		SetBkColor(DC, RGB(255, 255, 255));
		int YOffset = mDrawLines.length <= mMaxLines ? 0 : ClientYSize % mFontHeight;
		RECT Rect;
		if (YOffset != 0)
		{
			SetRect(&Rect, 0, 0, ClientXSize, YOffset);
			FillRect(DC, &Rect, cast(HBRUSH)GetStockObject(WHITE_BRUSH));
		}
		for (int i = 0; i < mMaxLines; i++)
		{
			if (mViewLine + i == mDrawLines.length)
				break;
			SetRect(&Rect, 0, i * mFontHeight + YOffset, ClientXSize, (i + 1) * mFontHeight + YOffset);
			FillRect(DC, &Rect, cast(HBRUSH)GetStockObject(WHITE_BRUSH));
			auto Line = &mDrawLines[mViewLine + i];
			string Text = mLines[Line.Line][Line.Start .. Line.Start + Line.Length];
			int DTResult = DrawText(DC, toUTF16z(Text), -1, &Rect, DT_LEFT | DT_TOP);
			assert(DTResult != 0);
		}
		SelectObject(DC, LastFont);
		EndPaint(mWindow, &Paint);
		return true;
	}



	private void RewrapLines(HDC aDC, int aClientXSize)
	{
		mDrawLines.length = 0;
		RECT Rect;
		SetRect(&Rect, 0, 0, aClientXSize, mFontHeight);
		SIZE Size;
		for (int i = 0; i < mLines.length; i++)
		{
			int StartIndex = 0;
			while (true)
			{
				wstring Text = to!wstring(mLines[i][StartIndex .. $]);
				bool GTEPResult = 0 != GetTextExtentPoint32(aDC, Text.ptr, Text.length, &Size);
				assert(GTEPResult);
				if (Size.cx < aClientXSize)
					break;
				int BreakIndex = -1;
				for (int j = StartIndex + 2; j <= mLines[i].length; j++)
				{
					Text = to!wstring(mLines[i][StartIndex .. j]);
					GTEPResult = 0 != GetTextExtentPoint32(aDC, Text.ptr, Text.length, &Size);
					assert(GTEPResult);
					if (Size.cx > aClientXSize)
					{
						BreakIndex = j - 1;
						break;
					}
				}
				assert(BreakIndex != -1 && BreakIndex != StartIndex);
				mDrawLines ~= SDrawLine(i, StartIndex, BreakIndex - StartIndex);
				StartIndex = BreakIndex;
			}
			mDrawLines ~= SDrawLine(i, StartIndex, mLines[i].length - StartIndex);
		}
	}



	private void UpdateScroll()
	{
		SCROLLINFO ScrollInfo;
		ScrollInfo.cbSize = SCROLLINFO.sizeof;
		ScrollInfo.fMask = SIF_DISABLENOSCROLL | SIF_PAGE | SIF_POS | SIF_RANGE;
		ScrollInfo.nMin = 0;
		ScrollInfo.nMax = mDrawLines.length - 1;
		ScrollInfo.nPage = mMaxLines;
		ScrollInfo.nPos = mViewLine;
		SetScrollInfo(mWindow, SB_VERT, &ScrollInfo, true);
	}



	public bool OnSize()
	{
		Redraw(true);
		return true;
	}



	public void ScrollBy(int aDelta)
	{
		if (mViewLine + aDelta < 0)
			aDelta = -mViewLine;
		else if (mViewLine + aDelta > mDrawLines.length - mMaxLines)
			aDelta = mDrawLines.length - mMaxLines - mViewLine;
		if (aDelta == 0)
			return;
		mViewLine += aDelta;
		Redraw(false);
		UpdateScroll();
	}



	public bool OnVScroll(int aScrollType, int aPos)
	{
		int OldViewLine = mViewLine;
		switch (aScrollType)
		{
			case SB_TOP:
				ScrollBy(-mViewLine);
				return true;
			case SB_BOTTOM:
				ScrollBy(mDrawLines.length - mMaxLines - mViewLine);
				return true;
			case SB_LINEUP:
				ScrollBy(-1);
				return true;
			case SB_LINEDOWN:
				ScrollBy(1);
				return true;
			case SB_PAGEUP:
				ScrollBy(-mMaxLines);
				return true;
			case SB_PAGEDOWN:
				ScrollBy(mMaxLines);
				return true;
			case SB_THUMBTRACK:
			case SB_THUMBPOSITION:
				ScrollBy(aPos - mViewLine);
				return true;
			default:
				return false;
		}
	}



	public bool OnMouseWheel(int aDelta)
	{
		mMouseWheelDelta += abs(aDelta);
		if (mMouseWheelDelta >= WHEEL_DELTA)
		{
			mMouseWheelDelta %= WHEEL_DELTA;
			int ScrollLines;
			bool SPIResult = 0 != SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &ScrollLines, 0);
			assert(SPIResult);
			ScrollBy(ScrollLines * -sgn(aDelta));
		}
		return true;
	}
}



extern (Windows) LRESULT ConsoleWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_CREATE:
			CREATESTRUCT *cs = cast(CREATESTRUCT*)lParam;
			SetLastError(0);
			LONG_PTR Result = SetWindowLongPtrW(hwnd, GWLP_USERDATA, cast(LONG_PTR)cs.lpCreateParams);
			if (Result == 0 && GetLastError() != 0)
				throw new Exception("SetWindowLongPtr() failed");
			return 0;
		case WM_DESTROY:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CConsoleWindow Window = cast(CConsoleWindow)(cast(LPVOID)Pointer);
			Window.OnWindowDestroyed();
			return 0;
		case WM_CLOSE:
			return 0;
		case WM_PAINT:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CConsoleWindow Window = cast(CConsoleWindow)(cast(LPVOID)Pointer);
			if (Window.OnPaint())
				return 0;
			break;
		case WM_SIZE:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CConsoleWindow Window = cast(CConsoleWindow)(cast(LPVOID)Pointer);
			if (Window.OnSize())
				return 0;
			break;
		case WM_VSCROLL:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CConsoleWindow Window = cast(CConsoleWindow)(cast(LPVOID)Pointer);
			if (Window.OnVScroll(LOWORD(wParam), HIWORD(wParam)))
				return 0;
			break;
		case WM_MOUSEWHEEL:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CConsoleWindow Window = cast(CConsoleWindow)(cast(LPVOID)Pointer);
			if (Window.OnMouseWheel(GET_WHEEL_DELTA_WPARAM(wParam)))
				return 0;
			break;
		default:
			break;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}