module Engine.CEngine;
import std.algorithm, std.conv, std.datetime, std.string, std.utf;
import win32.commctrl, win32.windows;
import DWin32.avrt;
import derelict.opengl3.gl3;
import glwh.glwh;
import Core.CCoreModule, Core.EKey;
import Engine.CConsoleWindow, Engine.CControllerWrapper, Engine.CFileSystem,
		Engine.CInputModule, Engine.CMainWindow, Engine.CTractor, Engine.ILogger;
import Fonts.CFontModule;
import Shaders.CShaderModule;
import Textures.CTextureModule;
import Tractor.Controller, Tractor.Engine;

class CEngine : IEngine, ILogger, ITimerFactory
{
	private HINSTANCE mInstance;
	private string mTitle;
	private CFileSystem mFileSystem;
	private CTractor mTractor;
	private CControllerWrapper mControllerWrapper;
	private CControllerOptions mControllerOptions;
	private CCoreModule mCoreModule;
	private CFontModule mFontModule;
	private CInputModule mInputModule;
	private CShaderModule mShaderModule;
	private CTextureModule mTextureModule;

	private HANDLE mAvTask = null;
	private HANDLE[] mWaitObjects;
	private CMainWindow mMainWindow;
	private CConsoleWindow mConsoleWindow;
	private HDC mDC;
	private HGLRC mRenderingContext = null;
	private HANDLE mWaitableTimer = null;
	private bool mTimerStarted = false;
	private CTimer[HANDLE] mTimers;
	private bool mShowCursor = true;

	public this(HINSTANCE aInstance, string aTitle, IController aController)
	{
		mInstance = aInstance;
		mTitle = aTitle;

		mFileSystem = new CFileSystem();
		mControllerWrapper = new CControllerWrapper(this, aController);
	}



	public HINSTANCE Instance()
	{
		return mInstance;
	}



	public string Title()
	{
		return mTitle;
	}



	public bool ShowCursor()
	{
		return mShowCursor;
	}



	public void Run()
	{
		if (Initialise())
		{
		MainLoop:
			while (true)
			{
				DWORD MWFMOResult = MsgWaitForMultipleObjects(mWaitObjects.length,
						mWaitObjects.ptr, false, INFINITE, QS_ALLINPUT);
				assert(MWFMOResult >= WAIT_OBJECT_0 && MWFMOResult <= WAIT_OBJECT_0 +
						mWaitObjects.length);
				if (MWFMOResult == WAIT_OBJECT_0)
					OnFrameTimer();
				else if (MWFMOResult < mWaitObjects.length)
					OnTimer(mWaitObjects[MWFMOResult - WAIT_OBJECT_0]);
				else
				{
					MSG Message;
					while (PeekMessage(&Message, null, 0, 0, PM_REMOVE))
					{
						if (Message.message == WM_QUIT)
							break MainLoop;
						TranslateMessage(&Message);
						DispatchMessage(&Message);
					}
				}
			}
		}
		Uninitialise();
	}



	private bool Initialise()
	{
		DWORD TaskIndex = 0;
		mAvTask = AvSetMmThreadCharacteristics(toUTF16z("Games"), &TaskIndex);
		assert(mAvTask != cast(HANDLE)0);

		mWaitableTimer = CreateWaitableTimer(null, false, null);
		assert(mWaitableTimer !is null);
		mWaitObjects ~= mWaitableTimer;

		INITCOMMONCONTROLSEX icc;
		bool ICCResult = 0 != InitCommonControlsEx(&icc);
		assert(ICCResult);

		mConsoleWindow = new CConsoleWindow(this);
		mConsoleWindow.Create();

		if (!mControllerWrapper.GetOptions(mControllerOptions))
		{
			ControllerError();
			return true;
		}

		if (mControllerOptions.LocalDirectoryName !is null)
			mFileSystem.SetLocalDirectoryName(mControllerOptions.LocalDirectoryName);
		mMainWindow = new CMainWindow(this);
		mMainWindow.Create(mControllerOptions.DisplayXSize, mControllerOptions.DisplayYSize);
		mConsoleWindow.Show();
		mDC = GetDC(mMainWindow.GetWindowHandle());
		mRenderingContext = cast(HGLRC)glwhCreateRenderingContextForDevice(GLVersion.GL31, mDC, 16);
		if (mRenderingContext == NULL)
		{
			mMainWindow.ShowError("Failed to initialise OpenGL 3.1");
			return false;
		}

		mCoreModule = new CCoreModule(mMainWindow, this, mFileSystem, this);
		mFontModule = new CFontModule(mFileSystem);
		mInputModule = new CInputModule(this);
		mMainWindow.SetInputHandler(mInputModule);
		mTextureModule = new CTextureModule(mFileSystem);
		mShaderModule = new CShaderModule(mMainWindow, mFileSystem, this);
		mTractor = new CTractor(mCoreModule, mFontModule, mShaderModule, mTextureModule);

		if (!mControllerWrapper.Initialise(mTractor))
		{
			ControllerError();
			return true;
		}

		return true;
	}



	private void Uninitialise()
	{
		assert(!mTimerStarted);

		mControllerWrapper.Uninitialise();
		mControllerWrapper = null;

		delete mCoreModule;
		delete mFontModule;
		delete mShaderModule;
		delete mTextureModule;

		if (mRenderingContext !is null)
			glwhDeleteRenderingContext(mRenderingContext);

		bool CHResult = 0 != CloseHandle(mWaitableTimer);
		assert(CHResult);
		foreach (const(HANDLE) Handle; mTimers.byKey)
		{
			CHResult = 0 != CloseHandle(cast(HANDLE)Handle);
			assert(CHResult);
		}

		bool RMTCResult = 0 != AvRevertMmThreadCharacteristics(mAvTask);
		assert(RMTCResult);
	}



	public void OnQuit()
	{
		if (mTimerStarted)
			StopTimer();
	}



	private void OnFrameTimer()
	{
		if (mControllerWrapper.IsError())
			return;
		if (!mControllerWrapper.OnFrame())
		{
			ControllerError();
			return;
		}
		SwapBuffers(mDC);
		assert(glGetError() == GL_NO_ERROR);
	}



	public void Quit()
	{
		mMainWindow.Close();
	}



	public void Log(string aMessage)
	{
		if (mConsoleWindow !is null)
			mConsoleWindow.Log(aMessage);
	}



	public void StartTimer()
	{
		assert(!mTimerStarted);
		LARGE_INTEGER DueTime;
		DueTime.QuadPart = -1;
		bool SWTResult = 0 != SetWaitableTimer(mWaitableTimer, &DueTime, 20, null, null, false);
		assert(SWTResult);
		mTimerStarted = true;
	}



	public void StopTimer()
	{
		assert(mTimerStarted);
		bool Result = 0 != CancelWaitableTimer(mWaitableTimer);
		assert(Result);
		mTimerStarted = false;
	}



	public void RegisterTimer(CTimer aTimer)
	{
		HANDLE Timer = CreateWaitableTimer(null, false, null);
		assert(Timer !is null);
		mTimers[Timer] = aTimer;
		mWaitObjects ~= Timer;
	}



	private void OnTimer(HANDLE aHandle)
	{
		if (mTimers[aHandle].Started)
			mTimers[aHandle].OnTick();
	}



	private HANDLE TimerHandle(CTimer aTimer)
	{
		return cast(HANDLE)mTimers.byKeyValue.find!(t => t.value == aTimer).front.key;
	}



	public void TimerStart(CTimer aTimer, int aPeriod)
	{
		LARGE_INTEGER DueTime;
		DueTime.QuadPart = aPeriod * -10000;
		bool SWTResult = 0 != SetWaitableTimer(TimerHandle(aTimer), &DueTime, aPeriod, null, null, false);
		assert(SWTResult);
	}



	public void TimerStop(CTimer aTimer)
	{
		bool CWTResult = 0 != CancelWaitableTimer(TimerHandle(aTimer));
		assert(CWTResult);
	}



	public string FormatDateTime(SysTime aTime)
	{
		auto SystemTime = SysTimeToSYSTEMTIME(aTime);
		return format("%s %s", FormatDate(SystemTime), FormatTime(SystemTime));
	}



	private string FormatDate(std.datetime.SYSTEMTIME aTime)
	{
		int GDFResult = GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE,
				cast(const(SYSTEMTIME*))&aTime, null, null, 0);
		assert(GDFResult != 0);
		wchar[] TimeStr = new wchar[GDFResult];
		GDFResult = GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE,
				cast(const(SYSTEMTIME*))&aTime, null, TimeStr.ptr, TimeStr.length);
		assert(GDFResult != 0);
		return to!string(TimeStr[0 .. GDFResult - 1]);
	}



	private string FormatTime(std.datetime.SYSTEMTIME aTime)
	{
		int GTFResult = GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS,
				cast(const(SYSTEMTIME*))&aTime, null, null, 0);
		assert(GTFResult != 0);
		wchar[] TimeStr = new wchar[GTFResult];
		GTFResult = GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS,
				cast(const(SYSTEMTIME*))&aTime, null, TimeStr.ptr, TimeStr.length);
		assert(GTFResult != 0);
		return to!string(TimeStr[0 .. GTFResult - 1]);
	}



	public bool OnPaint()
	{
		if (mControllerWrapper.IsError() || mTimerStarted)
			return false;
		if (!mControllerWrapper.OnFrame())
		{
			ControllerError();
			return true;
		}
		SwapBuffers(mDC);
		assert(glGetError() == GL_NO_ERROR);
		return true;
	}



	public bool OnKeyDown(EKey aKey)
	{
		if (!mControllerWrapper.OnKeyDown(aKey))
		{
			ControllerError();
			return false;
		}
		return true;
	}



	public bool OnKeyUp(EKey aKey)
	{
		if (!mControllerWrapper.OnKeyUp(aKey))
		{
			ControllerError();
			return false;
		}
		return true;
	}



	private void ControllerError()
	{
		if (mTimerStarted)
			StopTimer();
		mConsoleWindow.BringToFront();
	}
}