﻿module Engine.CMainWindow;
import std.utf;
import win32.windows;
import derelict.opengl3.gl3;
import Core.EKey;
import Engine.CEngine, Engine.CInputModule, Engine.IDisplay;

interface IInputHandler
{
	void OnKeyDown(uint aVirtualKey);
	void OnKeyUp(uint aVirtualKey);
}

class CMainWindow : IDisplay
{
	private CEngine mEngine;
	private IInputHandler mInputHandler = null;
	private HWND mWindow;
	private int mXSize = 0;
	private int mYSize;
	private bool mCursorVisible = true;
	private bool mCursorInWindow;

	public this(CEngine aEngine)
	{
		mEngine = aEngine;
	}



	public void SetInputHandler(IInputHandler aInputHandler)
	{
		mInputHandler = aInputHandler;
	}



	public HWND GetWindowHandle()
	{
		return mWindow;
	}



	public void Create(int aDisplayXSize, int aDisplayYSize)
	in
	{
		assert(aDisplayXSize >= 1 && aDisplayYSize >= 1);
	}
	body
	{
		mXSize = aDisplayXSize;
		mYSize = aDisplayYSize;

		WNDCLASSEX ClassInfo;
		ClassInfo.cbSize = WNDCLASSEX.sizeof;
		ClassInfo.style = CS_OWNDC;
		ClassInfo.lpfnWndProc = &MainWndProc;
		ClassInfo.hInstance = mEngine.Instance();
		ClassInfo.hIcon = cast(HICON)LoadImage(mEngine.Instance, MAKEINTRESOURCE(100),
				IMAGE_ICON, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
		ClassInfo.hCursor = LoadCursor(null, IDC_ARROW);
		assert(ClassInfo.hCursor != NULL);
		ClassInfo.hbrBackground = cast(HBRUSH)GetStockObject(BLACK_BRUSH);
		ClassInfo.lpszClassName = cast(wchar*)"TractorMainWindow"w;
		ATOM RCResult = RegisterClassEx(&ClassInfo);
		assert(RCResult != 0);

		RECT r;
		bool SRResult = 0 != SetRect(&r, 0, 0, mXSize, mYSize);
		assert(SRResult);
		DWORD Style = WS_MINIMIZEBOX | WS_POPUPWINDOW | WS_CAPTION;
		bool AWRResult = 0 != AdjustWindowRect(&r, Style, false);
		assert(AWRResult);
		int WindowXSize = r.right - r.left;
		int WindowYSize = r.bottom - r.top;
		int X = (GetSystemMetrics(SM_CXSCREEN) - WindowXSize) / 2;
		int Y = (GetSystemMetrics(SM_CYSCREEN) - WindowYSize) / 2;
		mWindow = CreateWindowEx(WS_EX_APPWINDOW, toUTF16z("TractorMainWindow"),
				toUTF16z(mEngine.Title()), Style | WS_VISIBLE, X, Y, WindowXSize, WindowYSize,
				null, null, null, cast(LPVOID)this);
		assert(mWindow != NULL && GetLastError() == 0);

		bool SWPResult = 0 != SetWindowPos(mWindow, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		assert(SWPResult);
	}



	public void GetSize(out int aDisplayXSize, out int aDisplayYSize)
	in
	{
		assert(mXSize != 0);
	}
	body
	{
		aDisplayXSize = mXSize;
		aDisplayYSize = mYSize;
	}



	public void Close()
	{
		DestroyWindow(mWindow);
	}



	public void ShowError(string aMessage)
	{
		MessageBox(mWindow, toUTF16z(aMessage), toUTF16z(mEngine.Title), MB_OK | MB_ICONERROR);
	}



	public void Redraw()
	{
		bool Result = 0 != InvalidateRect(mWindow, null, false);
		assert(Result);
	}



	public void OnWindowDestroyed()
	{
		mEngine.OnQuit();
		PostQuitMessage(0);
	}



	private void OnMouseMove(int aX, int aY)
	{
		RECT Client;
		bool GCRResult = 0 != GetClientRect(mWindow, &Client);
		assert(GCRResult);
		mCursorInWindow = aX >= 0 && aX < Client.right - Client.left && aY >= 0 && aY < Client.bottom - Client.top;
		if (!mCursorVisible)
			SetCursorVisible(true);
	}



	public bool OnSetCursor()
	{
		if (mEngine.ShowCursor())
			return false;
		int x, y;
		GetCursorPos(x, y);
		RECT Client;
		bool GCRResult = 0 != GetClientRect(mWindow, &Client);
		assert(GCRResult);
		if (x >= 0 && x < Client.right - Client.left && y >= 0 && y < Client.bottom - Client.top)
		{
			SetCursorVisible(false);
			return true;
		}
		return false;
	}



	public void GetCursorPos(out int aCursorX, out int aCursorY)
	{
		POINT Cursor;
		bool GCPResult = 0 != win32.winuser.GetCursorPos(&Cursor);
		assert(GCPResult);
		bool STCResult = 0 != ScreenToClient(mWindow, &Cursor);
		assert(STCResult);
		aCursorX = Cursor.x;
		aCursorY = Cursor.y;
	}



	private void SetCursorVisible(bool aVisible)
	{
		if (aVisible == mCursorVisible)
			return;
		mCursorVisible = aVisible;
		ShowCursor(aVisible);
	}



	public bool OnPaint()
	{
		if (mEngine.OnPaint())
		{
			ValidateRect(mWindow, cast(RECT*)null);
			return true;
		}
		return false;
	}



	private void OnKeyDown(uint aVirtualKey)
	{
		if (aVirtualKey == VK_F12)
		{
			TakeScreenshot();
			return;
		}
		if (mCursorInWindow)
			SetCursorVisible(false);
		if (mInputHandler !is null)
			mInputHandler.OnKeyDown(aVirtualKey);
	}



	private void TakeScreenshot()
	in
	{
		GLuint ReadFramebuffer;
		glGetIntegerv(GL_READ_FRAMEBUFFER_BINDING, cast(GLint*)&ReadFramebuffer);
		assert(ReadFramebuffer == 0);
	}
	body
	{
		glReadBuffer(GL_FRONT);
		ubyte[] Data = new ubyte[mXSize * mYSize * 3];
		glReadPixels(0, 0, mXSize, mYSize, GL_BGR, GL_UNSIGNED_BYTE, Data.ptr);
		assert(glGetError() == GL_NO_ERROR);
		bool CBResult = 0 != OpenClipboard(mWindow);
		assert(CBResult);
		CBResult = 0 != EmptyClipboard();
		assert(CBResult);
		int Pitch = mXSize * 3;
		if (Pitch % 4 != 0)
			Pitch += 4 - Pitch % 4;
		HGLOBAL CopyHandle = GlobalAlloc(GMEM_MOVEABLE, BITMAPINFOHEADER.sizeof + Pitch * mYSize);
		assert(CopyHandle !is null);
		ubyte* CopyData = cast(ubyte*)GlobalLock(CopyHandle);
		assert(CopyData !is null);
		BITMAPINFOHEADER* bih = cast(BITMAPINFOHEADER*)CopyData;
		bih.biSize = BITMAPINFOHEADER.sizeof;
		bih.biWidth = mXSize;
		bih.biHeight = mYSize;
		bih.biPlanes = 1;
		bih.biBitCount = 24;
		bih.biCompression = BI_RGB;
		bih.biSizeImage = Data.length;
		// TODO: use the display resolution
		bih.biXPelsPerMeter = 0;
		bih.biYPelsPerMeter = 0;
		bih.biClrUsed = 0;
		bih.biClrImportant = 0;
		CopyData[BITMAPINFOHEADER.sizeof .. Data.length + BITMAPINFOHEADER.sizeof] = Data[];
		GlobalUnlock(CopyHandle);
		HANDLE SCDResult = SetClipboardData(CF_DIB, CopyHandle);
		assert(SCDResult !is null);
		CBResult = 0 != CloseClipboard();
		assert(CBResult);
	}



	private void OnKeyUp(uint aVirtualKey)
	{
		if (mInputHandler !is null)
			mInputHandler.OnKeyUp(aVirtualKey);
	}
}



extern (Windows) LRESULT MainWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_CREATE:
			CREATESTRUCT *cs = cast(CREATESTRUCT*)lParam;
			SetLastError(0);
			LONG_PTR Result = SetWindowLongPtrW(hwnd, GWLP_USERDATA, cast(LONG_PTR)cs.lpCreateParams);
			if (Result == 0 && GetLastError() != 0)
				throw new Exception("SetWindowLongPtr() failed");
			return 0;
		case WM_DESTROY:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			Window.OnWindowDestroyed();
			return 0;
		case WM_CLOSE:
			DestroyWindow(hwnd);
			return 0;
		case WM_MOUSEMOVE:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			//TODO: should be GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam)
			Window.OnMouseMove(LOWORD(lParam), HIWORD(lParam));
			return 0;
		case WM_SETCURSOR:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			if (Window.OnSetCursor())
				return TRUE;
			break;
		case WM_PAINT:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			if (Window.OnPaint())
				return 0;
			break;
		case WM_KEYDOWN:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			Window.OnKeyDown(cast(uint)wParam);
			return 0;
		case WM_KEYUP:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CMainWindow Window = cast(CMainWindow)(cast(LPVOID)Pointer);
			Window.OnKeyUp(cast(uint)wParam);
			return 0;
		default:
			break;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}