﻿module Shaders.CShaderModule;
import std.conv, std.stream;
import derelict.opengl3.gl3;
import Engine.CEngine, Engine.CTractorException, Shaders.EBuiltinShader,
		Engine.IDisplay, Engine.IFileSystem, Engine.ILogger, Tractor.Shaders, Shaders.Shaders;

class CShaderModule : IShaderModule
{
	private IDisplay mDisplay;
	private IFileSystem mFileSystem;
	private ILogger mLogger;
	private CBuiltinVertexShader[EBuiltinShader] mBuiltinVertexShaders;
	private CBuiltinFragmentShader[EBuiltinShader] mBuiltinFragmentShaders;
	private CShaderProgram[EBuiltinShader] mBuiltinPrograms;
	private CCustomVertexShader[string] mCustomVertexShaders;
	private CCustomFragmentShader[string] mCustomFragmentShaders;
	private CShaderProgram[] mCustomPrograms;

	public this(IDisplay aDisplay, IFileSystem aFileSystem, ILogger aLogger)
	{
		mDisplay = aDisplay;
		mFileSystem = aFileSystem;
		mLogger = aLogger;
	}



	public ~this()
	{
		glUseProgram(0);
		foreach (CShaderProgram Program; mCustomPrograms)
			delete Program;
		foreach (CCustomVertexShader Shader; mCustomVertexShaders)
			delete Shader;
		foreach (CCustomFragmentShader Shader; mCustomFragmentShaders)
			delete Shader;
		foreach (CShaderProgram Program; mBuiltinPrograms)
			delete Program;
		foreach (CBuiltinVertexShader Shader; mBuiltinVertexShaders)
			delete Shader;
		foreach (CBuiltinFragmentShader Shader; mBuiltinFragmentShaders)
			delete Shader;
	}



	public CBuiltinVertexShader LoadBuiltinVertexShader(EBuiltinShader aShader)
	{
		if (!(aShader in mBuiltinVertexShaders))
		{
			GLuint VertexShader = glCreateShader(GL_VERTEX_SHADER);
			glShaderSource(VertexShader, 1, cast(GLchar**)[ eSimplestVertex.ptr ].ptr, null);
			glCompileShader(VertexShader);
			GLint CompileStatus;
			glGetShaderiv(VertexShader, GL_COMPILE_STATUS, &CompileStatus);
			if (CompileStatus == GL_FALSE)
			{
				glDeleteShader(VertexShader);
				mLogger.Log("Failed to compile built-in vertex shader: " ~ to!string(aShader));
				// TODO: log details about where compile error was?
				throw new CTractorException();
			}
			assert(glGetError() == GL_NO_ERROR);
			mBuiltinVertexShaders[aShader] = new CBuiltinVertexShader(VertexShader);
		}
		return mBuiltinVertexShaders[aShader];
	}



	public CBuiltinFragmentShader LoadBuiltinFragmentShader(EBuiltinShader aShader)
	{
		if (!(aShader in mBuiltinFragmentShaders))
		{
			GLuint FragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
			glShaderSource(FragmentShader, 1, cast(GLchar**)[ eSimplestFragment.ptr ].ptr, null);
			glCompileShader(FragmentShader);
			GLint CompileStatus;
			glGetShaderiv(FragmentShader, GL_COMPILE_STATUS, &CompileStatus);
			if (CompileStatus == GL_FALSE)
			{
				glDeleteShader(FragmentShader);
				mLogger.Log("Failed to compile built-in fragment shader: " ~ to!string(aShader));
				// TODO: log details about where compile error was?
				throw new CTractorException();
			}
			mBuiltinFragmentShaders[aShader] = new CBuiltinFragmentShader(FragmentShader);
		}
		return mBuiltinFragmentShaders[aShader];
	}



	private ubyte[] LoadShaderFile(string aFileName)
	{
		File ShaderFile;
		try
		{
			ShaderFile = new File(mFileSystem.GetDataFilePath(aFileName), FileMode.In);
		}
		catch (OpenException oe)
		{
			mLogger.Log("Shader file not found: " ~ aFileName);
			throw new CTractorException();
		}
		ubyte[] Data = new ubyte[cast(uint)ShaderFile.size + 1];
		ShaderFile.read(Data);
		Data[$ - 1] = 0;
		ShaderFile.close();
		return Data;
	}



	public CCustomVertexShader LoadCustomVertexShader(string aFileName)
	{
		if (!(aFileName in mCustomVertexShaders))
		{
			ubyte[] Data = LoadShaderFile(aFileName);
			GLuint VertexShader = glCreateShader(GL_VERTEX_SHADER);
			glShaderSource(VertexShader, 1, cast(GLchar**)[ Data.ptr ].ptr, null);
			glCompileShader(VertexShader);
			GLint CompileStatus;
			glGetShaderiv(VertexShader, GL_COMPILE_STATUS, &CompileStatus);
			if (CompileStatus == GL_FALSE)
			{
				glDeleteShader(VertexShader);
				mLogger.Log("Failed to compile custom vertex shader: " ~ aFileName);
				// TODO: log details about where compile error was?
				throw new CTractorException();
			}
			mCustomVertexShaders[aFileName] = new CCustomVertexShader(VertexShader);
		}
		return mCustomVertexShaders[aFileName];
	}



	public CCustomFragmentShader LoadCustomFragmentShader(string aFileName)
	{
		if (!(aFileName in mCustomFragmentShaders))
		{
			ubyte[] Data = LoadShaderFile(aFileName);
			GLuint FragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
			glShaderSource(FragmentShader, 1, cast(GLchar**)[ Data.ptr ].ptr, null);
			glCompileShader(FragmentShader);
			GLint CompileStatus;
			glGetShaderiv(FragmentShader, GL_COMPILE_STATUS, &CompileStatus);
			if (CompileStatus == GL_FALSE)
			{
				glDeleteShader(FragmentShader);
				mLogger.Log("Failed to compile custom fragment shader: " ~ aFileName);
				// TODO: log details about where compile error was?
				throw new CTractorException();
			}
			mCustomFragmentShaders[aFileName] = new CCustomFragmentShader(FragmentShader);
		}
		return mCustomFragmentShaders[aFileName];
	}



	private CShaderProgram CreateProgram(GLuint aVertexShader, GLuint aFragmentShader)
	{
		GLuint Program = glCreateProgram();
		glAttachShader(Program, aVertexShader);
		glAttachShader(Program, aFragmentShader);
		assert(glGetError() == GL_NO_ERROR);
		return new CShaderProgram(mLogger, Program);
	}



	public uint CreateBuiltinProgram(EBuiltinShader aProgram)
	{
		auto VertexShader = LoadBuiltinVertexShader(aProgram);
		auto FragmentShader = LoadBuiltinFragmentShader(aProgram);
		auto Program = CreateProgram(VertexShader.ID, FragmentShader.ID);
		Program.Link();
		mBuiltinPrograms[aProgram] = Program;
		return Program.ID;
	}



	public CShaderProgram CreateCustomProgram(IVertexShader aVertexShader,
			IFragmentShader aFragmentShader)
	{
		auto Program = CreateProgram(aVertexShader.ID, aFragmentShader.ID);
		mCustomPrograms ~= Program;
		return Program;
	}



	public void UseBuiltinProgram(EBuiltinShader aProgram)
	{
		auto Program = mBuiltinPrograms[aProgram];
		glUseProgram(Program.ID);
		switch (aProgram)
		{
			case EBuiltinShader.Simplest:
				int DisplayXSize;
				int DisplayYSize;
				mDisplay.GetSize(DisplayXSize, DisplayYSize);
				GLint ScreenXSizeLocation = glGetUniformLocation(Program.ID, "ScreenXSize");
				glUniform1f(ScreenXSizeLocation, DisplayXSize);
				GLint ScreenYSizeLocation = glGetUniformLocation(Program.ID, "ScreenYSize");
				glUniform1f(ScreenYSizeLocation, DisplayYSize);
				break;
			default:
				break;
		}
	}



	public void UseCustomProgram(IShaderProgram aProgram)
	{
		glUseProgram(aProgram.ID);

		// TODO: only do this for Simplest shader. although it might be useful for other
		// shaders as well?
		int DisplayXSize;
		int DisplayYSize;
		mDisplay.GetSize(DisplayXSize, DisplayYSize);
		GLint ScreenXSizeLocation = glGetUniformLocation(aProgram.ID, "ScreenXSize");
		glUniform1f(ScreenXSizeLocation, DisplayXSize);
		GLint ScreenYSizeLocation = glGetUniformLocation(aProgram.ID, "ScreenYSize");
		glUniform1f(ScreenYSizeLocation, DisplayYSize);
	}
}



private enum string eSimplestVertex =
"#version 140\n"
"\n"
"uniform float ScreenXSize;\n"
"uniform float ScreenYSize;\n"
"in vec2 Position;\n"
"\n"
"void main()\n"
"{\n"
"\tgl_Position = vec4((Position.x + 0.5) * 2.0 / ScreenXSize - 1.0, 1.0 - (Position.y + 0.5) * 2.0 / ScreenYSize, 0.0, 1.0);\n"
"}";

private enum string eSimplestFragment =
"#version 140\n"
"\n"
"out vec4 FragColor;\n"
"\n"
"void main()\n"
"{\n"
"\tFragColor = vec4(1.0, 1.0, 1.0, 1.0);\n"
"}";